/*-
 * Copyright (C) 2002-2003 Juergen Keil.
 * Copyright (C) 2002-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Copyright (c) 2001 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Yannick Montulet.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * A Solaris audio driver for CreativeLabs PCI sound cards based on
 * the EMU10Kx chipset.
 *
 * Creative technical specifications: NONE :-(
 *
 * A highlevel description of the card and the FX 8010 DSP engine can
 * be found in the following papers:
 *
 * The EMU10K1 Digital Audio Processor, Thomas C. Savell, Joint
 * E-mu/Creative Technology Center, 1999
 * ftp://ftp.alsa-project.org/pub/datasheets/creative/m2049.pdf
 *
 * FX8010 - A DSP Chip Architecture for Audio Effects, Steve Hoge,
 * Joint E-mu/Creative Technology Center
 * ftp://ftp.alsa-project.org/pub/datasheets/creative/HOG63.PS.gz
 */


#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#if	SUNOS >= 509
#include <sys/audio/ac97.h>
#else
#include <sys/ac97.h>
#endif
#include <sys/audio.h>
#include "audioio_compat.h"
#include "ac97_codec.h"
#include "audioemu.h"

#include <sys/ddi.h>

/*
 * hide gcc2 __attribute__ compiler extension from non-gcc compiler
 * and old gcc compiler
 */
#if     !defined(__GNUC__)
#define __attribute__(x)        /**/
#elif   __GNUC__ < 2
#define __attribute__(x)        /**/
#endif



static int audioemu_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int audioemu_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

static void audioemu_voice_halt(struct audioemu_voice *voice);

static audio_info_t audioemu_defaults;



/*
 * basic interface stuff for a STREAMS driver
 *
 * register the audio helper framework (audio_hlp_*) to handle all the
 * dirty low level STREAMS work.
 */

static struct module_info audioemu_minfo = {
    0x2710,		/* module ID number (0x2710 == 10000) */
    "audemu",		/* module name */
    0,			/* minimum packet size accepted */
    8192,		/* maximum packet size accepted */
    48000*4*2,		/* hi-water mark: ~2 secs of audio */
    0x8000		/* lo-water mark */
};


static struct qinit audioemu_rqueue = {
    audio_hlp_rput,
    audio_hlp_rsvc,
    audio_hlp_open,
    audio_hlp_close,
    NULL,
    &audioemu_minfo
};

static struct qinit audioemu_wqueue = {
    audio_hlp_wput,
    audio_hlp_wsvc,
    NULL,
    NULL,
    NULL,
    &audioemu_minfo
};


static struct streamtab audioemu_str_info = {
    &audioemu_rqueue,
    &audioemu_wqueue,
    NULL,
    NULL,
};


/*
 * char/block ops vector for a STREAMS driver
 */
static struct cb_ops audioemu_cb_ops = {
    nulldev,			/* cb_open */
    nulldev,			/* cb_close */
    nodev,			/* cb_strategy */
    nodev,			/* cb_print */
    nodev,			/* cb_dump */
    nodev,			/* cb_read */
    nodev,			/* cb_write */
    nodev,			/* cb_ioctl */
    nodev,			/* cb_devmap */
    nodev,			/* cb_mmap */
    nodev,			/* cb_segmap */
    nochpoll,			/* cb_chpoll */
    ddi_prop_op,		/* cb_prop_op */
    &audioemu_str_info,		/* cb_str */
    D_NEW|D_MTSAFE|D_64BIT,	/* cb_flag */
    CB_REV,			/* cb_rev */
    nodev,			/* cb_aread */
    nodev,			/* cb_awrite */
};

static struct dev_ops  audioemu_drv_ops = {
    DEVO_REV,
    0,

    audio_hlp_getinfo,
    nulldev,			/* devo_identify */
    nulldev,			/* devo_probe */
    audioemu_attach,
    audioemu_detach,
    nodev,			/* devo_reset */

    &audioemu_cb_ops,
    NULL,			/* devo_bus_ops */
    NULL,			/* devo_power */
};

/*
 * Module linkage information for the kernel.
 */
static struct modldrv audioemu_modldrv = {
    &mod_driverops,		/* Type of module.  This one is a driver */
    "EMU10Kx audio driver",
    &audioemu_drv_ops		/* driver ops */
};


static struct modlinkage audioemu_modlinkage = {
    MODREV_1,
    { &audioemu_modldrv, NULL },
};


/*
 * This describes the Audigy's DMA engines to the OS at hardware level
 * 
 * The first structure is for the "page table".  It consists of a
 * 32-Kbyte block of contiguous physical memory.  There are 8192
 * 4-byte entries in this page table, each pointing to 4096 byte buffer
 * for sample data.  So, the card can access a total of 8192*4K = 32MB
 * of sample data.
 *
 * The Audigy can perform 31-bit (?) PCI busmaster DMA. 
 *
 * (Physical addresses are stored in 32-bit words in the card / page
 * table shifted left one bit position, so I doubt that card can
 * access the full 32-bit physical address range.  Maybe the page
 * table base address and the recording buffers are an exception, but
 * for now let's just limit the dma addr_high to 31 bits for all DMA
 * accesses)
 */
static ddi_dma_attr_t ptb_dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0xffffffff,			/* DMA counter register */
    EMU_DMA_ALIGN,		/* DMA address alignment */
    				/*   align ptb on a 4K page boundary */
    0x3c,			/* DMA burstsizes ?? */
    4,				/* minxfer effective DMA size */
    0xffffffff,			/* max DMA xfer size */
    0xffffffff,			/* segment boundary*/
    1,				/* s/g length, contiguous */
    4,				/* granularity of device */
				/*   size of a ptb entry?? */
    0				/* flags */
};


/*
 * DMA attributes for playback sample data.  Multiple contiguous 4K
 * memory blocks each aligned on a 4K boundary.
 */
static ddi_dma_attr_t pmem_dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
#if	EMU_32BIT_PTE
    0xffffffff,			/* addr_high */
#else
    0x7fffffff,			/* addr_high */
				/*   due to the format of the PTE entries, */
				/*   playback mem must be located <2GB     */
#endif
    0xffffffff,			/* DMA counter register */
    EMU_DMA_ALIGN,		/* DMA address alignment */
    				/*   align sample data on 4k pages */
    0x3c,			/* DMA burstsizes */
    4,				/* minxfer effective DMA size */
    0xffffffff,			/* max DMA xfer size */
    0xffffffff,			/* segment boundary*/
    EMU_MAXPTE,			/* s/g length */
    4,				/* granularity of device */
    0				/* flags */
};

/*
 * DMA attributes for record sample data.  A single contiguous memory
 * blocks on a 4K boundary.
 */
static ddi_dma_attr_t rmem_dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0xffffffff,			/* DMA counter register */
    EMU_DMA_ALIGN,		/* DMA address alignment */
    				/*   align sample data on 4k pages */
    0x3c,			/* DMA burstsizes */
    4,				/* minxfer effective DMA size */
    0xffffffff,			/* max DMA xfer size */
    0xffffffff,			/* segment boundary*/
    1,				/* s/g length, contiguous */
    4,				/* granularity of device */
    0				/* flags */
};


/*
 * This describes things like byteorder internally for the Audigy 
 *
 * For the PTB data, we use the ddi_{get,put} access routines and let
 * them handle the byte order.
 */
static struct ddi_device_acc_attr dev_attr = {
    DDI_DEVICE_ATTR_V0,
    DDI_STRUCTURE_LE_ACC,
    DDI_STRICTORDER_ACC
};

/* byte swapping for pages with sample data is handled in this driver */
static struct ddi_device_acc_attr dev_sample_attr = {
    DDI_DEVICE_ATTR_V0,
    DDI_NEVERSWAP_ACC,
    /*DDI_STRICTORDER_ACC*/ DDI_MERGING_OK_ACC
};


/* available log levels */
#define	AUDEMU_LOG_L0	0		/* warnings, console & syslog buffer */
#define	AUDEMU_LOG_L1	1		/* errors, syslog buffer */
#define	AUDEMU_LOG_L2	2		/* recoverable errors, debug only */
#define	AUDEMU_LOG_L3	3		/* interesting data, debug only */
#define	AUDEMU_LOG_L4	4		/* tracing, debug only */


/* available log masks */
#define	AUDEMUM_AC97	0x00000020	/* AC97 register access */
#define	AUDEMUM_CODEC	0x00000010	/* CODEC IO */
#define	AUDEMUM_INTR	0x00000008	/* interrupt handling */
#define	AUDEMUM_DRIVER	0x00000004	/* audio driver functions */
#define	AUDEMUM_ATTACH	0x00000002	/* card setup at attach time */
#define	AUDEMUM_MODS	0x00000001	/* _init, _info, _fini */
#define	AUDEMUM_ALL	(~0u)

uint_t audioemu_errlevel = AUDEMU_LOG_L2;
uint_t audioemu_errmask =  AUDEMUM_ALL;


/* logging for log levels 4..0 */
#if	DEBUG
#define	DPRINTF4(...)	audioemu_dprintf4(__VA_ARGS__)
#define	DPRINTF3(...)	audioemu_dprintf3(__VA_ARGS__)
#define	DPRINTF2(...)	audioemu_dprintf2(__VA_ARGS__)
#else
#define	DPRINTF4(...)	/**/
#define	DPRINTF3(...)	/**/
#define	DPRINTF2(...)	/**/
#endif
#define	DPRINTF1(...)	audioemu_dprintf1(__VA_ARGS__)
#define	DPRINTF0(...)	audioemu_dprintf0(__VA_ARGS__)


static void
audioemu_vprintf(uint_t level, uint_t mask, char *fmt, va_list args)
{
    if (level <= audioemu_errlevel && (mask & audioemu_errmask))
	vcmn_err(CE_CONT, fmt, args);
}


#ifdef	DEBUG
void __attribute__ ((format (__printf__, 2, 3)))
audioemu_dprintf4(uint_t mask, char *fmt, ...)

{
    va_list args;
    va_start(args, fmt);
    audioemu_vprintf(AUDEMU_LOG_L4, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audioemu_dprintf3(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audioemu_vprintf(AUDEMU_LOG_L3, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audioemu_dprintf2(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audioemu_vprintf(AUDEMU_LOG_L2, mask, fmt, args);
    va_end(args);
}
#endif

void __attribute__ ((format (__printf__, 2, 3)))
audioemu_dprintf1(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audioemu_vprintf(AUDEMU_LOG_L1, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audioemu_dprintf0(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audioemu_vprintf(AUDEMU_LOG_L0, mask, fmt, args);
    va_end(args);
}


/* 
 * Global softstate for multiple emu audio controller instances
 */
static void *audioemu_state;


/* 
 * Sun driver module initialization and finalization code 
 */
int 
_init(void)
{
    int err;

    err = ddi_soft_state_init(&audioemu_state,
			      sizeof(struct audioemu_softstate), 1);
    if (err == 0) {
	if ((err = mod_install(&audioemu_modlinkage)) != 0)
	    ddi_soft_state_fini(&audioemu_state);
    }

    DPRINTF4(AUDEMUM_MODS, "_init, err=%d\n", err);

    return err;
}


int
_fini(void)
{
    int err;
    
    if ((err = mod_remove(&audioemu_modlinkage)) == 0)
	ddi_soft_state_fini(&audioemu_state);

    DPRINTF4(AUDEMUM_MODS, "_fini, err=%d\n", err);

    return err;
}


int
_info(struct modinfo *modinfop)
{
    return mod_info(&audioemu_modlinkage, modinfop);
}


/*
 * Dma memory mgmt
 */

static void
dmamem_delete(struct dmamem *mem)
{
    if (!mem)
	return;

    if (mem->acc_handle != NULL) {
	/* free DMA memory */
	if (mem->dma_bound) {
	    ddi_dma_unbind_handle(mem->dma_handle);
	    mem->dma_bound = 0;
	}
	ddi_dma_mem_free(&mem->acc_handle);
	mem->kaddr = NULL;
    }

    if (mem->dma_handle) {
	ddi_dma_free_handle(&mem->dma_handle);
    }

    kmem_free(mem, sizeof(struct dmamem));
}


static struct dmamem *
dmamem_alloc(struct audioemu_softstate *softstate,
	     ddi_dma_attr_t	       *dma_attr,
	     ddi_device_acc_attr_t     *dev_attr,
	     size_t			size)
{
    struct dmamem      *mem;
    int			ret;

    mem = kmem_zalloc(sizeof(struct dmamem), KM_SLEEP);
    
    if ((ret = ddi_dma_alloc_handle(softstate->dip, dma_attr,
				    DDI_DMA_DONTWAIT, NULL,
				    &mem->dma_handle)) != DDI_SUCCESS) {
	DPRINTF2(AUDEMUM_ATTACH,
		 "dmamem_alloc: alloc handle failed (%d)\n", ret);
	/*
	 * on sparc dma_alloc_handle() was failing with error
	 * DDI_DMA_BADATTR, but mem->dma_handle was initialized
	 * to a non-zero value.  Don't ddi_dma_free_handle() this
	 * bogus dma handle in our dmamem_delete() function.
	 */
	mem->dma_handle = 0;
	goto error;
    }

    if ((ret = ddi_dma_mem_alloc(mem->dma_handle, size, dev_attr,
				 DDI_DMA_CONSISTENT,
				 DDI_DMA_DONTWAIT, NULL,
				 &mem->kaddr, &mem->real_len,
				 &mem->acc_handle)) != DDI_SUCCESS) {
	DPRINTF2(AUDEMUM_ATTACH,
		 "dmamem_alloc: dma mem alloc of %lu bytes failed (%d)\n",
		 (unsigned long)size, ret);
	goto error;
    }
    mem->len = size;

    if ((ret = ddi_dma_addr_bind_handle(mem->dma_handle,
					NULL,
					mem->kaddr, mem->real_len,
					DDI_DMA_RDWR|DDI_DMA_CONSISTENT,
					DDI_DMA_DONTWAIT, NULL,
					&mem->cookie, &mem->ccount)) != DDI_DMA_MAPPED) {
	DPRINTF2(AUDEMUM_ATTACH, 
		 "dmamem_alloc: bind failed (%d)\n", ret);
	goto error;
    }
    mem->dma_bound = TRUE;

    /*
     * We should receive exactly one physical page mapping for the PTB
     * and the record sample buffers (=contiguous physical memory)
     */
    if (dma_attr->dma_attr_sgllen == 1 && mem->ccount != 1) {
	DPRINTF1(AUDEMUM_ATTACH, 
		 "unexpected dma cookie count > 1 (%u)\n", mem->ccount);
	goto error;
    }

    /*
     * for convenience, save the physical address of the first cookie.
     * this will only be useful for contiguous mappings (PTB & record buf)
     */
    mem->paddr = mem->cookie.dmac_address;
    return mem;

 error:
    dmamem_delete(mem);
    return NULL;
}


#define	PCI_VENDOR_CREATIVE	0x1102
#define	PCI_PRODUCT_CREATIVE_SBLIVE	0x0002
#define	PCI_PRODUCT_CREATIVE_AUDIGY	0x0004

/*
 * Configure the PCI device, so that the device is ready to 
 * accept I/O requests.  This includes enabling PCI busmaster,
 * and anabling access to the device's I/O registers.
 */
static int
audioemu_pci_setup(struct audioemu_softstate *softstate)
{
    dev_info_t *dip = softstate->dip;
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;    
    uint16_t pci_status;
    uint16_t pci_vendorid;
    uint16_t pci_deviceid;
    uint8_t  pci_revision;
    uint8_t  pci_intrinfo;
    uint16_t pci_commandreg;
    char *revstr = "";

    pci_status   = pci_config_get16(conf_handle, PCI_CONF_STAT);
    pci_vendorid = pci_config_get16(conf_handle, PCI_CONF_VENID);
    pci_deviceid = pci_config_get16(conf_handle, PCI_CONF_DEVID);
    pci_revision = pci_config_get8 (conf_handle, PCI_CONF_REVID);
    pci_intrinfo = pci_config_get8 (conf_handle, PCI_CONF_ILINE);    
    
    /* try to find out a nice hardware name for the given pci hw */
    softstate->audigy = 0;
    if (pci_vendorid == PCI_VENDOR_CREATIVE) {
	switch (pci_deviceid) {
	case PCI_PRODUCT_CREATIVE_SBLIVE:
	    revstr = "SB Live!"; break;
	case PCI_PRODUCT_CREATIVE_AUDIGY:
	    if (pci_revision == 0x04) {
		revstr = "SB Audigy 2";
		softstate->audigy = 2;
	    } else {
		revstr = "SB Audigy";
		softstate->audigy = 1;
	    }
	    break;
	}
    }
    softstate->rev = pci_revision;
    softstate->revstr = revstr;

    /* Announce the hw we're attaching to */
    ddi_report_dev(dip);
    cmn_err(CE_CONT, "?PCI-device: %s, audioemu%d, %s%s(rev. 0x%02x) at irq %d\n",
	    ddi_node_name(dip),
	    ddi_get_instance(dip), 
	    revstr, revstr[0] ? " " : "", 
	    pci_revision,
	    pci_intrinfo);

#if i386
    /*
     * reject interrupt line 255, per PCI spec 6.2.4, footnote 43,
     * this means "unknown" or "not connected" to the interrupt
     * controller on x86 based PCs.
     * Sun drivers (iprb) reject iline 0 as well.
     */
    if (pci_intrinfo == 0 || pci_intrinfo == 255) {
	cmn_err(CE_WARN, "interrupt line value out of range: %d", pci_intrinfo);
	return AUDIOEMU_FAILURE;
    }
#endif
    
    pci_commandreg = pci_config_get16(conf_handle, PCI_CONF_COMM);
    DPRINTF4(AUDEMUM_ATTACH, "pci command register 0x%x %s%s%s%s",
	    pci_commandreg,
	    pci_commandreg & PCI_COMM_SERR_ENABLE ? " serr#_en"  : "",
	    pci_commandreg & PCI_COMM_ME  ? " busmaster" : "",
	    pci_commandreg & PCI_COMM_MAE ? " mem_access" : "",
	    pci_commandreg & PCI_COMM_IO  ? " io_access" : "");
    if ((pci_commandreg & (PCI_COMM_IO|PCI_COMM_ME)) == (PCI_COMM_IO|PCI_COMM_ME)) {
	DPRINTF4(AUDEMUM_ATTACH, "busmastering and io_access already enabled");
    } else {
	DPRINTF4(AUDEMUM_ATTACH, "Enabling PCI I/O and bus mastering access");
	pci_config_put16(conf_handle, PCI_CONF_COMM, pci_commandreg | PCI_COMM_IO | PCI_COMM_ME);
    }

    return AUDIOEMU_SUCCESS;
}


/*
 * Read an AC97 primary codec mixer register
 *
 * Returns TRUE on success with the data read returned via the
 * *data pointer.  Returns FALSE in case the primary codec is not
 * ready any more for some reason.
 */
static int
audioemu_ac97_read(void *closure, uint16_t sub_addr, uint16_t *data)
{
    struct audioemu_softstate *softstate = closure;

    /*
     * may need a lock around these two register accesses, but at this
     * time the ac97 registers are only accessed during driver
     * initialization and from STREAMS callbacks; so there should be
     * no concurrent accesses.
     */
    EMU_PUT8(EMU_AC97ADDR, sub_addr);
    *data = EMU_GET16(EMU_AC97DATA);
    return TRUE;
}


/*
 * Write to an AC97 primary codec mixer register.
 *
 * Returns TRUE on success. Returns FALSE in case the primary codec is
 * not ready any more for some reason.
 */
static int
audioemu_ac97_write(void *closure, uint16_t sub_addr, uint16_t value)
{
    struct audioemu_softstate *softstate = closure;

    /* may need a lock around these two register accesses, see above */
    EMU_PUT8(EMU_AC97ADDR, sub_addr);
    EMU_PUT16(EMU_AC97DATA, value);

    return TRUE;
}


/*
 * reset the Audigy audiocontroller's AC'97 link and reset the AC'97
 * codec to default values
 */
static int
audioemu_reset_ac97(struct audioemu_softstate *softstate,
		    audio_info_t	      *audio_info,
		    int			       verbose)
{
    if (!CODEC_RESET(softstate->codec_handle, audio_info, verbose)) {
	DPRINTF2(AUDEMUM_CODEC, "Audio CODEC reset failed");
	return AUDIOEMU_FAILURE;
    }

    audioemu_ac97_read (softstate, AC97_EXTENDED_AUDIO_REGISTER, &softstate->ac97_ext_features);

    return AUDIOEMU_SUCCESS;
}


/* Misc stuff relative to emu10k1 */

static uint32_t
audioemu_rate_to_pitch(uint32_t rate)
{
    static const uint32_t logMagTable[128] = {
	0x00000, 0x02dfc, 0x05b9e, 0x088e6, 0x0b5d6, 0x0e26f, 0x10eb3, 0x13aa2,
	0x1663f, 0x1918a, 0x1bc84, 0x1e72e, 0x2118b, 0x23b9a, 0x2655d, 0x28ed5,
	0x2b803, 0x2e0e8, 0x30985, 0x331db, 0x359eb, 0x381b6, 0x3a93d, 0x3d081,
	0x3f782, 0x41e42, 0x444c1, 0x46b01, 0x49101, 0x4b6c4, 0x4dc49, 0x50191,
	0x5269e, 0x54b6f, 0x57006, 0x59463, 0x5b888, 0x5dc74, 0x60029, 0x623a7,
	0x646ee, 0x66a00, 0x68cdd, 0x6af86, 0x6d1fa, 0x6f43c, 0x7164b, 0x73829,
	0x759d4, 0x77b4f, 0x79c9a, 0x7bdb5, 0x7dea1, 0x7ff5e, 0x81fed, 0x8404e,
	0x86082, 0x88089, 0x8a064, 0x8c014, 0x8df98, 0x8fef1, 0x91e20, 0x93d26,
	0x95c01, 0x97ab4, 0x9993e, 0x9b79f, 0x9d5d9, 0x9f3ec, 0xa11d8, 0xa2f9d,
	0xa4d3c, 0xa6ab5, 0xa8808, 0xaa537, 0xac241, 0xadf26, 0xafbe7, 0xb1885,
	0xb3500, 0xb5157, 0xb6d8c, 0xb899f, 0xba58f, 0xbc15e, 0xbdd0c, 0xbf899,
	0xc1404, 0xc2f50, 0xc4a7b, 0xc6587, 0xc8073, 0xc9b3f, 0xcb5ed, 0xcd07c,
	0xceaec, 0xd053f, 0xd1f73, 0xd398a, 0xd5384, 0xd6d60, 0xd8720, 0xda0c3,
	0xdba4a, 0xdd3b4, 0xded03, 0xe0636, 0xe1f4e, 0xe384a, 0xe512c, 0xe69f3,
	0xe829f, 0xe9b31, 0xeb3a9, 0xecc08, 0xee44c, 0xefc78, 0xf148a, 0xf2c83,
	0xf4463, 0xf5c2a, 0xf73da, 0xf8b71, 0xfa2f0, 0xfba57, 0xfd1a7, 0xfe8df
    };
    static const uint8_t logSlopeTable[128] = {
	0x5c, 0x5c, 0x5b, 0x5a, 0x5a, 0x59, 0x58, 0x58,
	0x57, 0x56, 0x56, 0x55, 0x55, 0x54, 0x53, 0x53,
	0x52, 0x52, 0x51, 0x51, 0x50, 0x50, 0x4f, 0x4f,
	0x4e, 0x4d, 0x4d, 0x4d, 0x4c, 0x4c, 0x4b, 0x4b,
	0x4a, 0x4a, 0x49, 0x49, 0x48, 0x48, 0x47, 0x47,
	0x47, 0x46, 0x46, 0x45, 0x45, 0x45, 0x44, 0x44,
	0x43, 0x43, 0x43, 0x42, 0x42, 0x42, 0x41, 0x41,
	0x41, 0x40, 0x40, 0x40, 0x3f, 0x3f, 0x3f, 0x3e,
	0x3e, 0x3e, 0x3d, 0x3d, 0x3d, 0x3c, 0x3c, 0x3c,
	0x3b, 0x3b, 0x3b, 0x3b, 0x3a, 0x3a, 0x3a, 0x39,
	0x39, 0x39, 0x39, 0x38, 0x38, 0x38, 0x38, 0x37,
	0x37, 0x37, 0x37, 0x36, 0x36, 0x36, 0x36, 0x35,
	0x35, 0x35, 0x35, 0x34, 0x34, 0x34, 0x34, 0x34,
	0x33, 0x33, 0x33, 0x33, 0x32, 0x32, 0x32, 0x32,
	0x32, 0x31, 0x31, 0x31, 0x31, 0x31, 0x30, 0x30,
	0x30, 0x30, 0x30, 0x2f, 0x2f, 0x2f, 0x2f, 0x2f
    };
    int8_t          i;

    if (rate == 0)
	return 0;	/* Bail out if no leading "1" */
    rate *= 11185;		/* Scale 48000 to 0x20002380 */
    for (i = 31; i > 0; i--) {
	if (rate & 0x80000000) {	/* Detect leading "1" */
	    return (((uint32_t) (i - 15) << 20) +
		    logMagTable[0x7f & (rate >> 24)] +
		    (0x7f & (rate >> 17)) *
		    logSlopeTable[0x7f & (rate >> 24)]);
	}
	rate <<= 1;
    }

    return 0;		/* Should never reach this point */
}

/* Emu10k1 Low level */

/*
 * audioemu_{read/write}: reads/writes an "indirect" register on the
 * audigy.
 *
 * The lower 16 bits of "reg" contains a register number in the
 * range [0..7ff) (or [0..fff) for the audigy).
 *
 * The upper 16 bits of "reg" may contain an optional bitfield
 * descriptor, to restrict the read/write operation on a contiguous
 * block of bits of the given register.  The bits 16..20 contain
 * the bit position of the least significant bit in the bitfield,
 * and bits 24..29 contain the size of the bitfield.
 *
 * note: audioemu_{read/write} must be called with the intr_mutex locked,
 * to avoid concurrent access via the PTR/DATA register.
 */
static uint32_t
audioemu_read(struct audioemu_softstate *softstate,
	      uint16_t chano,
	      uint32_t reg)
{
    uint32_t       ptr, mask = 0xffffffff;
    uint8_t        size, offset = 0;
    uint32_t	   ptr_addr_mask;

    ptr_addr_mask = softstate->audigy 
	? EMU_A_PTR_ADDR_MASK : EMU_PTR_ADDR_MASK;
    ptr = ((reg << 16) & ptr_addr_mask) | (chano & EMU_PTR_CHNO_MASK);
    if (reg & 0xff000000) {
	/* bitfield position specified */
	size = (reg >> 24) & 0x3f;
	offset = (reg >> 16) & 0x1f;
	mask = ((1 << size) - 1) << offset;
    }

    EMU_PUT32(EMU_PTR, ptr);
    ptr = EMU_GET32(EMU_DATA);
    return (ptr & mask) >> offset;
}

static void
audioemu_write(struct audioemu_softstate *softstate,
	       uint16_t chano,
	       uint32_t reg,
	       uint32_t data)
{
    uint32_t       ptr, mask;
    uint8_t        size, offset;
    uint32_t	   ptr_addr_mask;

    ptr_addr_mask = softstate->audigy 
	? EMU_A_PTR_ADDR_MASK : EMU_PTR_ADDR_MASK;
    ptr = ((reg << 16) & ptr_addr_mask) | (chano & EMU_PTR_CHNO_MASK);
    if (reg & 0xff000000) {
	/* bitfield position specified */
	size = (reg >> 24) & 0x3f;
	offset = (reg >> 16) & 0x1f;
	mask = ((1 << size) - 1) << offset;
	data = ((data << offset) & mask) |
	    (audioemu_read(softstate, chano, reg & 0xffff) & ~mask);
    }

    EMU_PUT32(EMU_PTR, ptr);
    EMU_PUT32(EMU_DATA, data);
}


/* 
 * Update a 32-bit word of DSP Microcode 
 */
static void
audioemu_write_micro(struct audioemu_softstate *softstate,
		     uint32_t pc,
		     uint32_t data)
{
    pc += softstate->audigy ? EMU_A_MICROCODEBASE : EMU_MICROCODEBASE;
    audioemu_write(softstate, 0, pc, data);
}

/*
 * Store an FX8010 DSP 64-bit microcode instruction (EMU10K1 version)
 */
static void
audioemu_dsp_addop(struct audioemu_softstate *softstate,
		   uint16_t *pc, uint8_t op,
		   uint16_t r, uint16_t a, uint16_t x, uint16_t y)
{
    audioemu_write_micro(softstate, *pc << 1,
			 ((x << 10) & EMU_DSP_LOWORD_OPX_MASK) |
			 (y & EMU_DSP_LOWORD_OPY_MASK));
    audioemu_write_micro(softstate, (*pc << 1) + 1,
			 ((op << 20) & EMU_DSP_HIWORD_OPCODE_MASK) |
			 ((r << 10) & EMU_DSP_HIWORD_RESULT_MASK) |
			 (a & EMU_DSP_HIWORD_OPA_MASK));
    (*pc)++;
}


/*
 * Store an FX8010 DSP 64-bit microcode instruction (Audigy/EMU10K2
 * version)
 */
static void
audioemu_audigy_dsp_addop(struct audioemu_softstate *softstate,
			  uint16_t *pc, uint8_t op,
			  uint16_t r, uint16_t a, uint16_t x, uint16_t y)
{
    audioemu_write_micro(softstate, *pc << 1,
			 ((x << 12) & EMU_A_DSP_LOWORD_OPX_MASK) |
			 ( y        & EMU_A_DSP_LOWORD_OPY_MASK));
    audioemu_write_micro(softstate, (*pc << 1) + 1,
			 ((op << 24) & EMU_A_DSP_HIWORD_OPCODE_MASK) |
			 (( r << 12) & EMU_A_DSP_HIWORD_RESULT_MASK) |
			 (  a        & EMU_A_DSP_HIWORD_OPA_MASK));
    (*pc)++;
}


/* 
 * Init the FX8010 DSP (SB Live! / EMU10K1).
 * Init general purpose registers, "tank" mem and load microcode.
 */
static void
audioemu_initfx(struct audioemu_softstate *softstate)
{
    int		i;
    uint16_t    pc;

    DPRINTF4(AUDEMUM_ATTACH, "initfx: SB Live!\n");

    /* stop dsp while changing the microcode */
    audioemu_write(softstate, 0, EMU_DBG, EMU_DBG_SINGLE_STEP);

    /* Set all 256 GPRs to 0 */
    for (i = 0; i < 256; i++)
	audioemu_write(softstate, 0, EMU_DSP_GPR(i), 0);
    for (i = 0; i < 160; i++) {
	audioemu_write(softstate, 0, EMU_TANKMEMDATAREGBASE + i, 0);
	audioemu_write(softstate, 0, EMU_TANKMEMADDRREGBASE + i, 0);
    }

    pc = 0;
    /* AC97 Out (l/r) = AC97 In (l/r) + FX[0/1] * 4 */
    audioemu_dsp_addop(softstate, &pc, EMU_DSP_OP_MACINTS,
		       EMU_DSP_OUTL(EMU_DSP_OUT_AC97),
		       EMU_DSP_CST(0),
		       EMU_DSP_FX(0), EMU_DSP_CST(4));
    audioemu_dsp_addop(softstate, &pc, EMU_DSP_OP_MACINTS,
		       EMU_DSP_OUTR(EMU_DSP_OUT_AC97),
		       EMU_DSP_CST(0),
		       EMU_DSP_FX(1), EMU_DSP_CST(4));

#if 0
    /* Rear channel OUT (l/r) = FX[2/3] * 4 */
    audioemu_dsp_addop(softstate, &pc, EMU_DSP_OP_MACINTS,
		       EMU_DSP_OUTL(EMU_DSP_OUT_RCHAN),
		       EMU_DSP_OUTL(EMU_DSP_OUT_AC97),
		       EMU_DSP_FX(0), EMU_DSP_CST(4));
    audioemu_dsp_addop(softstate, &pc, EMU_DSP_OP_MACINTS,
		       EMU_DSP_OUTR(EMU_DSP_OUT_RCHAN),
		       EMU_DSP_OUTR(EMU_DSP_OUT_AC97),
		       EMU_DSP_FX(1), EMU_DSP_CST(4));
#endif
    /* zero out the rest of the microcode */
    while (pc < 512)
	audioemu_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			   EMU_DSP_CST(0), EMU_DSP_CST(0),
			   EMU_DSP_CST(0), EMU_DSP_CST(0));
    
    /* mircocode update is complete, restart dsp */
    audioemu_write(softstate, 0, EMU_DBG, 0);
}


/* 
 * Init the FX8010 DSP (Audigy version).
 * Init general purpose registers, "tank" mem and load microcode.
 */
static void
audioemu_audigy_initfx(struct audioemu_softstate *softstate)
{
    int		i;
    uint16_t    pc;

    DPRINTF4(AUDEMUM_ATTACH, "initfx: %s\n", softstate->revstr);

    /* stop dsp while changing the microcode */
    audioemu_write(softstate, 0, EMU_A_DBG, EMU_A_DBG_SINGLE_STEP);

    /* Set all GPRs to 0 */
    for (i = 0; i < 512; i++)
	audioemu_write(softstate, 0, EMU_A_DSP_GPR(i), 0);
    for (i = 0; i < 160; i++) {
	audioemu_write(softstate, 0, EMU_TANKMEMDATAREGBASE + i, 0);
	audioemu_write(softstate, 0, EMU_TANKMEMADDRREGBASE + i, 0);
    }

    pc = 0;

#if 1
#define	PVOLGPR	10
#define	RVOLGPR	12

    audioemu_write(softstate, 0, EMU_A_DSP_GPR(PVOLGPR+0), 0x7fffffff);
    audioemu_write(softstate, 0, EMU_A_DSP_GPR(PVOLGPR+1), 0x7fffffff);
    audioemu_write(softstate, 0, EMU_A_DSP_GPR(RVOLGPR+0), 0x7fffffff);
    audioemu_write(softstate, 0, EMU_A_DSP_GPR(RVOLGPR+1), 0x7fffffff);

    /*
     * AC'97 playback...
     */

    /* GPR 0/1 (l/r) := FX[0/1] * 1.0 */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_MACS,
			      EMU_A_DSP_GPR(0),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_FX(0), EMU_A_DSP_GPR(PVOLGPR+0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_MACS,
			      EMU_A_DSP_GPR(1),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_FX(1), EMU_A_DSP_GPR(PVOLGPR+1));
#define	PASSIVE_SPEAKERS	0
#if	PASSIVE_SPEAKERS
    /* Some extra boost, when using unamplified speakers */
    /* GPR 0/1 (l/r) := GPR 0/1 * 4.0 */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_MACINTS,
			      EMU_A_DSP_GPR(0),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(4));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_MACINTS,
			      EMU_A_DSP_GPR(1),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(4));
#endif
    if (softstate->audigy == 2) {
	/* Analog Front Out (l/r) := GPR 0/1 (l/r) */
	audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_AFRONT),
			      EMU_A_DSP_CST(0),
	    		      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
	audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_AFRONT),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
    } else {
	/* AC97 Out (l/r) := GPR 0/1 (l/r) */
	audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_AC97),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
	audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_AC97),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
    }
    /* Analog Rear Output (l/r) := GPR 0/1 (l/r) */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_AREAR),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_AREAR),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
    /* Analog Center + Subwoofer LFE := GPR 0/1 (l/r) */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_ACNT_LFE),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_ACNT_LFE),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));

#define	LOOPBACK_RECORD 0
#if	LOOPBACK_RECORD
    /*
     * loopback the audio output to the ADC recording buffer
     */

    /* ADC Capture (l/r) := GPR 0/1 (l/r) */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_ADC),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_ADC),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
#else
    /*
     * AC'97 recording...
     */

    /* GPR 0/1 (l/r) := AC'97 IN[0/1] * 1 */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_MACS,
			      EMU_A_DSP_GPR(0),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_INL(EMU_A_DSP_IN_AC97),
			      EMU_A_DSP_GPR(RVOLGPR+0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_MACS,
			      EMU_A_DSP_GPR(1),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_INR(EMU_A_DSP_IN_AC97),
			      EMU_A_DSP_GPR(RVOLGPR+1));
    /* ADC Capture (l/r) := GPR 0/1 (l/r) */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_ADC),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_ADC),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
#endif

#else
    /* GPR 0/1 (l/r) := AC97 In (l/r) + FX[0/1] * 1 */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_MACINTS,
			      EMU_A_DSP_GPR(0),
			      EMU_A_DSP_INL(EMU_A_DSP_IN_AC97),
			      EMU_A_DSP_FX(0), EMU_A_DSP_CST(1));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_MACINTS,
			      EMU_A_DSP_GPR(1),
			      EMU_A_DSP_INR(EMU_A_DSP_IN_AC97),
			      EMU_A_DSP_FX(1), EMU_A_DSP_CST(1));
    /* Digital Front Out (l/r) := GPR 0/1 (l/r) */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_DFRONT),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_DFRONT),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
    /* Analog Front Out (l/r) := GPR 0/1 (l/r) */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_AFRONT),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_AFRONT),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
    /* Analog Rear Out (l/r) := GPR 0/1 (l/r) */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_AREAR),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_AREAR),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
    /* Audigy Drive, Headphone (l/r) := GPR 0/1 (l/r) */
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTL(EMU_A_DSP_OUT_HEAD),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(0), EMU_A_DSP_CST(0));
    audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_ACC3,
			      EMU_A_DSP_OUTR(EMU_A_DSP_OUT_HEAD),
			      EMU_A_DSP_CST(0),
			      EMU_A_DSP_GPR(1), EMU_A_DSP_CST(0));
#endif

    /* zero out the rest of the microcode */
    while (pc < 1024)
	audioemu_audigy_dsp_addop(softstate, &pc, EMU_DSP_OP_SKIP,
				  EMU_A_DSP_CST(0),  EMU_A_DSP_CST(0),
				  EMU_A_DSP_CST(15), EMU_A_DSP_CST(0));
    
    /* mircocode update is complete, restart dsp */
    audioemu_write(softstate, 0, EMU_A_DBG, 0);
}


/*
 * Initialize Audigy hardware.
 */
static int
audioemu_hwinit(struct audioemu_softstate *softstate)
{
    int		     i;
    uint32_t	     spcs;
    uint32_t	     silentpage;
    ddi_acc_handle_t ptb_acc;
    uint32_t	    *ptb;
    uint32_t	     hcfg_ptb;

    softstate->ptb =
	dmamem_alloc(softstate, &ptb_dma_attr, &dev_attr,
		     EMU_MAXPTE * sizeof(uint32_t));
    if (!softstate->ptb) {
	DPRINTF4(AUDEMUM_ATTACH, "hwinit: ptb alloc failed\n");
	return AUDIOEMU_FAILURE;
    }

    DPRINTF4(AUDEMUM_DRIVER, "PTB at %p/%x, %d cookies\n",
	     softstate->ptb->kaddr, softstate->ptb->paddr,
	     softstate->ptb->ccount);

    softstate->silentpage = 
	dmamem_alloc(softstate, &pmem_dma_attr, &dev_sample_attr,
		     EMU_PTESIZE);
    if (!softstate->silentpage) {
	DPRINTF4(AUDEMUM_ATTACH, "hwinit: silentpage alloc failed\n");
	return AUDIOEMU_FAILURE;
    }

    DPRINTF4(AUDEMUM_DRIVER, "silentpage at %p/%x, %d cookies\n",
	     softstate->silentpage->kaddr, softstate->silentpage->paddr,
	     softstate->silentpage->ccount);

    if (!softstate->audigy && (softstate->silentpage->paddr & 0x80000000)) {
	cmn_err(CE_WARN, "invalid physical address for silentpage, failing");
	cmn_err(CE_NOTE, "SB Live! only works at physical addresses <2GB");
	return AUDIOEMU_FAILURE;
    }

    /* Zero out the silent page */
    bzero(softstate->silentpage->kaddr, EMU_PTESIZE);

    EMU_PUT32(EMU_HCFG,
	      EMU_HCFG_LOCKSOUNDCACHE | EMU_HCFG_LOCKTANKCACHE_MASK |
	      EMU_HCFG_AUTOMUTE);

    /* disable any channel interrupt */
    audioemu_write(softstate, 0, EMU_CLIEL, 0);
    audioemu_write(softstate, 0, EMU_CLIEH, 0);
    audioemu_write(softstate, 0, EMU_SOLEL, 0);
    audioemu_write(softstate, 0, EMU_SOLEH, 0);

    /* Set recording buffers sizes to zero */
    audioemu_write(softstate, 0, EMU_MICBS, EMU_RECBS_BUFSIZE_NONE);
    audioemu_write(softstate, 0, EMU_MICBA, 0);
    audioemu_write(softstate, 0, EMU_ADCBS, EMU_RECBS_BUFSIZE_NONE);
    audioemu_write(softstate, 0, EMU_ADCBA, 0);
    audioemu_write(softstate, 0, EMU_FXBS, EMU_RECBS_BUFSIZE_NONE);
    audioemu_write(softstate, 0, EMU_FXBA, 0);

    if (softstate->audigy) {
	audioemu_write(softstate, 0, EMU_SPBYPASS, EMU_SPBYPASS_24_BITS);
	audioemu_write(softstate, 0, EMU_AC97SLOT, EMU_AC97SLOT_CNTR | EMU_AC97SLOT_LFE);
    }

    /* Initialize all channels to stopped and no effects */
    for (i = 0; i < EMU_NUMCHAN; i++) {
	audioemu_write(softstate, i, EMU_CHAN_DCYSUSV, 0);
	audioemu_write(softstate, i, EMU_CHAN_IP, 0);
	audioemu_write(softstate, i, EMU_CHAN_VTFT, 0xffff);
	audioemu_write(softstate, i, EMU_CHAN_CVCF, 0xffff);
	audioemu_write(softstate, i, EMU_CHAN_PTRX, 0);
	audioemu_write(softstate, i, EMU_CHAN_CPF, 0);
	audioemu_write(softstate, i, EMU_CHAN_CCR, 0);
	audioemu_write(softstate, i, EMU_CHAN_CLP, 0);
	audioemu_write(softstate, i, EMU_CHAN_PSST, 0);
	audioemu_write(softstate, i, EMU_CHAN_DSL, 0x10);	/* Why 16 ? */
	audioemu_write(softstate, i, EMU_CHAN_CCCA, 0);
	audioemu_write(softstate, i, EMU_CHAN_Z1, 0);
	audioemu_write(softstate, i, EMU_CHAN_Z2, 0);
	audioemu_write(softstate, i, EMU_CHAN_FXRT, 0x32100000);
	if (softstate->audigy) {
	    audioemu_write(softstate, i, EMU_A_CHAN_FXRT1, 0x3f3f3f3f);
	    audioemu_write(softstate, i, EMU_A_CHAN_FXRT2, 0x3f3f3f3f);
	    audioemu_write(softstate, i, EMU_A_CHAN_SENDAMOUNTS, 0);
	}
	audioemu_write(softstate, i, EMU_CHAN_ATKHLDM, 0);
	audioemu_write(softstate, i, EMU_CHAN_DCYSUSM, 0);
	audioemu_write(softstate, i, EMU_CHAN_IFATN, 0xffff);
	audioemu_write(softstate, i, EMU_CHAN_PEFE, 0);
	audioemu_write(softstate, i, EMU_CHAN_FMMOD, 0);
	audioemu_write(softstate, i, EMU_CHAN_TREMFRQ, 24);
	audioemu_write(softstate, i, EMU_CHAN_FM2FRQ2, 24);
	audioemu_write(softstate, i, EMU_CHAN_TEMPENV, 0);
	
	/* these are last so OFF prevents writing */
	audioemu_write(softstate, i, EMU_CHAN_LFOVAL2, 0);
	audioemu_write(softstate, i, EMU_CHAN_LFOVAL1, 0);
	audioemu_write(softstate, i, EMU_CHAN_ATKHLDV, 0);
	audioemu_write(softstate, i, EMU_CHAN_ENVVOL, 0);
	audioemu_write(softstate, i, EMU_CHAN_ENVVAL, 0);
    }

    /* set digital outputs format */
    spcs = (EMU_SPCS_CLKACCY_1000PPM | EMU_SPCS_SAMPLERATE_48 |
	    EMU_SPCS_CHANNELNUM_LEFT | EMU_SPCS_SOURCENUM_UNSPEC |
	    EMU_SPCS_GENERATIONSTATUS | 0x00001200 /* Cat code. */ |
	    0x00000000 /* IEC-958 Mode */ | EMU_SPCS_EMPHASIS_NONE |
	    EMU_SPCS_COPYRIGHT);
    audioemu_write(softstate, 0, EMU_SPCS0, spcs);
    audioemu_write(softstate, 0, EMU_SPCS1, spcs);
    audioemu_write(softstate, 0, EMU_SPCS2, spcs);

    if (softstate->audigy == 2) {
	audioemu_write(softstate, 0, EMU_A2_SPDIF_SAMPLERATE, EMU_A2_SPDIF_UNKNOWN);

        EMU_PUT32(EMU_A2_PTR, EMU_A2_SRCSEL);
        EMU_PUT32(EMU_A2_DATA, EMU_A2_SRCSEL_ENABLE_SPDIF | EMU_A2_SRCSEL_ENABLE_SRCMULTI);

        EMU_PUT32(EMU_A2_PTR, EMU_A2_SRCMULTI);
        EMU_PUT32(EMU_A2_DATA, EMU_A2_SRCMULTI_ENABLE_INPUT);
    }

    if (softstate->audigy)
	EMU_PUT32(EMU_HCFG, EMU_HCFG_AUTOMUTE);
    else
	EMU_PUT32(EMU_HCFG, EMU_HCFG_AUTOMUTE | EMU_HCFG_LOCKTANKCACHE_MASK);

    /* Let's play with sound processor */
    if (softstate->audigy)
	audioemu_audigy_initfx(softstate);
    else
	audioemu_initfx(softstate);


    /*
     * Set all the PTB entries to the silent page.  The physical
     * addresses are ORed with the page number. I don't know what the
     * ORed index is for, might be a very useful unused feature...
     */
    silentpage = softstate->silentpage->paddr;
    ptb_acc = softstate->ptb->acc_handle;
    ptb = (uint32_t*)softstate->ptb->kaddr;
    for (i = 0; i < EMU_MAXPTE; i++)
	ddi_put32(ptb_acc, &ptb[i], EMU_MK_PTE(silentpage, i));

    /* Write PTB address and set TCB to none */
    audioemu_write(softstate, 0, EMU_PTB,  softstate->ptb->paddr);
    audioemu_write(softstate, 0, EMU_TCBS, EMU_TCBS_BUFFSIZE_16K);
    audioemu_write(softstate, 0, EMU_TCB, 0);	/* No TCB use for now */

    /*
     * Set channels MAPs to the silent page.
     * I don't know what MAPs are for.
     */
    for (i = 0; i < EMU_NUMCHAN; i++) {
	audioemu_write(softstate, i, EMU_CHAN_MAPA,
		       EMU_MK_CHMAP(silentpage, EMU_CHAN_MAP_PTI_MASK));
	audioemu_write(softstate, i, EMU_CHAN_MAPB, 
		       EMU_MK_CHMAP(silentpage, EMU_CHAN_MAP_PTI_MASK));
	softstate->channel[i] = NULL;
    }
    
#if	EMU_32BIT_PTE
    hcfg_ptb = EMU_HCFG_32BIT_PTB;
#else
    hcfg_ptb = 0;
#endif
    EMU_PUT32(EMU_HCFG,
	      EMU_GET32(EMU_HCFG) | hcfg_ptb | EMU_HCFG_AUDIOENABLE);
    EMU_PUT32(EMU_INTE,
	      EMU_GET32(EMU_INTE) |
	      EMU_INTE_VOLINCRENABLE | EMU_INTE_VOLDECRENABLE |
	      EMU_INTE_MUTEENABLE);

    if (softstate->audigy == 2) {
	EMU_PUT32(EMU_A_IOCFG, EMU_A_IOCFG_GPOUT0 | EMU_GET32(EMU_A_IOCFG));
    }

    return AUDIOEMU_SUCCESS;
}


/*
 * Cleanup all resources allocated by the driver during attach time
 * and during run-time, in reverse order.
 */
static int
audioemu_cleanup(struct audioemu_softstate *softstate)
{
    int i;

    if (softstate->audio_handle) {
	if (audio_hlp_detach(softstate->audio_handle, DDI_DETACH) != DDI_SUCCESS)
	    return DDI_FAILURE;
    }

    if (softstate->codec_handle) {
	CODEC_DELETE(softstate->codec_handle);
	softstate->codec_handle = NULL;    
    }

    if (softstate->regs_handle) {

	/* Disable any Channels interrupts */
	audioemu_write(softstate, 0, EMU_CLIEL, 0);
	audioemu_write(softstate, 0, EMU_CLIEH, 0);
	audioemu_write(softstate, 0, EMU_SOLEL, 0);
	audioemu_write(softstate, 0, EMU_SOLEH, 0);

	/*
	 * Should do some voice(stream) stopping stuff here, that's what will
	 * stop and deallocate all channels.
	 */

	/* Stop all channels */
	/* XXX This shouldn't be necessary, I'll remove once everything works */
	for (i = 0; i < EMU_NUMCHAN; i++)
	    audioemu_write(softstate, i, EMU_CHAN_DCYSUSV, 0);
	for (i = 0; i < EMU_NUMCHAN; i++) {
	    audioemu_write(softstate, i, EMU_CHAN_VTFT, 0);
	    audioemu_write(softstate, i, EMU_CHAN_CVCF, 0);
	    audioemu_write(softstate, i, EMU_CHAN_PTRX, 0);
	    audioemu_write(softstate, i, EMU_CHAN_CPF, 0);
	}

	/*
	 * Deallocate Emu10k1 caches and recording buffers. Again it will be
	 * removed because it will be done in voice shutdown.
	 */
	audioemu_write(softstate, 0, EMU_MICBS, EMU_RECBS_BUFSIZE_NONE);
	audioemu_write(softstate, 0, EMU_MICBA, 0);
	audioemu_write(softstate, 0, EMU_ADCBS, EMU_RECBS_BUFSIZE_NONE);
	audioemu_write(softstate, 0, EMU_ADCBA, 0);
	audioemu_write(softstate, 0, EMU_FXBS, EMU_RECBS_BUFSIZE_NONE);
	audioemu_write(softstate, 0, EMU_FXBA, 0);
	if (softstate->audigy) {
	    audioemu_write(softstate, 0, EMU_A_FXWC1, 0);
	    audioemu_write(softstate, 0, EMU_A_FXWC2, 0);
	} else
	    audioemu_write(softstate, 0, EMU_FXWC, 0);

	/*
	 * XXX I don't know yet how I will handle tank cache buffer,
	 * I don't even clearly  know what it is for.
	 */
	audioemu_write(softstate, 0, EMU_TCB, 0);
	audioemu_write(softstate, 0, EMU_TCBS, EMU_TCBS_BUFFSIZE_16K);
	
	if (softstate->audigy)		 	/* necessary ? */
	    audioemu_write(softstate, 0, EMU_A_DBG, EMU_A_DBG_SINGLE_STEP);
	else
	    audioemu_write(softstate, 0, EMU_DBG, EMU_DBG_SINGLE_STEP);
    }

    /*
     * remove interrupt handler before we free the card's mapped registers
     * otherwise an interrupt could access softstate->regs_handle in our
     * interrupt handler
     */
    if (softstate->intr_cookie != NULL) {
	ddi_remove_intr(softstate->dip, 0, softstate->intr_cookie);
    }

    dmamem_delete(softstate->silentpage);
    dmamem_delete(softstate->ptb);

    if (softstate->regs_handle != NULL) {
	ddi_regs_map_free(&softstate->regs_handle);
    }

    if (softstate->pci_conf_handle != NULL) {
	pci_config_teardown(&softstate->pci_conf_handle);
    }

    mutex_destroy(&softstate->voices_lock);
    mutex_destroy(&softstate->intr_mutex);

    if (softstate->lvoice)
	kmem_free(softstate->lvoice, sizeof(struct audioemu_voice));

    ddi_soft_state_free(audioemu_state, ddi_get_instance(softstate->dip));

    return DDI_SUCCESS;
}


/* Emu10k1 Memory managment */

static void
audioemu_mem_delete(struct audioemu_mem *mem)
{
    dmamem_delete(mem->dmamem);
    kmem_free(mem, sizeof(*mem));
}


static struct audioemu_mem *
audioemu_mem_new(struct audioemu_softstate *softstate,
		 int ptbidx, size_t size)
{
    struct audioemu_mem *mem;

    mem = kmem_alloc(sizeof(*mem), KM_SLEEP);

    mem->ptbidx = ptbidx;
    if ((mem->dmamem = dmamem_alloc(softstate,
				    ptbidx == EMU_RMEM
				        ? &rmem_dma_attr 
				        : &pmem_dma_attr,
				    &dev_sample_attr,
				    size)) == NULL) {
	kmem_free(mem, sizeof(*mem));
	return NULL;
    }

    if (!softstate->audigy && (mem->dmamem->paddr & 0x80000000)) {
	cmn_err(CE_WARN, "invalid physical address for playback page, failing");
	cmn_err(CE_NOTE, "SB Live! only works at physical addresses <2GB");
	audioemu_mem_delete(mem);
	return NULL;
    }

    return mem;
}


static void *
audioemu_pmem_alloc(struct audioemu_softstate *softstate, size_t size)
{
    int             i, j;
    size_t          numblocks;
    struct audioemu_mem *mem;
    ddi_acc_handle_t ptb_acc;
    uint32_t	   *ptb, silentpage;
    uint32_t	    paddr;
    size_t	    psize;

    ptb_acc = softstate->ptb->acc_handle;
    ptb = (uint32_t*)softstate->ptb->kaddr;
    silentpage = softstate->silentpage->paddr;
    numblocks = size / EMU_PTESIZE;
    if (size % EMU_PTESIZE)
	numblocks++;

    for (i = 0; i < EMU_MAXPTE; i++)
	if (EMU_PTE_ADR(ddi_get32(ptb_acc, &ptb[i])) == silentpage) {
	    /* We look for a free PTE */
	    mutex_enter(&softstate->intr_mutex);
	    for (j = 0; j < numblocks; j++)
		if (EMU_PTE_ADR(ddi_get32(ptb_acc, &ptb[i+j])) != silentpage)
		    break;
	    if (j == numblocks) {
		if ((mem = audioemu_mem_new(softstate, i, size)) == NULL) {
		    mutex_exit(&softstate->intr_mutex);
		    return NULL;
		}
		paddr = mem->dmamem->cookie.dmac_address;
		psize = mem->dmamem->cookie.dmac_size;
		for (j = 0; j < numblocks; j++) {
		    if (psize == 0) {
			ddi_dma_nextcookie(mem->dmamem->dma_handle, &mem->dmamem->cookie);
			paddr = mem->dmamem->cookie.dmac_address;
			psize = mem->dmamem->cookie.dmac_size;
		    }
		    if (psize < EMU_PTESIZE) {
			/*
			 * this should not happen, the dma allocation
			 * code returns page aligned memory.
			 */
			mutex_exit(&softstate->intr_mutex);
			cmn_err(CE_WARN, "incomplete page, %ld bytes",
			    (long)psize);
			return NULL;
		    }
		    ddi_put32(ptb_acc, &ptb[i + j], EMU_MK_PTE(paddr, i+j));
		    paddr += EMU_PTESIZE;
		    psize -= EMU_PTESIZE;
		}

		if (softstate->mem)
		    softstate->mem->prev = &mem->next;
		mem->next = softstate->mem;
		mem->prev = &softstate->mem;
		softstate->mem = mem;

		mutex_exit(&softstate->intr_mutex);
		return mem->dmamem->kaddr;
	    } else
		i += j;
	    mutex_exit(&softstate->intr_mutex);
	}
    return NULL;
}

static void *
audioemu_rmem_alloc(struct audioemu_softstate *softstate, size_t size)
{
    struct audioemu_mem *mem;

    if ((mem = audioemu_mem_new(softstate, EMU_RMEM, size)) == NULL)
	return NULL;

    mutex_enter(&softstate->intr_mutex);
    if (softstate->mem)
	softstate->mem->prev = &mem->next;
    mem->next = softstate->mem;
    mem->prev = &softstate->mem;
    softstate->mem = mem;
    mutex_exit(&softstate->intr_mutex);
    
    return mem->dmamem->kaddr;
}


static void
audioemu_freem(struct audioemu_softstate *softstate, void *ptr)
{
    int             i;
    struct audioemu_mem *mem;
    size_t          numblocks;
    uint32_t       *ptb, silentpage;
    ddi_acc_handle_t ptb_acc;

    ptb_acc = softstate->ptb->acc_handle;
    ptb = (uint32_t*)softstate->ptb->kaddr;
    silentpage = softstate->silentpage->paddr;
    for (mem = softstate->mem; mem != NULL; mem = mem->next) {
	if (mem->dmamem->kaddr != ptr)
	    continue;

	mutex_enter(&softstate->intr_mutex);
	if (mem->ptbidx != EMU_RMEM) {
	    numblocks = mem->dmamem->len / EMU_PTESIZE;
	    if (mem->dmamem->len % EMU_PTESIZE)
		numblocks++;
	    for (i = 0; i < numblocks; i++)
		ddi_put32(ptb_acc, &ptb[mem->ptbidx + i],
			  EMU_MK_PTE(silentpage, mem->ptbidx + i));
	}
	if ((*(mem->prev) = mem->next) != NULL)
	    mem->next->prev = mem->prev;

	mutex_exit(&softstate->intr_mutex);
	
	audioemu_mem_delete(mem);
	break;
    }
}


/*
 * audioemu_channel_* : Channel managment functions
 * audioemu_chanparms_* : Channel parameters modification functions
 */

static void
audioemu_chanparms_set_defaults(struct audioemu_channel *chan)
{
    chan->fxsend.a.level = chan->fxsend.b.level =
	chan->fxsend.c.level = chan->fxsend.d.level = 0xc0;	/* not max */
    chan->fxsend.a.dest = 0x0;
    chan->fxsend.b.dest = 0x1;
    chan->fxsend.c.dest = 0x2;
    chan->fxsend.d.dest = 0x3;

    chan->fxsend.e.level = chan->fxsend.f.level =
	chan->fxsend.g.level = chan->fxsend.h.level = 0xc0;	/* not max */
    chan->fxsend.e.dest = 0x4;
    chan->fxsend.f.dest = 0x5;
    chan->fxsend.g.dest = 0x6;
    chan->fxsend.h.dest = 0x7;

    chan->pitch.intial = 0x0000;	/* shouldn't it be 0xE000 ? */
    chan->pitch.current = 0x0000;	/* should it be 0x0400 */
    chan->pitch.target = 0x0000;	/* the unity pitch shift ? */
    chan->pitch.envelope_amount = 0x00;	/* none */

    chan->initial_attenuation = 0x00;	/* no attenuation */
    chan->volume.current = 0x0000;	/* no volume */
    chan->volume.target = 0xffff;
    chan->volume.envelope.current_state = 0x8000;/* 0 msec delay */
    chan->volume.envelope.hold_time = 0x7f;	/* 0 msec */
    chan->volume.envelope.attack_time = 0x7F;	/* 5.5msec */
    chan->volume.envelope.sustain_level = 0x7F;	/* full  */
    chan->volume.envelope.decay_time = 0x7F;	/* 22msec  */

    chan->filter.initial_cutoff_frequency = 0xff;	/* no filter */
    chan->filter.current_cutoff_frequency = 0xffff;	/* no filtering */
    chan->filter.target_cutoff_frequency = 0xffff;	/* no filtering */
    chan->filter.lowpass_resonance_height = 0x0;
    chan->filter.interpolation_ROM = 0x1;	/* full band */
    chan->filter.envelope_amount = 0x7f;	/* none */
    chan->filter.LFO_modulation_depth = 0x00;	/* none */

    chan->loop.start = 0x000000;
    chan->loop.end = 0x000010;	/* Why ? */

    chan->modulation.envelope.current_state = 0x8000;
    chan->modulation.envelope.hold_time = 0x00;	/* 127 better ? */
    chan->modulation.envelope.attack_time = 0x00;	/* infinite */
    chan->modulation.envelope.sustain_level = 0x00;	/* off */
    chan->modulation.envelope.decay_time = 0x7f;	/* 22 msec */
    chan->modulation.LFO_state = 0x8000;

    chan->vibrato_LFO.state = 0x8000;
    chan->vibrato_LFO.modulation_depth = 0x00;	/* none */
    chan->vibrato_LFO.vibrato_depth = 0x00;
    chan->vibrato_LFO.frequency = 0x00;	/* Why set to 24 when
					 * initialized ? */

    chan->tremolo_depth = 0x00;
}

/* only call it locked */
static struct audioemu_channel *
audioemu_channel_new(struct audioemu_voice *voice, uint8_t num)
{
    struct audioemu_channel *chan;

    chan = kmem_alloc(sizeof(struct audioemu_channel), KM_SLEEP);
    chan->voice = voice;
    chan->num = num;
    audioemu_chanparms_set_defaults(chan);
    chan->voice->softstate->channel[num] = chan;
    return chan;
}

/* only call it locked */
static void
audioemu_channel_delete(struct audioemu_channel *chan)
{
    chan->voice->softstate->channel[chan->num] = NULL;
    kmem_free(chan, sizeof(struct audioemu_channel));
}

static void
audioemu_channel_set_fxsend(struct audioemu_channel *chan,
			   struct audioemu_chanparms_fxsend *fxsend)
{
    bcopy(fxsend, &chan->fxsend, sizeof(struct audioemu_chanparms_fxsend));
}


#define PITCH_TARGET(x)	(((x) << 8) / 375 / 2)

static int
audioemu_channel_select_interprom(int pitch_target)
{
    if (pitch_target == PITCH_TARGET(48000))
	return EMU_CHAN_CCCA_INTERPROM_0;
    else if (pitch_target < PITCH_TARGET(48000))
	return EMU_CHAN_CCCA_INTERPROM_1;
    else if (pitch_target >= PITCH_TARGET(96000))
	return EMU_CHAN_CCCA_INTERPROM_0;
    else if (pitch_target >= PITCH_TARGET(85000))
	return EMU_CHAN_CCCA_INTERPROM_6;
    else if (pitch_target >= PITCH_TARGET(80726))
	return EMU_CHAN_CCCA_INTERPROM_5;
    else if (pitch_target >= PITCH_TARGET(67882))
	return EMU_CHAN_CCCA_INTERPROM_4;
    else if (pitch_target >= PITCH_TARGET(57081))
	return EMU_CHAN_CCCA_INTERPROM_3;
    else  
	return EMU_CHAN_CCCA_INTERPROM_2;
}

static void
audioemu_channel_set_srate(struct audioemu_channel *chan, uint32_t srate)
{
    int pitch_target;
    pitch_target = (srate << 8) / 375;
    pitch_target = (pitch_target >> 1) + (pitch_target & 1);
    pitch_target &= 0xffff;
    chan->pitch.target = pitch_target;
    chan->pitch.current = pitch_target;
    chan->pitch.intial = (audioemu_rate_to_pitch(srate) >> 8) & EMU_CHAN_IP_MASK;
    chan->filter.interpolation_ROM =
	audioemu_channel_select_interprom(pitch_target);
}

/* voice params must be set before calling this */
static void
audioemu_channel_set_bufparms(struct audioemu_channel *chan,
			      uint32_t start, uint32_t end)
{
    chan->loop.start = start & EMU_CHAN_PSST_LOOPSTARTADDR_MASK;
    chan->loop.end = end & EMU_CHAN_DSL_LOOPENDADDR_MASK;
}

static void
audioemu_channel_commit_parms(struct audioemu_channel *chan)
{
    struct audioemu_voice *voice = chan->voice;
    struct audioemu_softstate *softstate = voice->softstate;
    uint32_t start, mapval;
    uint8_t chano = chan->num;

    start = chan->loop.start +
	(voice->stereo ? 28 : 30) * (voice->b16 + 1);
    mapval = EMU_MK_CHMAP(softstate->silentpage->paddr, EMU_CHAN_MAP_PTI_MASK);

    mutex_enter(&softstate->intr_mutex);

    audioemu_write(softstate, chano, EMU_CHAN_CPF_STEREO, voice->stereo);
    if (softstate->audigy) {
	audioemu_write(softstate, chano, EMU_A_CHAN_FXRT1,
		       (chan->fxsend.d.dest << 24) |
		       (chan->fxsend.c.dest << 16) |
		       (chan->fxsend.b.dest <<  8) | 
		       (chan->fxsend.a.dest <<  0));
	audioemu_write(softstate, chano, EMU_A_CHAN_FXRT2,
		       (chan->fxsend.h.dest << 24) |
		       (chan->fxsend.g.dest << 16) |
		       (chan->fxsend.f.dest <<  8) | 
		       (chan->fxsend.e.dest <<  0));
	audioemu_write(softstate, chano, EMU_A_CHAN_SENDAMOUNTS,
		       (chan->fxsend.h.level <<  0) |
		       (chan->fxsend.g.level <<  8) |
		       (chan->fxsend.f.level << 16) | 
		       (chan->fxsend.e.level << 24));
    } else {
	audioemu_write(softstate, chano, EMU_CHAN_FXRT,
		       (chan->fxsend.d.dest << 28) |
		       (chan->fxsend.c.dest << 24) |
		       (chan->fxsend.b.dest << 20) | 
		       (chan->fxsend.a.dest << 16));
    }

    audioemu_write(softstate, chano, 0x10000000 | EMU_CHAN_PTRX,
		   (chan->fxsend.a.level << 8) | chan->fxsend.b.level);
    audioemu_write(softstate, chano, EMU_CHAN_DSL,
		   (chan->fxsend.d.level << 24) | chan->loop.end);
    audioemu_write(softstate, chano, EMU_CHAN_PSST,
		   (chan->fxsend.c.level << 24) | chan->loop.start);
    audioemu_write(softstate, chano, EMU_CHAN_CCCA,
		   (chan->filter.lowpass_resonance_height << 28) |
		   (chan->filter.interpolation_ROM << 25) |
		   (voice->b16 ? 0 : EMU_CHAN_CCCA_8BITSELECT) | start);
    audioemu_write(softstate, chano, EMU_CHAN_Z1, 0);
    audioemu_write(softstate, chano, EMU_CHAN_Z2, 0);
    audioemu_write(softstate, chano, EMU_CHAN_MAPA, mapval);
    audioemu_write(softstate, chano, EMU_CHAN_MAPB, mapval);
    audioemu_write(softstate, chano, EMU_CHAN_CVCF_CURRFILTER,
		   chan->filter.current_cutoff_frequency);
    audioemu_write(softstate, chano, EMU_CHAN_VTFT_FILTERTARGET,
		   chan->filter.target_cutoff_frequency);
    audioemu_write(softstate, chano, EMU_CHAN_ATKHLDM,
		   (chan->modulation.envelope.hold_time << 8) |
		   chan->modulation.envelope.attack_time);
    audioemu_write(softstate, chano, EMU_CHAN_DCYSUSM,
		   (chan->modulation.envelope.sustain_level << 8) |
		   chan->modulation.envelope.decay_time);
    audioemu_write(softstate, chano, EMU_CHAN_LFOVAL1,
		   chan->modulation.LFO_state);
    audioemu_write(softstate, chano, EMU_CHAN_LFOVAL2,
		   chan->vibrato_LFO.state);
    audioemu_write(softstate, chano, EMU_CHAN_FMMOD,
		   (chan->vibrato_LFO.modulation_depth << 8) |
		   chan->filter.LFO_modulation_depth);
    audioemu_write(softstate, chano, EMU_CHAN_TREMFRQ,
		   (chan->tremolo_depth << 8));
    audioemu_write(softstate, chano, EMU_CHAN_FM2FRQ2,
		   (chan->vibrato_LFO.vibrato_depth << 8) |
		   chan->vibrato_LFO.frequency);
    audioemu_write(softstate, chano, EMU_CHAN_ENVVAL,
		   chan->modulation.envelope.current_state);
    audioemu_write(softstate, chano, EMU_CHAN_ATKHLDV,
		   (chan->volume.envelope.hold_time << 8) |
		   chan->volume.envelope.attack_time);
    audioemu_write(softstate, chano, EMU_CHAN_ENVVOL,
		   chan->volume.envelope.current_state);
    audioemu_write(softstate, chano, EMU_CHAN_PEFE,
		   (chan->pitch.envelope_amount << 8) |
		   chan->filter.envelope_amount);
    mutex_exit(&softstate->intr_mutex);
}

static void
audioemu_channel_start(struct audioemu_channel *chan)
{
    struct audioemu_voice *voice = chan->voice;
    struct audioemu_softstate *softstate = voice->softstate;
    uint8_t	chano = chan->num;

    mutex_enter(&softstate->intr_mutex);
#if 1
    audioemu_write(softstate, chano, EMU_CHAN_CCR, 0);
#else
    {
    uint8_t     cache_sample, cache_invalid_size;
    uint32_t    sample;
#if 1
    cache_sample = voice->stereo ? 4 : 2;
    sample = voice->b16 ? 0x00000000 : 0x80808080;
    cache_invalid_size = (voice->stereo ? 28 : 30) * (voice->b16 + 1);

    while (cache_sample--) {
	audioemu_write(softstate, chano, EMU_CHAN_CD0 + cache_sample,
		       sample);
    }
    audioemu_write(softstate, chano, EMU_CHAN_CCR_CACHEINVALIDSIZE, 0);
    audioemu_write(softstate, chano, EMU_CHAN_CCR_READADDRESS, 64);
    audioemu_write(softstate, chano, EMU_CHAN_CCR_CACHEINVALIDSIZE,
		   cache_invalid_size);
#else
    cache_sample = voice->stereo ? 4 : 2;
    sample = voice->b16 ? 0x00000000 : 0x80808080;
    cache_invalid_size = (voice->stereo ? 12 : 14) * 4
	/ ((voice->stereo + 1) * (voice->b16 + 1));

    while (cache_sample--) {
	audioemu_write(softstate, chano, EMU_CHAN_CD0 + cache_sample,
		       sample);
    }
    audioemu_write(softstate, chano, EMU_CHAN_CCR, 
		   EMU_CHAN_CCR_LOOPFLAG |
		   (voice->dataloc.chan[0]->loop.start >> 16));
    audioemu_write(softstate, chano, EMU_CHAN_CLP,
		   voice->dataloc.chan[0]->loop.start & 0xffff);

    audioemu_write(softstate, chano, EMU_CHAN_CCR_READADDRESS, 0);
    audioemu_write(softstate, chano, EMU_CHAN_CCR_CACHEINVALIDSIZE,
		   cache_invalid_size);

    audioemu_write(softstate, chano, EMU_CHAN_CCR_CACHEINVALIDSIZE,
		   cache_invalid_size);
    
#endif
    }
#endif
    audioemu_write(softstate, chano, EMU_CHAN_IFATN,
		   (chan->filter.target_cutoff_frequency << 8) |
		   chan->initial_attenuation);
    audioemu_write(softstate, chano, EMU_CHAN_VTFT_VOLUMETARGET,
		   chan->volume.target);
    audioemu_write(softstate, chano, EMU_CHAN_CVCF_CURRVOL,
		   chan->volume.current);
    audioemu_write(softstate, 0,
		   EMU_MKSUBREG(1, chano, EMU_SOLEL + (chano >> 5)),
		   0);	/* Clear stop on loop */
    audioemu_write(softstate, 0,
		   EMU_MKSUBREG(1, chano, EMU_CLIEL + (chano >> 5)),
		   0);	/* Clear loop interrupt */
    audioemu_write(softstate, chano, EMU_CHAN_DCYSUSV,
		   (chan->volume.envelope.sustain_level << 8) |
		   chan->volume.envelope.decay_time);
    audioemu_write(softstate, chano, EMU_CHAN_PTRX_PITCHTARGET,
		   chan->pitch.target);
    if (chan == chan->voice->dataloc.chan[0])
	audioemu_write(softstate, chano, EMU_CHAN_CPF_PITCH,
		       chan->pitch.current);
    audioemu_write(softstate, chano, EMU_CHAN_IP, chan->pitch.intial);
    
    mutex_exit(&softstate->intr_mutex);
}

static void
audioemu_channel_stop(struct audioemu_channel *chan)
{
    uint8_t chano = chan->num;
    struct audioemu_softstate *softstate = chan->voice->softstate;

    mutex_enter(&softstate->intr_mutex);
    audioemu_write(softstate, chano, EMU_CHAN_PTRX_PITCHTARGET, 0);
    audioemu_write(softstate, chano, EMU_CHAN_CPF_PITCH, 0);
    audioemu_write(softstate, chano, EMU_CHAN_IFATN_ATTENUATION, 0xff);
    audioemu_write(softstate, chano, EMU_CHAN_VTFT_VOLUMETARGET, 0);
    audioemu_write(softstate, chano, EMU_CHAN_CVCF_CURRVOL, 0);
    audioemu_write(softstate, chano, EMU_CHAN_IP, 0);
    mutex_exit(&softstate->intr_mutex);
}

/*
 * Voices managment
 * audioemu_voice_dataloc : use(play or rec) independant dataloc union helpers
 * audioemu_voice_channel_* : play part of dataloc union helpers
 * audioemu_voice_recsrc_* : rec part of dataloc union helpers
 */

/* Allocate channels for voice in case of play voice */
static int
audioemu_voice_channel_create(struct audioemu_voice *voice)
{
    struct audioemu_softstate *softstate = voice->softstate;
    audioemu_channel_t **channel = softstate->channel;
    uint8_t i, stereo = voice->stereo;

    for (i = 0; i < EMU_NUMCHAN; i += stereo + 1) {
	if ((stereo && (channel[i + 1] != NULL)) ||
	    (channel[i] != NULL))	/* Looking for free channels */
	    continue;

	mutex_enter(&softstate->intr_mutex);
	if (stereo) {
	    voice->dataloc.chan[1] =
		audioemu_channel_new(voice, i + 1);
	    if (voice->dataloc.chan[1] == NULL) {
		mutex_exit(&softstate->intr_mutex);
		return ENOMEM;
	    }
	}
	voice->dataloc.chan[0] = audioemu_channel_new(voice, i);
	if (voice->dataloc.chan[0] == NULL) {
	    if (stereo) {
		audioemu_channel_delete(voice->dataloc.chan[1]);
		voice->dataloc.chan[1] = NULL;
	    }
	    mutex_exit(&softstate->intr_mutex);
	    return ENOMEM;
	}
	mutex_exit(&softstate->intr_mutex);
	return 0;
    }
    return EAGAIN;
}


/* When calling this function we assume no one can access the voice */
static void
audioemu_voice_channel_destroy(struct audioemu_voice *voice)
{
    audioemu_channel_delete(voice->dataloc.chan[0]);
    voice->dataloc.chan[0] = NULL;
    if (voice->stereo)
	audioemu_channel_delete(voice->dataloc.chan[1]);
    voice->dataloc.chan[1] = NULL;
}


static int
audioemu_voice_recsrc_reserve(struct audioemu_voice *voice, audioemu_recsrc_t source)
{
    if (source == EMU_RECSRC_NOTSET)
	return EINVAL;

    if (voice->softstate->recsrc[source] != NULL)
	return EBUSY;
    voice->softstate->recsrc[source] = voice;
    return 0;
}


/* When calling this function we assume the voice is stopped */
static void
audioemu_voice_recsrc_release(struct audioemu_voice *voice)
{
    struct audioemu_softstate *softstate = voice->softstate;
    audioemu_recsrc_t source = voice->dataloc.source;

    if (source != EMU_RECSRC_NOTSET)
    	softstate->recsrc[source] = NULL;
}


static int
audioemu_voice_dataloc_create(struct audioemu_voice *voice)
{
    int	error;
    audioemu_recsrc_t recsrc;

    if (voice->use & EMU_VOICE_USE_PLAY) {
	if ((error = audioemu_voice_channel_create(voice)))
	    return error;
    } else {
	recsrc = EMU_VOICE_USE_RECSRC(voice->use);
	if (recsrc == EMU_RECSRC_NOTSET)
	    return EINVAL;
	if ((error = audioemu_voice_recsrc_reserve(voice, recsrc)))
	    return error;
	voice->dataloc.source = recsrc;
    }
    return 0;
}

static void
audioemu_voice_dataloc_destroy(struct audioemu_voice *voice)
{
    if (voice->use & EMU_VOICE_USE_PLAY) {
	if (voice->dataloc.chan[0] != NULL)
	    audioemu_voice_channel_destroy(voice);
    } else {
	if (voice->dataloc.source != EMU_RECSRC_NOTSET) {
	    audioemu_voice_recsrc_release(voice);
	    voice->dataloc.source = EMU_RECSRC_NOTSET;
	}
    }
}


static struct audioemu_voice *
audioemu_voice_new(struct audioemu_softstate *softstate, uint8_t use)
{
    struct audioemu_voice *voice;

    mutex_enter(&softstate->voices_lock);
    voice = softstate->lvoice;
    softstate->lvoice = NULL;
    mutex_exit(&softstate->voices_lock);

    if (!voice)
	voice = kmem_zalloc(sizeof(*voice), KM_SLEEP);

    voice->softstate = softstate;
    voice->state = !EMU_VOICE_STATE_STARTED;
    voice->stereo = EMU_VOICE_STEREO_NOTSET;
    voice->b16 = 0;
    voice->sample_rate = 0;
    if (use & EMU_VOICE_USE_PLAY)
	voice->dataloc.chan[0] = voice->dataloc.chan[0] = NULL;
    else
	voice->dataloc.source = EMU_RECSRC_NOTSET;
    voice->buffer = NULL;
    voice->blksize = 0;
    voice->trigblk = 0;
    voice->blkmod = 0;
    voice->offs = 0;
    voice->used = 0;
    voice->inth = NULL;
    voice->inthparam = NULL;

    voice->use = use;

    mutex_enter(&softstate->voices_lock);
    /* add on voices list */
    if (softstate->voices)
	softstate->voices->prev = &voice->next;
    voice->next = softstate->voices;
    voice->prev = &softstate->voices;
    softstate->voices = voice;
    mutex_exit(&softstate->voices_lock);

    return (voice);
}


static void
audioemu_voice_delete(struct audioemu_voice *voice)
{
    struct audioemu_softstate *softstate = voice->softstate;
    struct audioemu_voice *lvoice;

    if (voice->state & EMU_VOICE_STATE_STARTED)
	audioemu_voice_halt(voice);

    audioemu_voice_dataloc_destroy(voice);

    mutex_enter(&softstate->voices_lock);
    /* remove from voices list */
    if ((*(voice->prev) = voice->next) != NULL)
	voice->next->prev = voice->prev;

    lvoice = softstate->lvoice;
    softstate->lvoice = voice;
    mutex_exit(&softstate->voices_lock);

    if (lvoice)
	kmem_free(lvoice, sizeof(struct audioemu_voice));
}

static int
audioemu_voice_set_stereo(struct audioemu_voice *voice, uint8_t stereo)
{
    int	error;
    struct audioemu_chanparms_fxsend fxsend;

    audioemu_voice_dataloc_destroy(voice);
    voice->stereo = stereo;
    if ((error = audioemu_voice_dataloc_create(voice)))
	return error;
    if (voice->use & EMU_VOICE_USE_PLAY) {
	if (voice->stereo) {
	    fxsend.a.dest = 0x0;
	    fxsend.b.dest = 0x1;
	    fxsend.c.dest = 0x2;
	    fxsend.d.dest = 0x3;
	    fxsend.e.dest = 0x4;
	    fxsend.f.dest = 0x5;
	    fxsend.g.dest = 0x6;
	    fxsend.h.dest = 0x7;
	    fxsend.a.level = fxsend.c.level = 0xc0;
	    fxsend.b.level = fxsend.d.level = 0x00;
	    fxsend.e.level = fxsend.f.level = 0x00;
	    fxsend.g.level = fxsend.h.level = 0x00;
	    audioemu_channel_set_fxsend(voice->dataloc.chan[0],	&fxsend);
	    fxsend.a.level = fxsend.c.level = 0x00;
	    fxsend.b.level = fxsend.d.level = 0xc0;
	    fxsend.e.level = fxsend.f.level = 0x00;
	    fxsend.g.level = fxsend.h.level = 0x00;
	    audioemu_channel_set_fxsend(voice->dataloc.chan[1], &fxsend);
	} /* No else : default is good for mono */	
    }
    return 0;
}

static uint16_t audioemu_valid_record_srate[] = { 
    48000, 44100, 32000, 24000,
    22050, 16000, 11025, 8000,
    0 
};
static uint16_t audioemu_audigy_valid_record_srate[] = {
    48000, 44100, 32000, 24000,
    22050, 16000, 12000, 11025,
    8000, 0
};

static int
audioemu_record_srate(uint16_t *srates, uint16_t sample_rate)
{
    int i;

    for (i = 0; srates[i]; i++)
	if (srates[i] == sample_rate)
	    return i;
    return -1;
}

static int
audioemu_voice_set_srate(struct audioemu_voice *voice, uint32_t srate)
{
    if (voice->use & EMU_VOICE_USE_PLAY) {
	if (srate < 4000 || srate > 48000)
	    return EINVAL;
	voice->sample_rate = srate;
	audioemu_channel_set_srate(voice->dataloc.chan[0], srate);
	if (voice->stereo)
	    audioemu_channel_set_srate(voice->dataloc.chan[1], srate);
    } else {
	if (audioemu_record_srate(voice->softstate->audigy
				  ? audioemu_audigy_valid_record_srate
				  : audioemu_valid_record_srate,
				  srate) < 0) {
	    return EINVAL;
	}
	voice->sample_rate = srate;
    }
    return 0;
}

static int
audioemu_voice_set_audioparms(struct audioemu_voice *voice, uint8_t stereo,
			      uint8_t b16, uint32_t srate)
{
    int error;
    int set_srate;	/* XXX: used to force a sample rate change,
			 * audioemu_voice_set_stereo() resets the
			 * channel to default params which messes up
			 * previous valid sample rate settings
			 */

    if (voice->stereo == stereo && voice->b16 == b16 &&
	voice->sample_rate == srate)
	return 0;

    DPRINTF3(AUDEMUM_DRIVER, 
	     "Setting %s voice params : %s, %u bits, %u hz\n",
	     (voice->use & EMU_VOICE_USE_PLAY) ? "play" : "record",
	     stereo ? "stereo" : "mono", (b16 + 1) * 8, srate);
	
    if ((set_srate = voice->stereo != stereo)) {
	if ((error = audioemu_voice_set_stereo(voice, stereo)))
	    return error;
    }
    voice->b16 = b16;
    if (voice->sample_rate != srate || set_srate) {
	if ((error = audioemu_voice_set_srate(voice, srate)))
	    return error;
    }
    return 0;
}


static int
audioemu_voice_record_start(struct audioemu_voice *voice)
{
    struct audioemu_softstate *softstate = voice->softstate;
    static int recbs_max[] = {
	0, 384, 448, 512, 640, 768, 896, 1024, 
	1280, 1536, 1792, 2048, 2560, 3072, 3584, 4096, 
	5120, 6144, 7168, 8192, 10240, 12288, 14366, 16384, 
	20480, 24576, 28672, 32768, 40960, 49152, 57344, 65536
    };
    int recbs;
    uint32_t recaddr;
    int inte = 0;
    int samplerate;
    int adc_cr;

    if (voice->use & EMU_VOICE_USE_PLAY)
	return EINVAL;
    if (!voice->buffer)
	return ENOMEM;

    recbs = 0;
    while (++recbs < 32 && recbs_max[recbs] <= voice->buffer->dmamem->len)
	;
    recbs--;
    recaddr = voice->buffer->dmamem->paddr;

    if (softstate->audigy) {
	adc_cr = EMU_A_ADCCR_LCHANENABLE;
	if (voice->stereo) 
	    adc_cr |= EMU_A_ADCCR_RCHANENABLE;
	samplerate = audioemu_record_srate(audioemu_audigy_valid_record_srate,
					   voice->sample_rate);
    } else {
	adc_cr = EMU_ADCCR_LCHANENABLE;
	if (voice->stereo) 
	    adc_cr |= EMU_ADCCR_RCHANENABLE;
	samplerate = audioemu_record_srate(audioemu_valid_record_srate,
					   voice->sample_rate);
    }
    if (samplerate < 0)
	return EINVAL;

    mutex_enter(&softstate->intr_mutex);

    switch (EMU_VOICE_USE_RECSRC(voice->use)) {
    case EMU_RECSRC_MIC:
	audioemu_write(softstate, 0, EMU_MICBA, recaddr);
	audioemu_write(softstate, 0, EMU_MICBS, recbs);
	inte = EMU_INTE_MICBUFENABLE; 
	break;
    case EMU_RECSRC_ADC:
	audioemu_write(softstate, 0, EMU_ADCBA, recaddr);
	audioemu_write(softstate, 0, EMU_ADCBS, recbs);
	inte = EMU_INTE_ADCBUFENABLE; 
	break;
    case EMU_RECSRC_FX:
	audioemu_write(softstate, 0, EMU_FXBA, recaddr);
	audioemu_write(softstate, 0, EMU_FXBS, recbs);
	if (softstate->audigy) {
	    audioemu_write(softstate, 0, EMU_A_FXWC1, 0);
	    audioemu_write(softstate, 0, EMU_A_FXWC2, 0);
	} else
	    audioemu_write(softstate, 0, EMU_FXWC, 0);
	inte = EMU_INTE_EFXBUFENABLE; 
	break;
    default:
	mutex_exit(&softstate->intr_mutex);
	DPRINTF2(AUDEMUM_DRIVER,
		 "cannot start recording: invalid source %d\n",
    		voice->dataloc.source);
	return EINVAL;
    }

    audioemu_write(softstate, 0, EMU_ADCCR, 0);
    audioemu_write(softstate, 0, EMU_ADCCR, adc_cr | samplerate);

    EMU_PUT32(EMU_INTE, EMU_GET32(EMU_INTE) | inte);

    mutex_exit(&softstate->intr_mutex);

    return 0;
}


static int
audioemu_voice_record_stop(struct audioemu_voice *voice)
{
    struct audioemu_softstate *softstate = voice->softstate;
    int inte = 0;

    if (voice->use & EMU_VOICE_USE_PLAY)
	return EINVAL;

    mutex_enter(&softstate->intr_mutex);

    switch (EMU_VOICE_USE_RECSRC(voice->use)) {
    case EMU_RECSRC_MIC:
	audioemu_write(softstate, 0, EMU_MICBS, EMU_RECBS_BUFSIZE_NONE);
	audioemu_write(softstate, 0, EMU_MICBA, 0);
	inte = EMU_INTE_MICBUFENABLE; 
	break;
    case EMU_RECSRC_ADC:
	audioemu_write(softstate, 0, EMU_ADCBS, EMU_RECBS_BUFSIZE_NONE);
	audioemu_write(softstate, 0, EMU_ADCBA, 0);
	inte = EMU_INTE_ADCBUFENABLE; 
	break;
    case EMU_RECSRC_FX:
	audioemu_write(softstate, 0, EMU_FXBS, EMU_RECBS_BUFSIZE_NONE);
	audioemu_write(softstate, 0, EMU_FXBA, 0);
	if (softstate->audigy) {
	    audioemu_write(softstate, 0, EMU_A_FXWC1, 0);
	    audioemu_write(softstate, 0, EMU_A_FXWC2, 0);
	} else
	    audioemu_write(softstate, 0, EMU_FXWC, 0);
	inte = EMU_INTE_EFXBUFENABLE;
	break;
    default:
	break;
    }
    audioemu_write(softstate, 0, EMU_ADCCR, 0);

    EMU_PUT32(EMU_INTE, EMU_GET32(EMU_INTE) & ~inte);

    mutex_exit(&softstate->intr_mutex);

    return 0;
}


/* voice audio parms (see just before) must be set prior to this */
static int
audioemu_voice_set_bufparms(struct audioemu_voice *voice, void *ptr,
			    uint32_t bufsize, uint16_t blksize)
{
    struct audioemu_mem *mem;
    struct audioemu_channel **chan;
    uint32_t start, end;
    uint8_t sample_size;
    int error = EFAULT;

    for (mem = voice->softstate->mem; mem != NULL; mem = mem->next) {
	if (mem->dmamem->kaddr != ptr)
	    continue;

	voice->buffer = mem;
	sample_size = (voice->b16 + 1) * (voice->stereo + 1);
	voice->blksize = blksize / sample_size;
	voice->trigblk = 0;	/* This shouldn't be needed */
	voice->blkmod = bufsize / blksize;
	if (bufsize % blksize) 	  /* This should not happen */
	    voice->blkmod++;
	error = 0;

	if (voice->use & EMU_VOICE_USE_PLAY) {
	    chan = voice->dataloc.chan;
	    start = (mem->ptbidx << 12) / sample_size;
	    end = start + bufsize / sample_size - 1;
	    audioemu_channel_set_bufparms(chan[0], start, end);
	    if (voice->stereo)
		audioemu_channel_set_bufparms(chan[1], start, end);
	    voice->timerate = (uint32_t) 48000 *
		voice->blksize / voice->sample_rate;
	    if (voice->timerate < 5)
		error = EINVAL;
	    else if (voice->timerate > 1024)
		voice->timerate = 1024;	
	}

	break;
    }

    return error;
}

static void
audioemu_voice_commit_parms(struct audioemu_voice *voice)
{
    if (voice->use & EMU_VOICE_USE_PLAY) {
	audioemu_channel_commit_parms(voice->dataloc.chan[0]);
	if (voice->stereo)
	    audioemu_channel_commit_parms(voice->dataloc.chan[1]);
	audioemu_channel_commit_parms(voice->dataloc.chan[0]);
	if (voice->stereo)
	    audioemu_channel_commit_parms(voice->dataloc.chan[1]);
    }
}

/* note: audioemu_voice_curaddr should be called with intr_mutex locked */
static uint32_t
audioemu_voice_curaddr(struct audioemu_voice *voice)
{
    if (voice->use & EMU_VOICE_USE_PLAY)
	return audioemu_read(voice->softstate,
			     voice->dataloc.chan[0]->num,
			     EMU_CHAN_CCCA_CURRADDR) -
	    voice->dataloc.chan[0]->loop.start;
    return 0;
}

static void
audioemu_resched_timer(struct audioemu_softstate *softstate)
{
    struct audioemu_voice *voice;
    uint16_t	timerate = 1024;
    uint8_t	active = 0;

    mutex_enter(&softstate->voices_lock);
    for (voice = softstate->voices; voice != NULL; voice = voice->next) {
	if ((voice->use & EMU_VOICE_USE_PLAY) == 0 ||
	    (voice->state & EMU_VOICE_STATE_STARTED) == 0)
	    continue;
	active = 1;
	if (voice->timerate < timerate)
	    timerate = voice->timerate;
    }
    mutex_exit(&softstate->voices_lock);

    if (timerate & ~EMU_TIMER_RATE_MASK)
	timerate = 0;
    EMU_PUT16(EMU_TIMER, timerate);
    if (!active && (softstate->timerstate & EMU_TIMER_STATE_ENABLED)) {
	EMU_PUT32(EMU_INTE, EMU_GET32(EMU_INTE) & ~EMU_INTE_INTERTIMERENB);
	softstate->timerstate &= ~EMU_TIMER_STATE_ENABLED;
    } else if (active && !(softstate->timerstate & EMU_TIMER_STATE_ENABLED)) {
	EMU_PUT32(EMU_INTE, EMU_GET32(EMU_INTE) | EMU_INTE_INTERTIMERENB);
	softstate->timerstate |= EMU_TIMER_STATE_ENABLED;
    }

    softstate->reschedule = 0;
}

static void
audioemu_voice_start(struct audioemu_voice *voice,
		     void (*inth) (void *), void *inthparam)
{
    voice->inth = inth;
    voice->inthparam = inthparam;
    if (voice->use & EMU_VOICE_USE_PLAY) {
	voice->trigblk = 1;
	audioemu_channel_start(voice->dataloc.chan[0]);
	if (voice->stereo)
	    audioemu_channel_start(voice->dataloc.chan[1]);
    } else {
	audioemu_voice_record_start(voice);
    }

    voice->state |= EMU_VOICE_STATE_STARTED;
    if (voice->use & EMU_VOICE_USE_PLAY)
	voice->softstate->reschedule |= EMU_TIMER_RESCHED_PENDING;
}

static void
audioemu_voice_halt(struct audioemu_voice *voice)
{
    if (voice->use & EMU_VOICE_USE_PLAY) {
	audioemu_channel_stop(voice->dataloc.chan[0]);
	if (voice->stereo)
	    audioemu_channel_stop(voice->dataloc.chan[1]);
    } else {
	audioemu_voice_record_stop(voice);
    }

    voice->state &= ~EMU_VOICE_STATE_STARTED;
    if (voice->use & EMU_VOICE_USE_PLAY)
	voice->softstate->reschedule |= EMU_TIMER_RESCHED_PENDING;
}


/*
 * Convert a number of bytes into a number of samples, according to
 * a given audio format 'prinfo'.
 */
static int
audioemu_bytes_to_samples(int byte_count, audio_prinfo_t *prinfo)
{
    int pcm16, stereo, sample_count;

    pcm16 = (prinfo->precision == AUDIO_PRECISION_16 ||
	     (prinfo->encoding != AUDIO_ENCODING_LINEAR &&
	      prinfo->encoding != AUDIO_ENCODING_LINEAR8));
    stereo = (prinfo->channels == AUDIO_CHANNELS_STEREO);

    sample_count = byte_count;
    if (pcm16)  sample_count >>= 1;
    if (stereo) sample_count >>= 1;
    return sample_count;
}


/*
 * Convert a number of bytes into a number of samples, according to
 * a given audio format 'prinfo'.
 */
static int
audioemu_samples_to_bytes(int sample_count, audio_prinfo_t *prinfo)
{
    int pcm16, stereo, byte_count;

    pcm16 = (prinfo->precision == AUDIO_PRECISION_16 ||
	     (prinfo->encoding != AUDIO_ENCODING_LINEAR &&
	      prinfo->encoding != AUDIO_ENCODING_LINEAR8));
    stereo = (prinfo->channels == AUDIO_CHANNELS_STEREO);

    byte_count = sample_count;
    if (pcm16)  byte_count <<= 1;
    if (stereo) byte_count <<= 1;
    return byte_count;
}


static void
audioemu_play_refill(void *inthparam)
{
    struct audioemu_voice *voice = inthparam;
    struct dmamem *dmamem = voice->buffer->dmamem;
    struct audioemu_softstate *softstate = voice->softstate;
    int bytes, samples, samples_read;
    int curaddr;
    caddr_t buf;
    int new_data;

    mutex_enter(&softstate->intr_mutex);
    curaddr = audioemu_samples_to_bytes(audioemu_voice_curaddr(voice),
					&softstate->audio_info.play);
    DPRINTF4(AUDEMUM_INTR, "audioemu_play_refill data %d/%d, trig %d, curr %d\n",
		voice->offs, voice->used, voice->trigblk, curaddr);

    /* avoid changing the active playback page */
    curaddr &= ~(EMU_PTESIZE-1);

    if (voice->offs + voice->used > dmamem->len) {
	/* valid sample data wraps around end of sample buffer */
	if (curaddr >= voice->offs + voice->used - dmamem->len
	    && curaddr < voice->offs) {
	    /*
	    printf("play_refill dry: data %d/%d, curr %d\n",
		voice->offs, voice->used, curaddr);
	    */
	    voice->used = 0;
	} else if (curaddr >= voice->offs)
	    voice->used -= curaddr - voice->offs;
	else
	    voice->used -= curaddr + dmamem->len - voice->offs;
	voice->offs = curaddr;
    } else {
	/* valid sample data does not cross end of sample buffer */
	if (curaddr < voice->offs || curaddr >= voice->offs + voice->used) {
	    /*
	    printf("play_refill dry: data %d/%d, curr %d\n",
		voice->offs, voice->used, curaddr);
	    */
	    voice->used = 0;
	} else
	    voice->used -= curaddr - voice->offs;
	voice->offs = curaddr;
    }

    /*
     * now try to fill in more data into the ring buffer
     */
    new_data = 0;
    while (voice->used < dmamem->len) {
	if (voice->offs + voice->used >= dmamem->len) {
	    bytes = dmamem->len - voice->used;
	} else {
	    bytes = dmamem->len - (voice->offs + voice->used);
	}
	samples = audioemu_bytes_to_samples(bytes, &softstate->audio_info.play);
	DPRINTF4(AUDEMUM_INTR, "space for %d samples / %d bytes\n", samples, bytes);

	buf = dmamem->kaddr + (voice->offs+voice->used) % dmamem->len;
	samples_read = audio_hlp_get_audio(softstate->audio_handle,
					   buf, 0, samples);
	if (samples_read <= 0) {
	    DPRINTF4(AUDEMUM_INTR, "got no samples\n");
	    break;
	}
	bytes = audioemu_samples_to_bytes(samples_read, 
					  &softstate->audio_info.play);
#if sparc
	if (softstate->audio_info.play.precision == AUDIO_PRECISION_16)
	    swab(buf, buf, bytes);
#endif

	new_data = 1;
	voice->used += bytes;
	
	DPRINTF4(AUDEMUM_INTR, "got %d samples / %d bytes\n", samples_read, bytes);
    }
    /*
    printf("play_refill: data %d/%d, curr %d\n",
	   voice->offs, voice->used, curaddr);
    */
    mutex_exit(&softstate->intr_mutex);

    if (new_data)
	ddi_dma_sync(dmamem->dma_handle, 0, 0, DDI_DMA_SYNC_FORDEV);

    if (voice->used <= 0) {
	/*
	 * running out of sample data, stop voice and tell audio
	 * helper about end of playback.
	 */
	audioemu_voice_halt(voice);
	audio_hlp_play_shutdown(softstate->audio_handle, 0);
    }
}


static void
audioemu_record_send(void *inthparam)
{
    struct audioemu_voice *voice = inthparam;
    struct dmamem *dmamem = voice->buffer->dmamem;
    struct audioemu_softstate *softstate = voice->softstate;
    int bufsize_half;
    caddr_t buf;
    int samples_half;

    DPRINTF4(AUDEMUM_INTR, "audioemu_record_send data, trig %d\n",
	     voice->trigblk);
    
    ddi_dma_sync(dmamem->dma_handle, 0, 0, DDI_DMA_SYNC_FORCPU);

    bufsize_half = dmamem->len / 2;
    buf = (char*)dmamem->kaddr + voice->trigblk * bufsize_half;
#if sparc
    if (softstate->audio_info.record.precision == AUDIO_PRECISION_16)
	swab(buf, buf, bufsize_half);
#endif
    samples_half = audioemu_bytes_to_samples(bufsize_half,
					     &softstate->audio_info.record);

    audio_hlp_send_audio(softstate->audio_handle, buf, 0, samples_half);
}


/*
 * interrupt handler for the audioemu device.
 */
static uint_t
audioemu_intr(caddr_t intr_arg)
{
    struct audioemu_softstate *softstate = (void*)intr_arg;
    struct audioemu_voice *voice;
    audioemu_recsrc_t recsrc;
    uint32_t ipr, ipr_mask;

    ipr = EMU_GET32(EMU_IPR);

    DPRINTF4(AUDEMUM_INTR, "audioemu_intr %x\n", ipr);

#if !sparc
    /*
     * sparc gets upset when we receive many interrupts but never claim them
     * XXX: Why are we receiving many interrupts with ipr == 0, anyway?
     */
    if (!ipr)
	return DDI_INTR_UNCLAIMED;
#endif

    if (ipr & EMU_IPR_INTERVALTIMER) {
	mutex_enter(&softstate->voices_lock);
	for (voice = softstate->voices; voice != NULL; voice = voice->next) {
	    if ((voice->use & EMU_VOICE_USE_PLAY) == 0 ||
		(voice->state & EMU_VOICE_STATE_STARTED) == 0)
		continue;

#if 1
	    mutex_exit(&softstate->voices_lock);
	    voice->inth(voice->inthparam);
	    mutex_enter(&softstate->voices_lock);
#else
	    {
		uint32_t curaddr;
		uint32_t curblk;

		mutex_enter(&softstate->intr_mutex);
		curaddr = audioemu_voice_curaddr(voice);
		mutex_exit(&softstate->intr_mutex);
		DPRINTF4(AUDEMUM_INTR,
			 "audioemu_intr timer, curaddr %x, blk %d, trig %d\n", 
			 curaddr, curaddr / voice->blksize, voice->trigblk);
		curblk = curaddr / voice->blksize;
		if (curblk == voice->trigblk) {
		    mutex_exit(&softstate->voices_lock);
		    voice->inth(voice->inthparam);
		    mutex_enter(&softstate->voices_lock);
		    voice->trigblk++;
		    voice->trigblk %= voice->blkmod;
		}
	    }
#endif
	}
	mutex_exit(&softstate->voices_lock);

	if (softstate->reschedule)
	    audioemu_resched_timer(softstate);
    }

    if (ipr & (EMU_IPR_MICBUFFULL | EMU_IPR_MICBUFHALFFULL |
	       EMU_IPR_ADCBUFFULL | EMU_IPR_ADCBUFHALFFULL |
	       EMU_IPR_EFXBUFFULL | EMU_IPR_EFXBUFHALFFULL)) {
	mutex_enter(&softstate->voices_lock);
	for (voice = softstate->voices; voice != NULL; voice = voice->next) {
	    if (voice->use & EMU_VOICE_USE_PLAY)
		continue;
	    switch (recsrc = EMU_VOICE_USE_RECSRC(voice->use)) {
	    case  EMU_RECSRC_MIC:
		ipr_mask = EMU_IPR_MICBUFFULL | EMU_IPR_MICBUFHALFFULL;
		break;
	    case  EMU_RECSRC_ADC:
		ipr_mask = EMU_IPR_ADCBUFFULL | EMU_IPR_ADCBUFHALFFULL;
		break;
	    case  EMU_RECSRC_FX:
		ipr_mask = EMU_IPR_EFXBUFFULL | EMU_IPR_EFXBUFHALFFULL;
		break;
	    default:
		ipr_mask = 0;
	    }
	    if (!(ipr & ipr_mask))
		continue;

	    voice->trigblk = (ipr & ipr_mask 
			      & (EMU_IPR_MICBUFFULL |
				 EMU_IPR_ADCBUFFULL |
				 EMU_IPR_EFXBUFFULL)) != 0;

	    mutex_exit(&softstate->voices_lock);
	    voice->inth(voice->inthparam);
	    mutex_enter(&softstate->voices_lock);
	}
	mutex_exit(&softstate->voices_lock);
    }

    /* Got interrupt */
    EMU_PUT32(EMU_IPR, ipr);

    return DDI_INTR_CLAIMED;
}


static audio_device_t audioemu_dev_info;
static ah_ad_entry_t audioemu_entry;
static ah_ad_ch_cap_t audioemu_ch_cap;
static ah_ad_cap_comb_t audioemu_cap_comb[];

static int
audioemu_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audioemu_softstate *softstate;
    int status;
    char *getdev_name;

    DPRINTF4(AUDEMUM_ATTACH, "attach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_ATTACH:
	break;

    case DDI_RESUME:
    case DDI_PM_RESUME:
    default:
	return DDI_FAILURE;
    }
	
    /*
     * this card is a bus master, reject any slave-only slot
     */
    if (ddi_slaveonly(dip) == DDI_SUCCESS) {
        cmn_err(CE_NOTE, "slot does not support PCI bus-master");
        return DDI_FAILURE;
    }

    /*
     * Make sure we're not using a high level interrupt handler.
     */
    if (ddi_intr_hilevel(dip, 0)) {
	cmn_err(CE_NOTE, "Driver does not support high level interrupt.");
	cmn_err(CE_NOTE, "You should probably add 'interrupt-priorities=9;' to audioemu.conf");
	return DDI_FAILURE;
    }

    if (ddi_soft_state_zalloc(audioemu_state, instance) != DDI_SUCCESS)
	return DDI_FAILURE;

    softstate = ddi_get_soft_state(audioemu_state, instance);
    softstate->dip = dip;

    if (pci_config_setup(softstate->dip, &softstate->pci_conf_handle) != DDI_SUCCESS) {
	DPRINTF4(AUDEMUM_ATTACH, "audioemu%d: pci_config_setup failed", instance);
	audioemu_cleanup(softstate);
	return DDI_FAILURE;
    }

    if (audioemu_pci_setup(softstate) != AUDIOEMU_SUCCESS) {
	DPRINTF4(AUDEMUM_ATTACH, "audioemu%d: audioemu_pci_setup failed", instance);
	audioemu_cleanup(softstate);
	return DDI_FAILURE;	
    }

    bcopy(&audioemu_defaults, &softstate->audio_info, sizeof(audio_info_t));
    bcopy(&audioemu_dev_info, &softstate->dev_info, sizeof(audio_device_t));

    /* allow spoofing of audio_device.name in AUDIO_GETDEV ioctl */
    if (ddi_prop_lookup_string(DDI_DEV_T_ANY, dip,
			       DDI_PROP_DONTPASS | DDI_PROP_NOTPROM,
			       "ioctl_audio_getdev_name",
			       &getdev_name) == DDI_PROP_SUCCESS) {

	strncpy(softstate->dev_info.name, getdev_name,
		sizeof(softstate->dev_info.name));
	ddi_prop_free(getdev_name);
    }

    sprintf(softstate->dev_info.version, "%02X,%s",
	    softstate->rev, softstate->revstr);

    softstate->ad_info.ad_int_vers = AD_VERS1;
    softstate->ad_info.ad_defaults = &audioemu_defaults;
    bcopy(&audioemu_ch_cap, &softstate->ad_info.ad_play,
	  sizeof(ah_ad_ch_cap_t));
    bcopy(&audioemu_ch_cap, &softstate->ad_info.ad_record,
	  sizeof(ah_ad_ch_cap_t));
    softstate->ad_info.ad_play_comb = audioemu_cap_comb;
    softstate->ad_info.ad_rec_comb = audioemu_cap_comb;
    softstate->ad_info.ad_entry    = &audioemu_entry;
    softstate->ad_info.ad_dev_info = &softstate->dev_info;
    softstate->ad_info.ad_translate_flags =
				AD_MISC_8_P_TRANSLATE | AD_MISC_8_R_TRANSLATE;

    softstate->audio_handle = audio_hlp_attach(dip, cmd, &softstate->ad_info);
    if (!softstate->audio_handle) {
	cmn_err(CE_WARN, "failed to attach audio helper module");
	audioemu_cleanup(softstate);
	return DDI_FAILURE;
    }

    audio_hlp_set_private(softstate->audio_handle, softstate);


    /* map the audigy's register block */
    status = ddi_regs_map_setup(dip, 1,
				(caddr_t *)&softstate->regs, 0,
				softstate->audigy == 2
				    ? EMU_A2_REGS_SIZE : EMU_REGS_SIZE,
				&dev_attr, &softstate->regs_handle);

    if (status != DDI_SUCCESS) {
	if (status == DDI_REGS_ACC_CONFLICT) {
	    cmn_err(CE_WARN, "regs mapping conflicted with other mapping");
	} else {
	    cmn_err(CE_WARN, "regs mapping failed");
	}
	audioemu_cleanup(softstate);
	return DDI_FAILURE;
    }


    if (ddi_get_iblock_cookie(dip, 0, &softstate->intr_cookie) != DDI_SUCCESS) {
	cmn_err(CE_WARN, "Could not get iblock cookie");
	audioemu_cleanup(softstate);
	return DDI_FAILURE;
    }
    mutex_init(&softstate->voices_lock, NULL,
	       MUTEX_DRIVER, (void*)softstate->intr_cookie);
    mutex_init(&softstate->intr_mutex, NULL,
	       MUTEX_DRIVER, (void*)softstate->intr_cookie);

    if (ddi_add_intr(dip, 0,
		     &softstate->intr_cookie, NULL,
		     audioemu_intr, (caddr_t)softstate) != DDI_SUCCESS) {

	cmn_err(CE_WARN, "Could not add interrupt handler");
	audioemu_cleanup(softstate);
	return DDI_FAILURE;
    }

    if (audioemu_hwinit(softstate) != AUDIOEMU_SUCCESS) {
	cmn_err(CE_WARN, "hardware initialization of emu audio controller failed");
	audioemu_cleanup(softstate);
	return DDI_FAILURE;	
    }

    /*
     * try to reset the audio controller, this verifies basic operation of the
     * sound hardware.
     */

    {
	static struct ac97_codec_ops codec_ops = {
	    audioemu_ac97_read,
	    audioemu_ac97_write
	};

	softstate->codec_handle =
	    ac97_codec_init(softstate->audio_handle, &codec_ops, softstate);
    }

    if (audioemu_reset_ac97(softstate, &audioemu_defaults, TRUE) != AUDIOEMU_SUCCESS) {
	cmn_err(CE_WARN, "reset of emu ac97 codec failed");
	audioemu_cleanup(softstate);
	return DDI_FAILURE;	
    }

    return DDI_SUCCESS;
}


static int
audioemu_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audioemu_softstate *softstate = ddi_get_soft_state(audioemu_state, instance);

    DPRINTF4(AUDEMUM_ATTACH, "detach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_DETACH:
	break;

    default:
	return DDI_FAILURE;
    }

    /* All voices should be stopped now but add some code here if not */

    EMU_PUT32(EMU_HCFG,
	      EMU_HCFG_LOCKSOUNDCACHE | EMU_HCFG_LOCKTANKCACHE_MASK |
	      EMU_HCFG_MUTEBUTTONENABLE);

    /* Disable all interrupts */
    EMU_PUT32(EMU_INTE, 0);

    if (audioemu_cleanup(softstate) != DDI_SUCCESS) 
	return DDI_FAILURE;

    return DDI_SUCCESS;
}


/*
 * Audio Driver functions (for ah_ad_entry_t)
 */

static int    
audioemu_setup(audiohdl_t handle, int stream, int flags)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDEMUM_DRIVER, "audioemu_setup, stream %d, flags %x\n", stream, flags);

    softstate->pvoice = audioemu_voice_new(softstate, EMU_VOICE_USE_PLAY);
    if (softstate->pvoice == NULL)
	return AUDIO_FAILURE;
    
    softstate->pbuf = audioemu_pmem_alloc(softstate, SAMPLEBUF_SIZE);
    if (softstate->pbuf == NULL)
	return AUDIO_FAILURE;

    softstate->rvoice = audioemu_voice_new(softstate, EMU_VOICE_USE_REC_ADC);
    if (softstate->rvoice == NULL)
	return AUDIO_FAILURE;

    softstate->rbuf = audioemu_rmem_alloc(softstate,
					  min(SAMPLEBUF_SIZE, 64*1024));
    if (softstate->rbuf == NULL)
	return AUDIO_FAILURE;

    return AUDIO_SUCCESS;
}


static void
audioemu_teardown(audiohdl_t handle, int stream, int dir)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDEMUM_DRIVER, "audioemu_teardown, stream %d, dir %d\n", stream, dir);

    if (softstate->pbuf) {
	audioemu_freem(softstate, softstate->pbuf);
	softstate->pbuf = NULL;
    }
    if (softstate->rbuf) {
	audioemu_freem(softstate, softstate->rbuf);
	softstate->rbuf = NULL;
    }

    if (softstate->pvoice) {
	audioemu_voice_delete(softstate->pvoice);
	softstate->pvoice = NULL;
    }
    if (softstate->rvoice) {
	audioemu_voice_delete(softstate->rvoice);
	softstate->rvoice = NULL;
    }
}


static int
audioemu_set_config(audiohdl_t handle, int stream,
		     int command, int dir, int arg1, int arg2)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);
    
    DPRINTF4(AUDEMUM_DRIVER, "audioemu_set_config, stream %d, "
	    "command %d, dir %d, arg1 %d, arg2 %d\n",
	    stream, command, dir, arg1, arg2);

    CODEC_SET_CONFIG(softstate->codec_handle, command, dir, arg1, arg2);

    return AUDIO_SUCCESS;
}


static int
audioemu_set_format(audiohdl_t handle, int stream,
		 int dir, int sample_rate, int channels,
		 int precision, int encoding)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);
    struct audioemu_voice *voice;
    audio_prinfo_t *prinfo;
    int err;

    DPRINTF4(AUDEMUM_DRIVER, "audioemu_set_format, stream %d, dir %d, "
	    "sample_rate %d, channels %d, precision %d, encoding %d\n",
	    stream, dir, sample_rate, channels, precision, encoding);

    if (dir == AUDIO_PLAY) {
	prinfo = &softstate->audio_info.play;
	voice = softstate->pvoice;
    } else {
	prinfo = &softstate->audio_info.record;
	voice = softstate->rvoice;
    }

    if (!voice) {
	/* XXX: this should not happen, but it does... */
        DPRINTF2(AUDEMUM_DRIVER,
		 "audioemu_set_format: no voice %p/%p, dir %d\n",
		 softstate->pvoice, softstate->rvoice, dir);
	return AUDIO_FAILURE;
    }

    prinfo->sample_rate = sample_rate;
    prinfo->channels = channels;
    prinfo->precision = precision;
    prinfo->encoding = encoding;

    err = audioemu_voice_set_audioparms(voice,
					channels == 2,
					precision == 16,
					sample_rate);
    if (err)
	return AUDIO_FAILURE;

    if (dir == AUDIO_PLAY) {
	err = audioemu_voice_set_bufparms(voice, softstate->pbuf,
					  SAMPLEBUF_SIZE,
					  SAMPLEBUF_INTR_CHUNK);
    } else {
	err = audioemu_voice_set_bufparms(voice, softstate->rbuf,
					  min(SAMPLEBUF_SIZE, 64*1024),
					  SAMPLEBUF_INTR_CHUNK);
    }
    if (err)
	return AUDIO_FAILURE;

    audioemu_voice_commit_parms(voice);

    return AUDIO_SUCCESS;
}


/*
 * Audio framework has seen the first audio samples for playback,
 * or playback resumes after a 'pause_play'. 
 */
static int
audioemu_start_play(audiohdl_t handle, int stream)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDEMUM_DRIVER, "audioemu_start_play, stream %d\n", stream);

    if (!softstate->pvoice) {		/* this should not happen */
        DPRINTF2(AUDEMUM_DRIVER, "audioemu_start_play: no voice\n");
	return AUDIO_FAILURE;
    }

    softstate->pvoice->offs = 0;
    softstate->pvoice->used = 0;

    audioemu_play_refill(softstate->pvoice);

    audioemu_voice_start(softstate->pvoice,
			 audioemu_play_refill, softstate->pvoice);
    if (softstate->reschedule)
	audioemu_resched_timer(softstate);

    return AUDIO_SUCCESS;
}


static void
audioemu_pause_play(audiohdl_t handle, int stream)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDEMUM_DRIVER, "audioemu_pause_play, stream %d\n", stream);

    if (!softstate->pvoice) {		/* this should not happen */
        DPRINTF2(AUDEMUM_DRIVER, "audioemu_pause_play: no voice\n");
	return;
    }

    audioemu_voice_halt(softstate->pvoice);
    if (softstate->reschedule)
	audioemu_resched_timer(softstate);
}


static void
audioemu_stop_play(audiohdl_t handle, int stream)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);

    if (!softstate->pvoice) {		/* this should not happen */
        DPRINTF2(AUDEMUM_DRIVER, "audioemu_stop_play: no voice\n");
	return;
    }

    audioemu_voice_halt(softstate->pvoice);
    if (softstate->reschedule)
	audioemu_resched_timer(softstate);
}


/*
 * audioemu_get_delay: returns the number of samples buffered in the
 * sound card and that are still waiting for playback.
 */
static int
audioemu_get_delay(audiohdl_t handle, int stream)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);
    struct audioemu_voice *voice;
    int offs_sample;
    int used_sample;
    int curr_sample;
    int samples;

    if ((voice = softstate->pvoice) == NULL) {	/* this should not happen */
        DPRINTF2(AUDEMUM_DRIVER, "audioemu_get_delay: no voice\n");
	return 0;
    }

    mutex_enter(&softstate->intr_mutex);
    if (voice->state & EMU_VOICE_STATE_STARTED) {
	offs_sample = audioemu_bytes_to_samples(voice->offs,
						&softstate->audio_info.play);
	used_sample = audioemu_bytes_to_samples(voice->used,
						&softstate->audio_info.play);
	curr_sample = audioemu_voice_curaddr(voice);
    } else {
	offs_sample = used_sample = curr_sample = 0;
    }
    mutex_exit(&softstate->intr_mutex);

    if (curr_sample < offs_sample) {
	/* handle wrap arround at buffer end */
	curr_sample += audioemu_bytes_to_samples(voice->buffer->dmamem->len,
						 &softstate->audio_info.play);
    }

    if (curr_sample >= offs_sample && curr_sample < offs_sample + used_sample) {
	/* curr_sample points inside the region with valid samples */
	samples = offs_sample + used_sample - curr_sample;
    } else {
	samples = 0;
    }

    return samples;
}


static int
audioemu_start_record(audiohdl_t handle, int stream)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDEMUM_DRIVER, "audioemu_start_record, stream %d\n", stream);

    if (!softstate->rvoice) {	/* this should not happen */
        DPRINTF2(AUDEMUM_DRIVER, "audioemu_start_record: no voice\n");
	return AUDIO_FAILURE;
    }

    audioemu_voice_start(softstate->rvoice,
			 audioemu_record_send, softstate->rvoice);
    if (softstate->reschedule)
	audioemu_resched_timer(softstate);

    return AUDIO_SUCCESS;
}


static void
audioemu_stop_record(audiohdl_t handle, int stream)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);

    if (!softstate->rvoice) {	/* this should not happen */
        DPRINTF2(AUDEMUM_DRIVER, "audioemu_stop_record: no voice\n");
	return;
    }

    audioemu_voice_halt(softstate->rvoice);
    if (softstate->reschedule)
	audioemu_resched_timer(softstate);
}


static int
audioemu_ioctl(audiohdl_t handle, int channel,
		queue_t *q, mblk_t *mp, int *error)
{
    struct audioemu_softstate *softstate = audio_hlp_get_private(handle);
    struct iocblk *iocp = (struct iocblk *)mp->b_rptr;
    mblk_t *mp2;
    caddr_t arg = 0;
    int vol;
    int gpr_bit;
    uint16_t codecval;

    if (iocp->ioc_count == TRANSPARENT && (mp2 = mp->b_cont) != NULL)
	/* get the "arg" value from the streams msg */
	bcopy(mp2->b_rptr, &arg, sizeof(arg));

    DPRINTF4(AUDEMUM_DRIVER, "audioemu_ioctl ioc_cmd=%x, arg=%p\n", iocp->ioc_cmd, arg);

    switch (iocp->ioc_cmd) {
    case AUDIOEMU_SET_PCMOUT_VOL:
	vol = ac97_codec_play_vol((int) (long)arg, AUDIO_MAX_GAIN, PCMOVR_RIGHT_GAIN_MASK);
	codecval = (vol << 8) | vol;
	audioemu_ac97_write(softstate, AC97_PCM_OUT_VOLUME_REGISTER, codecval);
	DPRINTF4(AUDEMUM_AC97, "PCMVOL=%x", codecval);
	return AD_ACK;

    case AUDIOEMU_SET_BEEP_VOL:
	vol = ac97_codec_play_vol((int) (long)arg, AUDIO_MAX_GAIN, 15);
	codecval = vol << 1;
	audioemu_ac97_write(softstate, AC97_PC_BEEP_REGISTER, codecval);
	DPRINTF4(AUDEMUM_AC97, "PCBEEPVOL=%x", codecval);
	return AD_ACK;

    case AUDIOEMU_SET_LOUDNESS_ENABLE:
	gpr_bit = GPR_BASS_BOOST;
	goto set_gpr_register;

    case AUDIOEMU_SET_3D_STEREO_ENABLE:
	gpr_bit = GPR_3D_STEREO_ENHANCE;
	goto set_gpr_register;

    case AUDIOEMU_SET_STEREO_ENHANCE_ENABLE:
	gpr_bit = GPR_ST;

    set_gpr_register:	
	audioemu_ac97_read(softstate, AC97_GENERAL_PURPOSE_REGISTER, &codecval);
	if ((long)arg)
	    codecval |= gpr_bit;
	else
	    codecval &= ~gpr_bit;
	audioemu_ac97_write(softstate, AC97_GENERAL_PURPOSE_REGISTER, codecval);
	return AD_ACK;

    case AUDIOEMU_SET_3D_STEREO_DEPTH:
	audioemu_ac97_read(softstate, AC97_THREE_D_CONTROL_REGISTER, &codecval);
	codecval &= ~TDCR_DEPTH_MASK;
	codecval |= (long)arg & TDCR_DEPTH_MASK;
	    
	audioemu_ac97_write(softstate, AC97_THREE_D_CONTROL_REGISTER, codecval);
	return AD_ACK;

    case AUDIOEMU_SET_3D_STEREO_CENTER:
	audioemu_ac97_read(softstate, AC97_THREE_D_CONTROL_REGISTER, &codecval);
	codecval &= ~TDCR_CENTER_MASK;
	codecval |= (long)arg & TDCR_CENTER_MASK;
	    
	audioemu_ac97_write(softstate, AC97_THREE_D_CONTROL_REGISTER, codecval);
	return AD_ACK;


    default:
	return AD_NACK;
    }
}


static int
audioemu_iocdata(audiohdl_t handle, int channel,
		  queue_t *q, mblk_t *mp, int *error)
{
    struct copyresp *copyresp = (struct copyresp *)mp->b_rptr;
    mblk_t *data_mp;
    caddr_t arg = (caddr_t) copyresp->cp_private;

    if (mp->b_cont == NULL) return AD_NACK;
    data_mp = mp->b_cont;

    DPRINTF4(AUDEMUM_DRIVER, "audioemu_iocdata cp_cmd=%x, arg=%p\n", copyresp->cp_cmd, arg);

    switch (copyresp->cp_cmd) {
    default:
	return AD_NACK;
    }
}


static audio_info_t audioemu_defaults = {
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	/*AUDIO_MID_GAIN*/ 255*75/100,		/* gain level: 0 - 255 */
	AUDIO_SPEAKER,		/* selected I/O port */
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT | AUDIO_SPDIF_OUT,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT | AUDIO_SPDIF_OUT,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_MICROPHONE,		/* selected I/O port */
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD | AUDIO_SPDIF_IN |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD | AUDIO_SPDIF_IN |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    AUDIO_MIN_GAIN,		/* monitor_gain */
    0,				/* output_muted */
#ifdef	AUDIOIO_HAS_REFCNT
    0,				/* driver reference count */
#endif
    {0, 0},			/* reserved fields */
#ifdef	AUDIOIO_HAS_FEATURES
    AUDIO_HWFEATURE_DUPLEX | AUDIO_HWFEATURE_IN2OUT | \
    	AUDIO_HWFEATURE_PLAY | AUDIO_HWFEATURE_RECORD,
    AUDIO_SWFEATURE_MIXER,
    0,
#endif
};


/*
 * Audio Driver ops vector definition
 */
static ah_ad_entry_t audioemu_entry = {
    audioemu_setup,
    audioemu_teardown,
    audioemu_set_config,
    audioemu_set_format,
    audioemu_start_play,
    audioemu_pause_play,
    audioemu_stop_play,
    audioemu_start_record,
    audioemu_stop_record,
    audioemu_ioctl,
    audioemu_iocdata,
    audioemu_get_delay,
};


static audio_device_t audioemu_dev_info = {
    "TOOLS,EMU10Kx",		/* name */
    "1.0",			/* version */
    "EMU10Kx"			/* config */
};


/* 0 terminated list of supported sample rates (used as limits) */
static uint_t audioemu_srs[] = {
    4000, 48000, 0
};


/*
 * 0 terminated list of supported channels. The Audigy audio
 * controller supports 8-bit unsigned and 16-bit signed PCM
 * mono/stereo in the hardware.
 */
static uint_t audioemu_chs[] = {
    AUDIO_CHANNELS_MONO,
    AUDIO_CHANNELS_STEREO,
    0
};

static ah_ad_cap_comb_t audioemu_cap_comb[] = {
    { AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR },  /* 16-bit signed PCM */
    { AUDIO_PRECISION_8,  AUDIO_ENCODING_LINEAR8 }, /* 8-bit unsigned PCM */
    { 0 },
};



static ah_ad_ch_cap_t audioemu_ch_cap = {
    { AD_SRS_FLAG_SR_LIMITS, audioemu_srs },
    audioemu_chs,		/* list of channel types */
    0,				/* interrupt rate */
    32,				/* max channels */
    SAMPLEBUF_SIZE,		/* buffer size */
};
