/*
    Soundblaster 16 PCI driver for Solaris.
    @(#) sbpci.c 2.1@(#)

    Written by Philip Brown <phil@bolthole.com>
    Rewritten by Jrgen Keil <jk@tools.de>

    Copyright  2000      Philip Brown
    Copyright  2000-2002 Jrgen Keil

    This driver is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This driver is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this driver; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
*/

/*
 * This file contains chipset specific code for the Ensoniq 1371 audio
 * controller.  This includes the AC'97 audio codec interface, the
 * hardware samplerate converter and the power management functions.
 */

#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#include <sys/audio.h>
#include "audioio_compat.h"
#include "audio1371.h"
#include "ac97_codec.h"

#include <sys/ddi.h>


/************************************************************/
/* Start of CODEC section.                                  */
/************************************************************/


/*
 * Waits for a while until the SBPCI_CODEC_READBSY is clear.
 *
 * Returns AUDIO1371_SUCCESS if the code is idle or AUDIO1371_FAILURE if
 * the codec is still busy after a maximum wait of 40msecs.
 */
static int
ens1371_wait_for_codecclear(struct audio1371_softstate *softstate)
{
    int loop;
    uint32_t value;

    for (loop = 2000; --loop >= 0; ) {
	value = SBPCI_GET32(SBPCI_1371_CODEC);
	if ((value & SBPCI_CODEC_READBSY) == 0) {
	    /* codec is idle now */
	    return AUDIO1371_SUCCESS;
	}
	drv_usecwait(20);
    }

    return AUDIO1371_FAILURE;
}


/*
 * This writes "through" the codec interface register, to the codec
 * chip hidden "behind" it.
 * It's up the the caller to call wait_for_codecclear() and
 * guarantee the chip is not busy.
 */
static void
ens1371_write_through_codec(struct audio1371_softstate *softstate,
			    uint16_t sub_addr,
			    uint16_t value) 
{
    uint32_t codecvalue = 0;

    codecvalue = (sub_addr&0x7f)<<16;
    codecvalue |= value;

    SBPCI_PUT32(SBPCI_1371_CODEC, codecvalue);
}


/*
 * Waits for a while until SBPCI_CODEC_READBSY is clear and
 * SBPCI_CODEC_READRDY is set.  This condition indicates a 
 * successful codec register read.
 *
 * Returns AUDIO1371_SUCCESS for a successful codec register read -
 * the 16bit codec register value is returned in *data.
 * Returns AUDIO1371_FAILURE if no data is available after after
 * a maximum wait for 40msec.
 */
static int
ens1371_wait_for_codecdata(struct audio1371_softstate *softstate, uint16_t *data)
{
    int loop;
    uint32_t value;

    for (loop = 2000; --loop >= 0; ) {
	value = SBPCI_GET32(SBPCI_1371_CODEC);
	if ((value & (SBPCI_CODEC_READRDY|SBPCI_CODEC_READBSY))
		     == SBPCI_CODEC_READRDY) {

	    /* codec not busy any more + our requested data is ready */
	    *data = value & 0xffff;
	    return AUDIO1371_SUCCESS;
	}
	drv_usecwait(20);
    }

    return AUDIO1371_FAILURE;
}


/*
 * Attempts to read codec register at given subaddress.
 * Note that the subaddress is internal to the codec: NOT addressable normally
 *
 * Returns FALSE on timeout.
 * Returns TRUE otherwise, and the codec register's value in *data.
 * 
 */
static int
ens1371_ac97_read(void *closure, uint16_t sub_addr, uint16_t *data)
{
    struct audio1371_softstate *softstate = closure;
    uint32_t codecreg = 0;
    uint16_t subreg;

    codecreg = (sub_addr&0x7f) << 16;
    codecreg |= SBPCI_CODEC_READBIT;

    if (ens1371_wait_for_codecclear(softstate) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC, "codec not idle before ac97_read");
	return FALSE;
    }

    SBPCI_PUT32(SBPCI_1371_CODEC, codecreg);

    /* wait and read the codec register */
    if (ens1371_wait_for_codecdata(softstate, &subreg) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC, "codec read timeout, reg=%x", sub_addr);
	return FALSE;
    }
    *data = subreg;
    return TRUE;
}


/*
 * like write_though_codec, but this one waits/checks for busy first.
 * returns TRUE on okay, FALSE on failure
 */
static int 
ens1371_ac97_write(void *closure, uint16_t sub_addr, uint16_t value)
{
    struct audio1371_softstate *softstate = closure;

    if (ens1371_wait_for_codecclear(softstate) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC, "ens1371_ac97_write: codec not idle");
	return FALSE;
    }
    ens1371_write_through_codec(softstate, sub_addr, value);
    return TRUE;
}


static void
ens1371_codec_interface_reset(struct audio1371_softstate *softstate)
{
    DPRINTF4(AUD1371M_CODEC, "Resetting CODEC");

    if (softstate->ct5880) {
	SBPCI_PUT32(SBPCI_STATUS, SBPCI_STAT_CT5880_AC97_RESET);
	delay(drv_usectohz(50000));
        /* SBPCI_PUT32(SBPCI_STATUS, 0); ??? not necessary to reset bit? */
    }

    /* raise the AC97WARMRESET bit for 50 usec */
    SBPCI_PUT32(SBPCI_CONTROL, SBPCI_CTRL_AC97WARMRESET);
    drv_usecwait(50);
    SBPCI_PUT32(SBPCI_CONTROL, 0);
    drv_usecwait(2);

    DPRINTF4(AUD1371M_CODEC, "Done with CODEC reset");
}


static audio_codec_handle_t
ens1371_codec_init(struct audio1371_softstate *softstate)
{
    static struct ac97_codec_ops codec_ops = {
	ens1371_ac97_read,
	ens1371_ac97_write
    };

    return ac97_codec_init(softstate->audio_handle, &codec_ops, softstate);
}


/*
 * This configures the power management state of the AudioPCI
 * card.
 */
static int
ens1371_configure_pm(struct audio1371_softstate *softstate, int state)
{
    ddi_acc_handle_t conf_handle;
    uint32_t ctrlreg;
    uint16_t pcistat;

    if ((conf_handle = softstate->pci_conf_handle) == NULL)
	return AUDIO1371_FAILURE;

    /* disable power level interrupts while we change the power level */
    ctrlreg = SBPCI_GET32(SBPCI_CONTROL);
    SBPCI_PUT32(SBPCI_CONTROL, ctrlreg & ~SBPCI_CTRL_PWR_INTRM);

    pcistat = pci_config_get16(conf_handle, PCI_CONF_STAT);
#ifdef	PCI_STAT_CAP
    if (pcistat & PCI_STAT_CAP) {
    	/* set the requested pci power level in pci configuration space */
	uint8_t capptr, capid, capnext;
	uint16_t pcipmstat;

	capptr = pci_config_get8(conf_handle, PCI_CONF_CAP_PTR);
	for (; (capptr &= ~3) != PCI_CAP_NEXT_PTR_NULL; capptr = capnext) {
	    capid   = pci_config_get8(conf_handle, capptr + PCI_CAP_ID);
	    capnext = pci_config_get8(conf_handle, capptr + PCI_CAP_NEXT_PTR);

	    switch (capid) {
	    case PCI_CAP_ID_PM:
		pcipmstat = pci_config_get16(conf_handle, capptr + PCI_PMCSR);
		pcipmstat &= ~PCI_PMCSR_STATE_MASK;
		pcipmstat |= state & 3;
		pci_config_put16(conf_handle, capptr + PCI_PMCSR, pcipmstat);
		break;
	    }
	}
    }
#endif

    /* set the audiopci card's power level */
    ctrlreg &= ~SBPCI_CTRL_PDLEV_MASK;
    ctrlreg |= (state & 3) << SBPCI_CTRL_PDLEV_SHIFT;
    SBPCI_PUT32(SBPCI_CONTROL, ctrlreg);

    return AUDIO1371_SUCCESS;
}


/************************************************************/
/*    Start of SampleRate section                           */
/************************************************************/

/* You have to wait until the rate register is clear of busybit before
 * you can set it! :-)
 *
 * Unfortunately, the "sample rate converter" is a "PIO" type interface.
 * You set the addr you want to write to in the high bits. (0xff000000)
 * And you set the values you want to WRITE, in the low word (0xffff)
 * FURTHERMORE, you have to set the "write enable" bit if you want to write
 * ARRG.
 *
 * This routine returns full value of rate register in the *data
 * pointer, if data is != NULL.
 *
 * This routine returns AUDIO1371_SUCCESS/AUDIO1371_FAILURE.  If AUDIO1371_FAILURE
 * is returned, a timeout has happened.
 */
static int
ens1371_wait_for_rateclear(struct audio1371_softstate *softstate, uint32_t *data) 
{
    int loop;
    uint32_t value = 0;

    for (loop = 1000; --loop >= 0; ) {
	value = SBPCI_GET32(SBPCI_1371_SAMPLERATE);
	if ((value & SBPCI_SAMPLE_BUSY) == 0) {
	    break;
	}
	drv_usecwait(10);
    }
    if (data != NULL) *data = value;
    return loop >= 0 ? AUDIO1371_SUCCESS : AUDIO1371_FAILURE;
}


/*
 * Write to "Sample Rate Converter" register.
 * See comments with wait_for_rateclear().
 * You probably want to 'or' your data with what comes out of there before
 * writing it here.
 *
 * There is no "read_from_rate()". Just use wait_for_rateclear()
 * BEFORE CALLING THIS ROUTINE
 */
static void
ens1371_write_to_rate(struct audio1371_softstate *softstate, uint32_t value)
{
    SBPCI_PUT32(SBPCI_1371_SAMPLERATE, value);
}


/*
 * This uses the Sample Rate Converter register, to read "through" to
 * the actual S.R.C. RAM hidden behind it.
 *
 * The 16-bit word from the S.R.C. RAM is returned in *data.
 * This routine returns AUDIO1371_FAILURE on sample rate register
 * timesouts (*data not valid), or AUDIO1371_SUCCESS otherwise.
 */
static int
ens1371_read_through_rate(struct audio1371_softstate *softstate,
			  uint8_t subaddr,
			  uint16_t *data)
{
    uint32_t ratereg;

    if (ens1371_wait_for_rateclear(softstate, &ratereg) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC,
		 "read_through_rate: sample reg busy: subreg %x", subaddr);
	return AUDIO1371_FAILURE;
    }
    subaddr &= 0x7f;

    /* yes, keep anything disabled that was disabled previously */
    ratereg &= SBPCI_SAMPLE_DISMASK;
    ratereg |= (subaddr << 25);
    ens1371_write_to_rate(softstate, ratereg);

    if (ens1371_wait_for_rateclear(softstate, &ratereg) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC, "read_through_rate: timeout");
	return AUDIO1371_FAILURE;
    }

    *data = ratereg & 0xffff;

    return AUDIO1371_SUCCESS;
}

/* This uses the Sample Rate Converter register, to write "through" to
 * the actual S.R.C. RAM hidden behind it.
 * It also actually tries to wait for the samplerate register to not
 * be busy before writing.
 */
static int
ens1371_write_through_rate(struct audio1371_softstate *softstate,
			   uint8_t subaddr,
			   uint16_t value)
{

    uint32_t ratereg;

    if (ens1371_wait_for_rateclear(softstate, &ratereg) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC, "write_through_rate: sample reg busy: subreg %x := %x",
		 subaddr, value);
	return AUDIO1371_FAILURE;
    }

    subaddr &= 0x7f;
    ratereg &= SBPCI_SAMPLE_DISMASK;
    ratereg |= (subaddr << 25) | (value & 0xffff);
    ratereg |= SBPCI_SAMPLE_WR_EN;

    ens1371_write_to_rate(softstate, ratereg);
    if (ens1371_wait_for_rateclear(softstate, NULL) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC, "write_through_rate: sample reg wr-busy: subreg %x := %x",
		 subaddr, value);
	return AUDIO1371_FAILURE;
    }

    return AUDIO1371_SUCCESS;
}


/*
 * enable/disable the "Sample Rate Converter" for particular gizmo.
 * dacnum: 0	"global sample rate converted enable/disable"
 *	   1,2	DAC1/DAC2
 *	   3	ADC
 * enable_flag: 1 enables sample rate conversion for a particular
 *		dac,
 *		0 disables it
 * old_enable_flag: returns the original enable/disable status
 */
static int 
ens1371_enable_sample_rate_conv(struct audio1371_softstate *softstate,
				int dacnum,
				int enable_flag,
				int *old_enable_flag)
{
    uint32_t cur_val;
    uint32_t disable_bit;

    switch (dacnum) {
    case 0:
	disable_bit = SBPCI_SAMPLE_DISABLE;
	break;
    case 1:
	disable_bit = SBPCI_SAMPLE_DIS_DAC1;
	break;
    case 2:
	disable_bit = SBPCI_SAMPLE_DIS_DAC2;
	break;
    case 3:
	disable_bit = SBPCI_SAMPLE_DIS_ADC;
	break;
    default:
	DPRINTF4(AUD1371M_CODEC, "invalid dacnum %d", dacnum);
	return AUDIO1371_FAILURE;
    }
    
    if (ens1371_wait_for_rateclear(softstate, &cur_val) != AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC, "enable_sample_rate_conv: sample reg busy");
	return AUDIO1371_FAILURE;
    }
	
    if (old_enable_flag != NULL)
	*old_enable_flag = !(cur_val & disable_bit);

    if (enable_flag)
	cur_val &= ~disable_bit;
    else
	cur_val |= disable_bit;

    ens1371_write_to_rate(softstate, cur_val);
    return AUDIO1371_SUCCESS;
}



static int
ens1371_init_sample_rate_conv(struct audio1371_softstate *softstate)
{
    int count;

    if (ens1371_enable_sample_rate_conv(softstate, 0, FALSE, NULL) != AUDIO1371_SUCCESS ||
	ens1371_enable_sample_rate_conv(softstate, 1, FALSE, NULL) != AUDIO1371_SUCCESS ||
	ens1371_enable_sample_rate_conv(softstate, 2, FALSE, NULL) != AUDIO1371_SUCCESS ||
	ens1371_enable_sample_rate_conv(softstate, 3, FALSE, NULL) != AUDIO1371_SUCCESS) {

	DPRINTF4(AUD1371M_CODEC, "ens1371_enable_sample_rate_conv() timed out!!");
	return AUDIO1371_FAILURE;
    }

    /* zero out sample rate conversion memory and sample rate config info */
    for (count = 0; count < 0x80; count++) {
	if (ens1371_write_through_rate(softstate, count, 0) != AUDIO1371_SUCCESS) {
	    DPRINTF4(AUD1371M_CODEC, "zeroing of sample rate memory failed");
	    return AUDIO1371_FAILURE;
	}
    }

    DPRINTF4(AUD1371M_CODEC, "starting default dac1/dac2 stuff now");
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_DAC1_TRUNC, ENS_SAMPLE_SET_N(16));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_DAC1_INT,   ENS_SAMPLE_SET_VFI(16));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_DAC2_TRUNC, ENS_SAMPLE_SET_N(16));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_DAC2_INT,   ENS_SAMPLE_SET_VFI(16));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_ADC_TRUNC,  ENS_SAMPLE_SET_N(16));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_ADC_INT,    ENS_SAMPLE_SET_VFI(16));

    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_DAC1_VOL,  ENS_SAMPLE_SET_DAC_VOLI(1));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_DAC1_VOL2, ENS_SAMPLE_SET_DAC_VOLI(1));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_DAC2_VOL,  ENS_SAMPLE_SET_DAC_VOLI(1));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_DAC2_VOL2, ENS_SAMPLE_SET_DAC_VOLI(1));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_ADC_VOL,   ENS_SAMPLE_SET_ADC_VOL(16));
    ens1371_write_through_rate(softstate,
		       ENS_SAMPLE_ADC_VOL2,  ENS_SAMPLE_SET_ADC_VOL(16));

    return AUDIO1371_SUCCESS;
}


/* rate is between 4000 and 48000 (hz) */
static void
ens1371_setADCfreq(struct audio1371_softstate *softstate, int rate)
{
    uint_t freq, truncm, n;
    uint16_t tmp16;

    if (rate > 48000)
	rate = 48000;
    if (rate < 4000)
	rate = 4000;
    
    if (rate >= 24000) {
	/*
	 * ignore rounding with the (rate / 3000) division for high
	 * sample rates.  It somehow messes up recording at 44.1kHz.
	 */
	n = rate / 3000;
	truncm = (21 * n - 1) | 1;    
	freq = (48000 << 15) / rate * n;
	if (truncm > 239)
	    truncm = 239;
	truncm = ENS_SAMPLE_SET_TRUNC((239 - truncm) / 2);
    } else {
	/*
	 * use the original code with rounding for lower sample rates.
	 */
	n = (rate + 1500) / 3000;
	truncm = (21 * n - 1) | 1;    
	freq = (48000 << 15) / rate * n;
	if (truncm > 119)
	    truncm = 119;
	truncm = ENS_SAMPLE_SMF | ENS_SAMPLE_SET_TRUNC((119 - truncm) / 2);
    }
    truncm |= ENS_SAMPLE_SET_N(n);
    ens1371_write_through_rate(softstate, ENS_SAMPLE_ADC_TRUNC, truncm);

    if (ens1371_read_through_rate(softstate, ENS_SAMPLE_ADC_INT, &tmp16)
					!= AUDIO1371_SUCCESS) {
	DPRINTF4(AUD1371M_CODEC, "setADCfreq: cannot read sample mem");
    }
    ens1371_write_through_rate(softstate, ENS_SAMPLE_ADC_INT,
		       ENS_SAMPLE_SET_VFI(freq >> 15) | (tmp16 & 0xff));
    ens1371_write_through_rate(softstate, ENS_SAMPLE_ADC_VFREQ,
		       freq & 0x7fff);
    ens1371_write_through_rate(softstate, ENS_SAMPLE_ADC_VOL,
		       ENS_SAMPLE_SET_ADC_VOL(n));
    ens1371_write_through_rate(softstate, ENS_SAMPLE_ADC_VOL2,
		       ENS_SAMPLE_SET_ADC_VOL(n));

    DPRINTF4(AUD1371M_CODEC, "sample rate=%d: freq=%x, n=%x", rate, freq, n);
}


/* rate is between 4000 and 48000. dacnum is 1 or 2 */
static void
ens1371_setDACfreq(struct audio1371_softstate *softstate, int rate, int dacnum)
{

    int freq;
    int oldval;
    uint16_t tmp16;

    if (dacnum < 1 || dacnum > 2) {
	DPRINTF4(AUD1371M_CODEC, "setDACfreq called with wrong dacnum");
	return;
    }
    if (rate > 48000)
	rate = 48000;
    if (rate < 4000)
	rate = 4000;
#if 0	/* removed this hack after DMA re-write */
    /*
     * sometimes the hardware does not interrupt when 48000 is an
     * exact multiple of the playback sample rate.  Avoid these
     * frequencies.
     */
    if (rate < 48000 && 48000 % rate == 0)
	rate++;
#endif
    freq = (rate << 15) / 3000;
	
    ens1371_enable_sample_rate_conv(softstate, dacnum, FALSE, &oldval);
    
    /*
     * I feel like I should be doing the following additional register
     * set. But I don't understand this stuff, and the other drivers
     * just do this once at init. So I won't do it
     *
    write_through_rate(softstate, SBPCI_SAMPLE_DAC1_TRUNC,0x100);
     */

    switch (dacnum) {
    case 1:
	if (ens1371_read_through_rate(softstate, ENS_SAMPLE_DAC1_INT, &tmp16)
		!= AUDIO1371_SUCCESS) {
	    break;
	}
	ens1371_write_through_rate(softstate, ENS_SAMPLE_DAC1_INT,
			   ENS_SAMPLE_SET_VFI(freq >> 15) | (tmp16 & 0xff));
	ens1371_write_through_rate(softstate, ENS_SAMPLE_DAC1_VFREQ,
			   freq & 0x7fff);
	break;
    case 2:
	if (ens1371_read_through_rate(softstate, ENS_SAMPLE_DAC2_INT, &tmp16)
		!= AUDIO1371_SUCCESS) {
	    break;
	}
	ens1371_write_through_rate(softstate, ENS_SAMPLE_DAC2_INT, 
			   ENS_SAMPLE_SET_VFI(freq >> 15) | (tmp16 & 0xff));
	ens1371_write_through_rate(softstate, ENS_SAMPLE_DAC2_VFREQ,
			   freq & 0x7fff);
	break;
    }

    ens1371_enable_sample_rate_conv(softstate, dacnum, oldval, NULL);
}

/************************************************************/
/*    End of  SampleRate section.                           */
/************************************************************/

ens_chip_ops_t ens1371_ops = {
    ens1371_codec_init,

    ens1371_codec_interface_reset,

    ens1371_configure_pm,

    ens1371_init_sample_rate_conv,
    ens1371_enable_sample_rate_conv,
    ens1371_setADCfreq,
    ens1371_setDACfreq
};
