/*-
 * Copyright (C) 2002 Juergen Keil.
 * Copyright (C) 2002 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#include <sys/audio.h>
#include "audioio_compat.h"
#include "akm4531.h"
#include "akm4531_codec.h"

#include <sys/ddi.h>

#if 0
#define	DPRINTF(...)	printf(__VA_ARGS__)
#else
#define	DPRINTF(...)	/* */
#endif


/*
 * Cleanup routine, when a codec object is no longer used.
 */
static void
akm4531_delete(struct audio_codec *base)
{
    struct akm4531_codec *ac = (struct akm4531_codec *)base;
    kmem_free(ac, sizeof(struct akm4531_codec));
}



/*
 * helper subroutine to convert sun audio_info gain values
 * to akm4531 codec volume levels.  akm4531 volume is an
 * "attenuation",  lower values means "louder".
 *
 * we use a linear mapping of [0..maxvol] -> [codec_max..0]
 */
static int
akm4531_codec_vol(int vol, int maxvol, int codec_max)
{
    if (vol < 0 || vol > maxvol) {
	/*
	 * openwin audiotool sends illegal "balance" values,
	 * handle them gracefully.
	 *
	DPRINTF4("illegal volume %d", vol);
	 */
	if (vol < 0)
	    vol = 0;
	else if (vol > maxvol)
	    vol = maxvol;
    }

    return (maxvol - vol) * codec_max / maxvol;
}


static void
akm4531_configure_playback_vol(struct akm4531_codec *ac,
			       audio_info_t         *audio_info)
{
    audio_prinfo_t *play = &audio_info->play;
    int voll, volr;
    int mute;
    int lfactor, rfactor;	/* 0.0 .. 1.0 (scaled integer) */
#define	BALANCE_SCALE	(8)

    if (play->balance >= AUDIO_MID_BALANCE) {
	/* 
	 * MID_BALANCE..RIGHT_BALANCE
	 * full gain on right channel, reduced gain on left channel.
	 */
	lfactor = AUDIO_RIGHT_BALANCE - play->balance;
	lfactor = (lfactor << BALANCE_SCALE) / AUDIO_MID_BALANCE;
	rfactor = 1 << BALANCE_SCALE;
    } else {
	/* 
	 * LEFT_BALANCE..MID_BALANCE
	 * full gain on left channel, reduced gain on right channel.
	 */
	lfactor = 1 << BALANCE_SCALE;
	rfactor = play->balance;
	rfactor = (rfactor << BALANCE_SCALE) / AUDIO_MID_BALANCE;
    }
    voll = akm4531_codec_vol((play->gain * lfactor) >> BALANCE_SCALE,
			     AUDIO_MAX_GAIN, AK_ATT_MASK);
    volr = akm4531_codec_vol((play->gain * rfactor) >> BALANCE_SCALE,
			     AUDIO_MAX_GAIN, AK_ATT_MASK);

    mute = (audio_info->output_muted
	    || !(play->port & (AUDIO_SPEAKER|AUDIO_HEADPHONE)))
	? AK_MUTE
	: 0;

    AKM4531_WRITE(ac, AK_MASTER_VOLUME_LCH, mute | voll);
    AKM4531_WRITE(ac, AK_MASTER_VOLUME_RCH, mute | volr);
    DPRINTF("MASTERVOL=%02x %02x", mute|voll, mute|volr);


    voll = akm4531_codec_vol(play->gain,
			     AUDIO_MAX_GAIN, AK_MONO_ATT_MASK);

    AKM4531_WRITE(ac, AK_MONOOUT_VOLUME, mute | voll);
}


static void
akm4531_configure_record_vol(struct akm4531_codec *ac,
			     audio_info_t	  *audio_info)
{
    audio_prinfo_t *record = &audio_info->record;
    int lfactor, rfactor;	/* 0.0 .. 1.0 (scaled integer) */
    int vol, voll, volr;
    uint8_t mute;
    uint8_t om;
    uint8_t recl, recr;

    /*
     * the following AKM4541 codec registers are for *playback*
     * (monitoring of recording)
     * the akm4531 mixer can output all of these (+ the PCM playback)
     * concurrently.  But there's only one gain control for both
     * recording and monitoring available.  We use the record->gain
     * value to program the codec's input gain, and we use
     * the field monitor_gain as a toggle to either enable or disable
     * monitoring of any input source.
     */
    om = AK_OM_FML | AK_OM_FMR;
    if (audio_info->monitor_gain >= AUDIO_MID_GAIN) {
	if (record->port & AUDIO_MICROPHONE)
	    om |= AK_OM_MIC;
	if (record->port & AUDIO_CD)
	    om |= AK_OM_CDL | AK_OM_CDR;
	if (record->port & AUDIO_LINE_IN)
	    om |= AK_OM_LINEL | AK_OM_LINER;
    }
    AKM4531_WRITE(ac, AK_OUTPUT_MIXER_SW1, om);
    DPRINTF("OUTMIX SW1=%02x", om);

    om = AK_OM_VOICEL | AK_OM_VOICER;
    if (audio_info->monitor_gain >= AUDIO_MID_GAIN) {
	if (record->port & AUDIO_AUX1_IN)
	    om |= AK_OM_AUXL | AK_OM_AUXR;
    }
    AKM4531_WRITE(ac, AK_OUTPUT_MIXER_SW2, om);
    DPRINTF("OUTMIX SW2=%02x", om);

    /*
     * The following AKM4531 codec registers are for *recording*.
     */
    vol = akm4531_codec_vol(record->gain, AUDIO_MAX_GAIN, AK_GAIN_MASK);

    if (record->balance >= AUDIO_MID_BALANCE) {
	/* 
	 * MID_BALANCE..RIGHT_BALANCE
	 * full gain on right channel, reduced gain on left channel.
	 */
	lfactor = AUDIO_RIGHT_BALANCE - record->balance;
	lfactor = (lfactor << BALANCE_SCALE) / AUDIO_MID_BALANCE;
	rfactor = 1 << BALANCE_SCALE;
    } else {
	/* 
	 * LEFT_BALANCE..MID_BALANCE
	 * full gain on left channel, reduced gain on right channel.
	 */
	lfactor = 1 << BALANCE_SCALE;
	rfactor = record->balance;
	rfactor = (rfactor << BALANCE_SCALE) / AUDIO_MID_BALANCE;
    }

    voll = akm4531_codec_vol((record->gain * lfactor) >> BALANCE_SCALE,
			     AUDIO_MAX_GAIN, AK_GAIN_MASK);
    volr = akm4531_codec_vol((record->gain * rfactor) >> BALANCE_SCALE,
			     AUDIO_MAX_GAIN, AK_GAIN_MASK);

    mute = (record->port & AUDIO_CD) ? 0 : AK_MUTE;
    AKM4531_WRITE(ac, AK_CD_AUDIO_VOLUME_LCH, mute | voll);
    AKM4531_WRITE(ac, AK_CD_AUDIO_VOLUME_RCH, mute | volr);
    DPRINTF("CDVOL=%02x %02x", mute|voll, mute|volr);

    mute = (record->port & AUDIO_LINE_IN) ? 0 : AK_MUTE;
    AKM4531_WRITE(ac, AK_LINE_VOLUME_LCH, mute | voll);
    AKM4531_WRITE(ac, AK_LINE_VOLUME_RCH, mute | volr);
    DPRINTF("LINEINVOL=%02x %02x", mute|voll, mute|volr);

    mute = (record->port & AUDIO_AUX1_IN) ? 0 : AK_MUTE;
    AKM4531_WRITE(ac, AK_AUX_VOLUME_LCH, mute | voll);
    AKM4531_WRITE(ac, AK_AUX_VOLUME_RCH, mute | volr);
    DPRINTF("AUXVOL=%02x %02x", mute|voll, mute|volr);

    /* mono1 & mono2 omitted here */

    mute = (record->port & AUDIO_MICROPHONE) ? 0 : AK_MUTE;
    AKM4531_WRITE(ac, AK_MIC_VOLUME, mute | vol);
    DPRINTF("MICVOL=%02x", mute|vol);


    recl = recr = 0;
    if (record->port & AUDIO_MICROPHONE) {
	recl |= AK_IM_MIC;
	recr |= AK_IM_MIC;
    }
    if (record->port & AUDIO_CD) {
	recl |= AK_IM_CDL;
	recr |= AK_IM_CDR;
    }
    if (record->port & AUDIO_LINE_IN) {
	recl |= AK_IM_LINEL;
	recr |= AK_IM_LINER;
    }
    AKM4531_WRITE(ac, AK_LCH_INPUT_MIXER_SW1, recl);
    AKM4531_WRITE(ac, AK_RCH_INPUT_MIXER_SW1, recr);
    DPRINTF("INMIX SW1=%02x %02x", recl, recr);

    recl = recr = 0;
    if (record->port & AUDIO_AUX1_IN) {
	recl |= AK_IM_AUXL;
	recr |= AK_IM_AUXR;
    }
    AKM4531_WRITE(ac, AK_LCH_INPUT_MIXER_SW2, recl);
    AKM4531_WRITE(ac, AK_RCH_INPUT_MIXER_SW2, recr);
    DPRINTF("INMIX SW2=%02x %02x", recl, recr);
}


/*
 * First time initialization of the audio codec.
 *
 * We reset and power cycle the AK4531 audio codec here. This
 * will initialize all codec register to default values.
 * 
 * Some global audio codec settings are performed as well.
 */
static int
akm4531_reset(struct audio_codec *base, audio_info_t *audio_info, int verbose)
{
    struct akm4531_codec *ac = (struct akm4531_codec *)base;

    DPRINTF("reset akm4531 codec\n");

    /* reset and power down the codec */
    if (!AKM4531_WRITE(ac, AK_RESET_POWER_DOWN, 0))
	return 0;
    drv_usecwait(10);

    /* back to normal operation */
    if (!AKM4531_WRITE(ac, AK_RESET_POWER_DOWN, AK_RST_NEG|AK_PD_NEG))
	return 0;

    /* 
     * There's a short period while the codec is busy initializing.
     * Wait until the codec is done with it's initialization and is
     * idle, waiting for commands.
     */
    AKM4531_WAIT_IDLE(ac);

    /* Enable the PLL clock for DAC2 */
    AKM4531_WRITE(ac, AK_CLOCK_SELECT, 0);

#if 0
    /*
     * Messing with the clocks seems to trigger another long codec-busy
     * period.
     */
    AKM4531_WAIT_IDLE(ac);
#endif

    /*
     * Setup the mixer registers, according to defaults from the
     * audio driver's "audio_info" structure
     */
    akm4531_configure_playback_vol(ac, audio_info);
    akm4531_configure_record_vol(ac, audio_info);
    
    /*
     * select the codec's input mixer as record source
     * (CD/Microphone/Line-in) 
     */
    AKM4531_WRITE(ac, AK_AD_INPUT_SELECT, AK_ADSEL_INPUT_MIXER);

    /* enable the optional microphone amplifier */
    AKM4531_WRITE(ac, AK_MIC_AMP_GAIN, 
		  ac->akm4531_mic_30db_boost ? AK_MGAIN_30dB : AK_MGAIN_0dB);
    
    /*
     * Un-mute and set the default gain for the "PCM wave-out"
     * channel.  The audio_info structure does not contain a field to
     * control this gain for the pcm wave-out channel, so we
     * configure a global value here.
     */

    /* DAC2, PLL variable sample rates */
    AKM4531_WRITE(ac, AK_VOICE_VOLUME_LCH, ac->akm4531_pcmvol);
    AKM4531_WRITE(ac, AK_VOICE_VOLUME_RCH, ac->akm4531_pcmvol);

    /* DAC1, Codec FM, 4 fixed sample rates */
    AKM4531_WRITE(ac, AK_FM_VOLUME_LCH, ac->akm4531_pcmvol);
    AKM4531_WRITE(ac, AK_FM_VOLUME_RCH, ac->akm4531_pcmvol);

    AKM4531_WRITE(ac, AK_MONO1_VOLUME, ac->akm4531_pcmvol);
    AKM4531_WRITE(ac, AK_MONO2_VOLUME, ac->akm4531_pcmvol);

    return 1;
}


/*
 * Handle audio driver "set_config" callbacks from the audio
 * framework.  The audio driver typically just dispatches the
 * "set_config" call to this codec specific implementation.
 */
static int
akm4531_set_config(struct audio_codec *base,
		   int command, int dir, int arg1, int arg2)
{
    struct akm4531_codec *ac = (struct akm4531_codec *)base;
    int config_vol = 0;

    switch (command) {
    case AD_SET_GAIN:
	if (dir == AUDIO_PLAY)
	    ac->audio_info.play.gain = arg1;
	else
	    ac->audio_info.record.gain = arg1;
	config_vol = dir;
	break;
    case AD_SET_GAIN_BAL:
	if (dir == AUDIO_PLAY)
	    ac->audio_info.play.balance = arg1;
	else
	    ac->audio_info.record.balance = arg1;
	config_vol = dir;
	break;
    case AD_SET_PORT:
	if (dir == AUDIO_PLAY)
	    ac->audio_info.play.port = arg1;
	else
	    ac->audio_info.record.port = arg1;
	config_vol = dir;
	break;
    case AD_SET_MONITOR_GAIN:
	ac->audio_info.monitor_gain = arg1;
	config_vol = AUDIO_RECORD;
	break;
    case AD_OUTPUT_MUTE:
	ac->audio_info.output_muted = arg1;
	config_vol = AUDIO_PLAY;
	break;
    case AD_MONO_MIC:
	break;
    case AD_MIC_BOOST:
	break;
    case AD_BASS_BOOST:
	break;
    case AD_MID_BOOST:
	break;
    case AD_TREBLE_BOOST:
	break;
    case AD_LOUDNESS:
	break;
    case AD_SET_DIAG_MODE:
	break;
    }
    
    if (config_vol & AUDIO_PLAY) {
	akm4531_configure_playback_vol(ac, &ac->audio_info);
    } else if (config_vol & AUDIO_RECORD) {
	akm4531_configure_record_vol(ac, &ac->audio_info);
    }

    return 0;
}


static struct audio_codec_ops akm4531_codec_ops = {
    akm4531_delete,
    akm4531_reset,
    akm4531_set_config
};


/*
 * Initialize and return a "codec object" for an Asahi Kasei 4531 
 * audio codec.
 */
audio_codec_handle_t
akm4531_codec_init(audiohdl_t handle,
		   struct akm4531_codec_ops *akm4531_ops, void *akm4531_data)
{
    struct akm4531_codec *ac;
    dev_info_t *dip = audio_hlp_get_dip(handle);

    ac = kmem_alloc(sizeof(struct akm4531_codec), KM_NOSLEEP);
    ac->base.ops = &akm4531_codec_ops;
    ac->audio_handle = handle;
    ac->akm4531_ops = akm4531_ops;
    ac->akm4531_data = akm4531_data;

    /*
     * the akm4531 has a 30db microphone boost, while the ordinary AC97
     * codec used 20db.  For now, I keep the AC97 "mic20db" option and
     * reuse it for the akm4531 codec as well.
     */
    ac->akm4531_mic_30db_boost =
	ddi_getprop(DDI_DEV_T_ANY, dip,
		    DDI_PROP_DONTPASS | DDI_PROP_CANSLEEP | DDI_PROP_NOTPROM,
		    "mic20db", 1);

    ac->akm4531_pcmvol =
	ddi_getprop(DDI_DEV_T_ANY, dip,
		    DDI_PROP_DONTPASS | DDI_PROP_CANSLEEP | DDI_PROP_NOTPROM,
		    "pcmvol", 6) & AK_GAIN_MASK;

    return (audio_codec_handle_t)ac;
}



