/*-
 * Copyright (C) 2001-2003 Juergen Keil.
 * Copyright (C) 2001-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if	_KERNEL

#include <sys/sunddi.h>
#include <sys/ksynch.h>
#include <sys/note.h>
#include <sys/mixer.h>
#include <sys/audio.h>
#include <sys/audio/audio_helper.h>

#define	SAMPLEBUF_INTR_CHUNK	0x2000	/* one intr every 8k data */
#define	SAMPLEBUF_SIZE		(3*SAMPLEBUF_INTR_CHUNK)

#define	V8233_NUM_SGD	32


/*
 * An 8K (SAMPLEBUF_INTR_CHUNK) chunk of data of DMA-able memory
 * for which we generate one interrupt from the VIA audio controller.
 */
struct sgd_chunk {
    caddr_t		data;		/* kernel vaddr of data */

    ddi_dma_handle_t	dma_handle;
    uint8_t		dma_bound;
    size_t		reallen;
    ddi_acc_handle_t	acc_handle;

    struct sgd_chunk   *next;		/* next DMA data chunk, when on free list */
};


/*
 * 
 */
struct sample_buffer {
    uint8_t		io_started;
    uint8_t		head;		/* VIA audio controller accesses sample */
    					/* buffer starting with this index */
    uint8_t		tail;		/* first free slot after the end of valid  */
    					/* buffers */
    uint8_t		buffers_available;
    struct sgd_chunk   *active[V8233_NUM_SGD];
					/* while chunks are in use by the VIA audio
					 * controller, they are placed in the active 
					 * array under their current bdl slot index
					 */
    struct sgd_chunk   *free;		/* list of unused sample chunks */
};
typedef struct sample_buffer sample_buffer_t;


/*
 * SGD buffer descriptor list for DMA
 * From Page 104 of the VIA VT82C686A data sheet
 */

struct v8233_buf_descr {
    uint32_t	bd_base;	/* location of the data buffer (must be an even addr) */
    uint32_t    bd_len:24;	/* length of buffer, in bytes? */
    uint8_t	:5;
    uint8_t	bd_stop:1;	/*  */
    uint8_t	bd_flag:1;	/* interrupt on completion flag */
    uint8_t	bd_eol:1;	/* interrupt, and restart SGD DMA  */
};
typedef struct v8233_buf_descr v8233_buf_descr_t;


/*
 * We allocate the 32-entry buffer descriptor lists for each 
 * of the three DMA engines available in the VIA audio controller
 * in one single 4k physical page:
 */
struct v8233_sgd {
    v8233_buf_descr_t	pcm_out_sgd[V8233_NUM_SGD];
    v8233_buf_descr_t	pcm_in_sgd[V8233_NUM_SGD];
    v8233_buf_descr_t	fm_out_sgd[V8233_NUM_SGD];
};
typedef struct v8233_sgd v8233_sgd_t;


typedef struct audiov8233_softstate {
    dev_info_t         *dip;            /* device info node for your sound hw */
    audiohdl_t		audio_handle;	/* opaque handle to audio support module */

    audio_codec_handle_t codec_handle;

    kmutex_t            instance_mutex;
    kmutex_t            intr_mutex;
    ddi_iblock_cookie_t	intr_cookie;

    caddr_t		regs;		/* AC97 audio I/O space registers */
    ddi_acc_handle_t	regs_handle;

    ddi_acc_handle_t	pci_conf_handle;

    char               *revstr;         /* pci revision of card as printable string */
    uint8_t             rev;            /* pci revision byte of card */

    uint16_t		ac97_ext_features;
    int			playback_reg_base; /* offs for playback register block: */
    					/* V8233_SGD_RDCHN_IOBASE/V8233_SGD_MP_IOBASE */


    ddi_dma_handle_t	sgd_dma_handle;	/* DMA handle for buffer descriptor list */
    int			sgd_dma_bound;
    ddi_acc_handle_t	sgd_acc_handle;
    v8233_sgd_t	       *sgd;		/* virtual kernel addr of sgd */
    v8233_sgd_t	       *sgd_paddr;	/* physical   	    "	      */
    size_t		sgdlen;

    sample_buffer_t	record_buf;
    sample_buffer_t	play_buf;
    timeout_id_t	play_timeout_id;

    audio_info_t	audio_info;	/* driver local audio state. Note: */
    					/* we maintain only a few fields */
    					/* of the structure in the driver */

    audio_device_t	dev_info;
    ah_ad_info_t	ad_info;
} audiov8233_softstate_t;

#define	TRUE	1
#define	FALSE	0

/* general success/failure return codes */
#define	AUDIOV8233_SUCCESS	0
#define	AUDIOV8233_FAILURE	(-1)


/*
 * VIA audio codec specific PCI configuration registers
 */
#define PCI_CONF_V8233_AC97_STAT	0x40	/* AC97 interface status, 1 byte */
#define PCI_CONF_V8233_AC97_CTRL	0x41	/* AC Link interface control, 1 byte */
#define	PCI_CONF_V8233_FUNC_CTRL	0x42	/* Function enable, 1 byte */
#define	PCI_CONF_V8233_PNP_CTRL	0x43	/* Plug&Play control, 1 byte */
#define	PCI_CONF_V8233_MC97_CTRL	0x44	/* MC97 interface control, 1 byte */
#define	PCI_CONF_V8233_FM_CTRL	0x48	/* FM NMI control, 1 byte */


/*
 * AC97 interface status register (0x40) 
 */
#define	PCI_V8233_ACSTAT_SECONDARY_CODEC_READY	4
#define	PCI_V8233_ACSTAT_CODEC_LOWPOWER		2
#define PCI_V8233_ACSTAT_PRIMARY_CODEC_READY	1

/*
 * AC Link interface control (0x41)
 */
#define	PCI_V8233_ACCTL_ENAC97			0x80
#define	PCI_V8233_ACCTL_ACRST			0x40
#define	PCI_V8233_ACCTL_RSYNC_HI			0x20
#define	PCI_V8233_ACCTL_SDO_HI			0x10
#define	PCI_V8233_ACCTL_VSR_ON_DEMAND		0x08
#define	PCI_V8233_ACCTL_SGD_READ_CHANNEL		0x04
#define	PCI_V8233_ACCTL_FM_CHANNEL		0x02
#define	PCI_V8233_ACCTL_SB_CHANNEL		0x01

/*
 * Function enable (0x42)
 */
#define	PCI_V8233_FCTL_MIDI_PNP			0x40
#define	PCI_V8233_FCTL_Rx2C_WRITE		0x20
#define	PCI_V8233_FCTL_SB_PCM_WHEN_FIFO_EMPTY	0x10
#define	PCI_V8233_FCTL_ENGAME			0x08
#define	PCI_V8233_FCTL_ENFM			0x04
#define	PCI_V8233_FCTL_ENMIDI			0x02
#define	PCI_V8233_FCTL_ENSB			0x01

/*
 * macros to access the VIA I/O space registers 
 */
#define	V8233_GET8(offs)						\
	ddi_get8(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)))
#define	V8233_GET16(offs)					\
	ddi_get16(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)))
#define	V8233_GET32(offs)					\
	ddi_get32(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)))

#define	V8233_PUT8(offs, val)					\
	ddi_put8(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)),	\
		  (val))
#define	V8233_PUT16(offs, val)					\
	ddi_put16(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)),	\
		  (val))
#define	V8233_PUT32(offs, val)					\
	ddi_put32(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)),	\
		  (val))

#define	V8233_REGS_SIZE	0x100	/* size of audio controller register block */


#define	V8233_RDCHN_REG(offs)	(softstate->playback_reg_base + (offs))
#define	V8233_WRCHN_REG(offs)	(V8233_SGD_WRCHN_IOBASE + (offs))

/* SGD read channel == playback channel */
#define	V8233_SGD_RDCHN_IOBASE		0x00		/* I/O base */

#define V8233_SGD_RDCHN_STAT		0x00		/* 1 byte */
#define	  V8233_SGD_STAT_ACTIVE		  0x80
#define	  V8233_SGD_STAT_PAUSED		  0x40
#define	  V8233_SGD_STAT_TRIGGER	  0x08
#define	  V8233_SGD_STAT_STOPPED	  0x04
#define	  V8233_SGD_STAT_EOL		  0x02
#define	  V8233_SGD_STAT_FLAG		  0x01
#define	V8233_SGD_RDCHN_CTRL		0x01		/* 1 byte */
#define	  V8233_SGD_CTRL_START		  0x80
#define	  V8233_SGD_CTRL_TERMINATE	  0x40
#define	  V8233_SGD_CTRL_AUTOSTART	  0x20
#define	  V8233_SGD_CTRL_PAUSE		  0x08
#define	  V8233_SGD_CTRL_INTR_STOP	  0x04
#define	  V8233_SGD_CTRL_INTR_EOL	  0x02
#define	  V8233_SGD_CTRL_INTR_FLAG	  0x01
#define	V8233_SGD_RDCHN_VOLL		0x02		/* 1 byte */
#define	V8233_SGD_RDCHN_VOLR		0x03		/* 1 byte */
#define	V8233_SGD_RDCHN_PTR		0x04		/* 4 bytes */
#define	V8233_SGD_RDCHN_TYPE		0x08		/* 1 byte? */
#define	  V8233_SGD_TYPE_STOP_IDX_MASK	  0xFF000000
#define	  V8233_SGD_TYPE_STOP_IDX_SHIFT   24
#define   V8233_SGD_TYPE_16BIT		  0x00200000
#define   V8233_SGD_TYPE_STEREO		  0x00100000
#define	  V8233_SGD_TYPE_VRA_FREQ_MASK	  0x000FFFFF	
#define	V8233_SGD_RDCHN_STOP_IDX	0x0B		/* 1 byte? */
#define	V8233_SGD_RDCHN_CUR_CNT		0x0C		/* 4 bytes */
#define	V8233_SGD_RDCHN_CUR_IDX		0x0F		/* 1 byte */
#define	  V8233_SGD_CHN_CUR_CNT_MASK	  0x00ffffff
#define	  V8233_SGD_CHN_CUR_CNT_SHIFT	  0
#define	  V8233_SGD_CHN_CUR_IDX_MASK	  0xff000000
#define	  V8233_SGD_CHN_CUR_IDX_SHIFT	  24

/* SGD multichannel playback */
#define	V8233_SGD_MP_IOBASE		0x40		/* I/O base */

#define	V8233_SGD_MP_TYPE		0x02		/* 1 byte */
#define	  V8233_SGD_MP_TYPE_16BIT	  0x80
#define	  V8233_SGD_MP_TYPE_8BIT	  0x00
#define	  V8233_SGD_MP_TYPE_CHANNELS(x)	  (((x)&7) << 4)/* #channels: 1,2,4 or 6 */
#define	V8233_SGD_MP_SCRATCH		0x03		/* 1 byte (unused) */
#define	V8233_SGD_MP_SLOT_SELECT	0x08		/* 4 byte */
#define   V8233_SGD_MP_SLOT_MASK	  0x00FFFFFF
#define   V8233_SGD_MP_SLOT3(x)           (x)
#define   V8233_SGD_MP_SLOT4(x)           ((x) << 4)
#define   V8233_SGD_MP_SLOT7(x)           ((x) << 8)
#define   V8233_SGD_MP_SLOT8(x)           ((x) << 12)
#define   V8233_SGD_MP_SLOT6(x)           ((x) << 16)
#define   V8233_SGD_MP_SLOT9(x)           ((x) << 20)

/* SGD write channel == record channel */
#define V8233_SGD_WRCHN_IOBASE		0x60		/* I/O base */

#define V8233_SGD_WRCHN_STAT		0x00		/* 1 byte */
#define	V8233_SGD_WRCHN_CTRL		0x01		/* 1 byte */
#define	V8233_SGD_WRCHN_FIFO		0x02		/* 1 byte */
#define	  V8233_SGD_FIFO_ENABLE		  0x40
#define	V8233_SGD_WRCHN_SEL		0x03		/* 1 byte */
#define	V8233_SGD_WRCHN_PTR		0x04		/* 4 bytes */
#define	V8233_SGD_WRCHN_TYPE		0x08		/* 4 bytes */
#define	V8233_SGD_WRCHN_STOP_IDX	0x0B		/*???*/
#define	V8233_SGD_WRCHN_CUR_CNT		0x0C		/* 4 bytes */
#define	V8233_SGD_WRCHN_CUR_IDX		0x0F		/* 1 byte */

/* AC97 Controller Command/Status */
#define	V8233_AC97_COMMAND		0x80		/* 4 bytes */
#define   V8233_AC97_CMD_CODEC_MASK	  0xC0000000
#define   V8233_AC97_CMD_PRIMARY_CODEC	  0x00000000
#define   V8233_AC97_CMD_SECONDARY_CODEC  0x40000000
#define   V8233_AC97_CMD_SECONDARY_VALID  0x08000000
#define   V8233_AC97_CMD_PRIMARY_VALID	  0x02000000
#define   V8233_AC97_CMD_CONTROLLER_BUSY  0x01000000
#define	  V8233_AC97_CMD_READ_ACCESS	  0x00800000
#define	  V8233_AC97_CMD_WRITE_ACCESS	  0x00000000
#define	  V8233_AC97_CMD_REG_IDX_MASK	  0x007f0000
#define	  V8233_AC97_CMD_REG_IDX_SHIFT	  16
#define	  V8233_AC97_CMD_DATA_MASK	  0x0000ffff

#define	V8233_AC97_CMD_REG_IDX(r)	(((r)&0x7e) << V8233_AC97_CMD_REG_IDX_SHIFT)

#define	V8233_SGD_STATUS_SHADOW		0x84		/* 4 bytes */

#endif  /* _KERNEL */

/* custom 'ioctl' supported by the audioi810 driver */
#define	AUDIOV8233_IOC		('v'<<8)
#define	AUDIOV8233_SET_VSR_VOL (AUDIOV8233_IOC | 0)
