/*-
 * Copyright (C) 2001-2003 Juergen Keil.
 * Copyright (C) 2001-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/audio/audio_helper.h>

#include "audiohlp.h"
#include <sys/ddi.h>


/*
** This routine converts from linear to ulaw.
**
** Craig Reese: IDA/Supercomputing Research Center
** Joe Campbell: Department of Defense
** 29 September 1989
**
** References:
** 1) CCITT Recommendation G.711  (very difficult to follow)
** 2) "A New Digital Technique for Implementation of Any
**     Continuous PCM Companding Law," Villeret, Michel,
**     et al. 1973 IEEE Int. Conf. on Communications, Vol 1,
**     1973, pg. 11.12-11.17
** 3) MIL-STD-188-113,"Interoperability and Performance Standards
**     for Analog-to_Digital Conversion Techniques,"
**     17 February 1987
**
** Input: Signed 16 bit linear sample
** Output: 8 bit ulaw sample
*/

#define ZEROTRAP    /* turn on the trap as per the MIL-STD */
#define BIAS 0x84		/* define the add-in bias for 16 bit samples */
#define CLIP 32635

static unsigned char
linear_to_ulaw(int sample)
{
    static int exp_lut[256] = {0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
			       4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
			       5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
			       5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
			       6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
			       6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
			       6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
			       6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
			       7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
			       7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
			       7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
			       7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
			       7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
			       7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
			       7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
			       7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};
    int sign, exponent, mantissa;
    unsigned char ulawbyte;
	
    /* Get the sample into sign-magnitude. */
    sign = (sample >> 8) & 0x80;		/* set aside the sign */
    if ( sign != 0 ) sample = -sample;		/* get magnitude */
    if ( sample > CLIP ) sample = CLIP;		/* clip the magnitude */
	
    /* Convert from 16 bit linear to ulaw. */
    sample = sample + BIAS;
    exponent = exp_lut[( sample >> 7 ) & 0xFF];
    mantissa = ( sample >> ( exponent + 3 ) ) & 0x0F;
    ulawbyte = ~ ( sign | ( exponent << 4 ) | mantissa );
#ifdef ZEROTRAP
    if ( ulawbyte == 0 ) ulawbyte = 0x02;	/* optional CCITT trap */
#endif
	
    return ulawbyte;
}

/*
** This routine converts from ulaw to 16 bit linear.
**
** Craig Reese: IDA/Supercomputing Research Center
** 29 September 1989
**
** References:
** 1) CCITT Recommendation G.711  (very difficult to follow)
** 2) MIL-STD-188-113,"Interoperability and Performance Standards
**     for Analog-to_Digital Conversion Techniques,"
**     17 February 1987
**
** Input: 8 bit ulaw sample
** Output: signed 16 bit linear sample
*/

static int
ulaw_to_linear(unsigned char ulawbyte)
{
    static int exp_lut[8] = { 0, 132, 396, 924, 1980, 4092, 8316, 16764 };
    int sign, exponent, mantissa, sample;
	
    ulawbyte = ~ ulawbyte;
    sign = ( ulawbyte & 0x80 );
    exponent = ( ulawbyte >> 4 ) & 0x07;
    mantissa = ulawbyte & 0x0F;
    sample = exp_lut[exponent] + ( mantissa << ( exponent + 3 ) );
    if ( sign != 0 ) sample = -sample;
	
    return sample;
}


#define	SIGN_BIT	(0x80)		/* Sign bit for a A-law byte. */
#define	QUANT_MASK	(0xf)		/* Quantization field mask. */
#define	NSEGS		(8)		/* Number of A-law segments. */
#define	SEG_SHIFT	(4)		/* Left shift for segment number. */
#define	SEG_MASK	(0x70)		/* Segment field mask. */

static short seg_end[8] = { 0xFF,  0x1FF,  0x3FF,  0x7FF,
			   0xFFF, 0x1FFF, 0x3FFF, 0x7FFF};

static int
search(int    val,
       short *table,
       int    size)
{
    int	i;

    for (i = 0; i < size; i++) {
	if (val <= *table++)
	    return i;
    }
    return size;
}

/*
 * linear2alaw() - Convert a 16-bit linear PCM value to 8-bit A-law
 *
 * linear2alaw() accepts an 16-bit integer and encodes it as A-law data.
 *
 *		Linear Input Code	Compressed Code
 *	------------------------	---------------
 *	0000000wxyza			000wxyz
 *	0000001wxyza			001wxyz
 *	000001wxyzab			010wxyz
 *	00001wxyzabc			011wxyz
 *	0001wxyzabcd			100wxyz
 *	001wxyzabcde			101wxyz
 *	01wxyzabcdef			110wxyz
 *	1wxyzabcdefg			111wxyz
 *
 * For further information see John C. Bellamy's Digital Telephony, 1982,
 * John Wiley & Sons, pps 98-111 and 472-476.
 */
static unsigned char
linear_to_alaw(int pcm_val)	/* 2's complement (16-bit range) */
{
    int		mask;
    int		seg;
    unsigned char	aval;

    if (pcm_val >= 0) {
	mask = 0xD5;		/* sign (7th) bit = 1 */
    } else {
	mask = 0x55;		/* sign bit = 0 */
	pcm_val = -pcm_val - 8;
    }

    /* Convert the scaled magnitude to segment number. */
    seg = search(pcm_val, seg_end, 8);

    /* Combine the sign, segment, and quantization bits. */

    if (seg >= 8)		/* out of range, return maximum value. */
	return 0x7F ^ mask;
    else {
	aval = seg << SEG_SHIFT;
	if (seg < 2)
	    aval |= (pcm_val >> 4) & QUANT_MASK;
	else
	    aval |= (pcm_val >> (seg + 3)) & QUANT_MASK;
	return aval ^ mask;
    }
}

/*
 * alaw2linear() - Convert an A-law value to 16-bit linear PCM
 *
 */
static int
alaw_to_linear(unsigned char	a_val)
{
    int		t;
    int		seg;

    a_val ^= 0x55;

    t = (a_val & QUANT_MASK) << 4;
    seg = ((unsigned)a_val & SEG_MASK) >> SEG_SHIFT;
    switch (seg) {
    case 0:
	t += 8;
	break;
    case 1:
	t += 0x108;
	break;
    default:
	t += 0x108;
	t <<= seg - 1;
    }
    return (a_val & SIGN_BIT) ? t : -t;
}


static int
get_ulaw(void **data)
{
    unsigned char *s = *data;
    int sample = ulaw_to_linear(*s++);
    *data = s;
    return sample;
}


static void
put_ulaw(void **data, int sample)
{
    unsigned char *s = *data;
    *s++ = linear_to_ulaw(sample);
    *data = s;
}


static int
get_alaw(void **data)
{
    unsigned char *s = *data;
    int sample = alaw_to_linear(*s++);
    *data = s;
    return sample;
}


static void
put_alaw(void **data, int sample)
{
    unsigned char *s = *data;
    *s++ = linear_to_alaw(sample);
    *data = s;
}


static int
get_s8(void **data)
{
    char *s = *data;
    int sample = *s++ << 8;
    *data = s;
    return sample;
}


static void
put_s8(void **data, int sample)
{
    char *s = *data;
    *s++ = sample >> 8;
    *data = s;
}


static int
get_s16(void **data)
{
    short *s = *data;
    int sample = *s++;
    *data = s;
    return sample;
}


static void
put_s16(void **data, int sample)
{
    short *s = *data;
    *s++ = sample;
    *data = s;
}


static int
get_u8(void **data)
{
    unsigned char *s = *data;
    int sample = (*s++ - 128) << 8;
    *data = s;
    return sample;
}


static void
put_u8(void **data, int sample)
{
    unsigned char *s = *data;
    *s++ = (sample >> 8) + 128;
    *data = s;
}


#if 0
static void
mono_to_mono(void	    *dest,
	     put_sample_func put,
	     void	    *src,
	     get_sample_func get,
	     int	     nsamples)
{
    int sample;
    while (--nsamples >= 0) {
	sample = get(&src);
	put(&dest, sample);
    }
}

static void
mono_to_stereo(void	      *dest,
	       put_sample_func put,
	       void	      *src,
	       get_sample_func get,
	       int	       nsamples)
{
    int sample;
    while (--nsamples >= 0) {
	sample = get(&src);
	put(&dest, sample);
	put(&dest, sample);
    }
}

static void
stereo_to_mono(void	      *dest,
	       put_sample_func put,
	       void	      *src,
	       get_sample_func get,
	       int	       nsamples)
{
    int sample_l;
    int sample_r;
    while (--nsamples >= 0) {
	sample_l = get(&src);
	sample_r = get(&src);
	put(&dest, (sample_l + sample_r) >> 1);
    }
}

static void
stereo_to_stereo(void	        *dest,
		 put_sample_func put,
		 void	        *src,
		 get_sample_func get,
		 int		 nsamples)
{
    int sample_l;
    int sample_r;
    while (--nsamples >= 0) {
	sample_l = get(&src);
	sample_r = get(&src);
	put(&dest, sample_l);
	put(&dest, sample_r);
    }
}
#endif


static void
sample_access_func(device_format_t *fmt,
		   get_sample_func *get,
		   put_sample_func *put)
{
    switch (fmt->encoding) {
    case AUDIO_ENCODING_ULAW:
	if (get) *get = get_ulaw;
	if (put) *put = put_ulaw;
	break;
    case AUDIO_ENCODING_ALAW:
	if (get) *get = get_alaw;
	if (put) *put = put_alaw;
	break;
    case AUDIO_ENCODING_LINEAR:
	if (fmt->precision == AUDIO_PRECISION_8) {
	    if (get) *get = get_s8;
	    if (put) *put = put_s8;
	} else {
	    if (get) *get = get_s16;
	    if (put) *put = put_s16;
	}
	break;
    case AUDIO_ENCODING_LINEAR8:
	if (get) *get = get_u8;
	if (put) *put = put_u8;
	break;
    }
}

#undef	abs
#define	abs(a)	((a) < 0 ? -(a) : (a))
#define	sample_rate_similar(a, b)	(abs((a) - (b)) <= (a) / 1000)

static int
fmt_equal(device_format_t *fmt1, device_format_t *fmt2)
{
    return fmt1->precision == fmt2->precision
	&& fmt1->encoding  == fmt2->encoding
	&& fmt1->channels  == fmt2->channels
	&& sample_rate_similar(fmt1->sample_rate, fmt2->sample_rate);
}

void
convert_samples(struct conv_ctx *cctx,
		void *dest, int dest_samples, int *ret_dest_samples,
		void *src,  int src_samples,  int *ret_src_samples)
{
    int i, j;
    int si, di;
    int nsamples;
    int max_chan;

    if (cctx->no_conv) {
	nsamples = min(src_samples, dest_samples);
	bcopy(src, dest, nsamples << cctx->src_shift);
	*ret_dest_samples = *ret_src_samples = nsamples;
	return;
    }

    if (cctx->no_src) {
	nsamples = min(src_samples, dest_samples);
	*ret_dest_samples = *ret_src_samples = nsamples;
	max_chan = max(cctx->src_fmt.channels, cctx->dst_fmt.channels);
	while (--nsamples >= 0) {
	    for (i = 0; i < cctx->src_fmt.channels; i++)
		cctx->cur_sample[i] = cctx->src_get_sample(&src);
	    for (j = 0; i < max_chan; i++, j++)
		cctx->cur_sample[i] = cctx->cur_sample[j];

	    if (cctx->dst_fmt.channels < cctx->src_fmt.channels) {
		for (i = 0; i < cctx->dst_fmt.channels; i++) {
		    int sum = 0;
		    int k = 0;
		    for (j = i; j < max_chan; j += cctx->dst_fmt.channels, k++) 
			sum += cctx->cur_sample[j];
		    cctx->cur_sample[j] = sum / k;
		}
	    }

	    for (i = 0; i < cctx->dst_fmt.channels; i++) 
		cctx->dst_put_sample(&dest, cctx->cur_sample[i]);
	}
	return;
    }

    max_chan = max(cctx->src_fmt.channels, cctx->dst_fmt.channels);

    for (si = di = 0; di < dest_samples; di++) {
	cctx->pos += cctx->dst_step;
	while (cctx->pos > cctx->src_step) {
	    if (si >= src_samples) 
		goto stop_src;

	    bcopy(cctx->cur_sample, cctx->last_sample, sizeof(cctx->cur_sample));
	    /* get new cur_sample */
	    for (i = 0; i < cctx->src_fmt.channels; i++)
		cctx->cur_sample[i] = cctx->src_get_sample(&src);
	    for (j = 0; i < max_chan; i++, j++)
		cctx->cur_sample[i] = cctx->cur_sample[j];

	    if (cctx->dst_fmt.channels < cctx->src_fmt.channels) {
		for (i = 0; i < cctx->dst_fmt.channels; i++) {
		    int sum = 0;
		    int k = 0;
		    for (j = i; j < max_chan; j += cctx->dst_fmt.channels, k++) 
			sum += cctx->cur_sample[j];
		    cctx->cur_sample[j] = sum / k;
		}
	    }

	    si++;
	    cctx->pos -= cctx->src_step;
	}
	/* interpolate tmp_sample from last_sample&cur_sample (pos ./. src_step) */
	for (i = 0; i < cctx->dst_fmt.channels; i++) {
	    int sample;
	    sample = cctx->last_sample[i];
	    sample += (cctx->cur_sample[i] - cctx->last_sample[i]) * cctx->pos / cctx->src_step;
	    cctx->dst_put_sample(&dest, sample);
	}

    }
 stop_src:
    *ret_src_samples = si;
    *ret_dest_samples = di;
}


/*
 * compute greatest common divisor
 */
static uint_t
gcd(uint_t a, uint_t b)
{
    uint_t tmp;

    while ((tmp = a%b) != 0) {
	a = b;
	b = tmp;
    }
    return b;
}


/*
 * compute least common multiple
 */
static uint_t
lcm(uint_t a, uint_t b)
{
    /* divide before multiply, to ovoid overflow, if possible */
    return a / gcd(a,b) * b;
}


/*
 * initialize the sample converter context, converting from
 * |srcfmt| to |destfmt|.
 */
void
convert_init(struct conv_ctx *cctx,
	     device_format_t *destfmt,
	     device_format_t *srcfmt)
{
    cctx->src_fmt.sample_rate = srcfmt->sample_rate;
    cctx->src_fmt.channels    = srcfmt->channels;
    cctx->src_fmt.precision   = srcfmt->precision;
    cctx->src_fmt.encoding    = srcfmt->encoding;

    cctx->src_shift = 0;
    if (cctx->src_fmt.precision == 16) cctx->src_shift++;
    if (cctx->src_fmt.channels  ==  2) cctx->src_shift++;

    cctx->dst_fmt.sample_rate = destfmt->sample_rate;
    cctx->dst_fmt.channels    = destfmt->channels;
    cctx->dst_fmt.precision   = destfmt->precision;
    cctx->dst_fmt.encoding    = destfmt->encoding;

    cctx->dst_shift = 0;
    if (cctx->dst_fmt.precision == 16) cctx->dst_shift++;
    if (cctx->dst_fmt.channels  ==  2) cctx->dst_shift++;

    if (!(cctx->no_conv = fmt_equal(&cctx->src_fmt, &cctx->dst_fmt))) {
	sample_access_func(&cctx->src_fmt, &cctx->src_get_sample, NULL);
	sample_access_func(&cctx->dst_fmt, NULL, &cctx->dst_put_sample);

	/*
	 * we don't use the software sample rate converter when the
	 * difference between the two sample rates is very small
	 */
	if (!(cctx->no_src = sample_rate_similar(cctx->src_fmt.sample_rate,
						 cctx->dst_fmt.sample_rate))) {
	    /*
	     * "optimize" the rates a bit, to avoid overflows in the
	     * lcm computation, if possible.
	     */
	    cctx->src_fmt.sample_rate &= ~3;
	    cctx->dst_fmt.sample_rate &= ~3;
	    cctx->base_freq = lcm(cctx->src_fmt.sample_rate,
				  cctx->dst_fmt.sample_rate);
	    cctx->src_step = cctx->base_freq / cctx->src_fmt.sample_rate;
	    cctx->dst_step = cctx->base_freq / cctx->dst_fmt.sample_rate;
	    cctx->pos = cctx->src_step - cctx->dst_step;
	    bzero(cctx->last_sample, sizeof(cctx->last_sample));
	    bzero(cctx->cur_sample,  sizeof(cctx->cur_sample));

	    if (!cctx->src_step || !cctx->dst_step) {
		/* still got an overflow, give up... */
		cctx->no_src = 1;
	    }
	}
    }
}

void
convert_cleanup(struct conv_ctx *cctx)
{
}
