/*-
 * Copyright (C) 2001-2003 Juergen Keil.
 * Copyright (C) 2001-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if	MIXER
#include <sys/mixer.h>
#endif
#include <sys/audio/audio_helper.h>

#include "audiohlp.h"
#include <sys/ddi.h>

#define	SRC_ENABLE 1	/* 1: enable the software sample rate converter */


static int
audio_hlp_check_sr(ah_ad_sample_rates_t *ad_srs, uint_t sample_rate,
		   device_format_t *device_fmt)
{
    int i;
    uint_t max_sample_rate;

    switch (ad_srs->ad_limits) {
    case AD_SRS_FLAG_SR_NOT_LIMITS:
	max_sample_rate = 0;
	for (i = 0; ad_srs->ad_srs[i]; i++) {
	    if (sample_rate == ad_srs->ad_srs[i]) {
		device_fmt->sample_rate = sample_rate;
		return AUDIO_SUCCESS;
	    }
	    if (ad_srs->ad_srs[i] > max_sample_rate)
		max_sample_rate = ad_srs->ad_srs[i];
	}
#if	SRC_ENABLE
	if (max_sample_rate > 0) {
	    device_fmt->sample_rate = max_sample_rate;
	    return AUDIO_SUCCESS;
	}
#endif
	return AUDIO_FAILURE;
    case AD_SRS_FLAG_SR_LIMITS:
	if (sample_rate >= ad_srs->ad_srs[0]
	    && sample_rate <= ad_srs->ad_srs[1]) {
	    device_fmt->sample_rate = sample_rate;
	    return AUDIO_SUCCESS;
	}
#if	SRC_ENABLE
	device_fmt->sample_rate = ad_srs->ad_srs[1];
	return AUDIO_SUCCESS;
#else
	return AUDIO_FAILURE;
#endif

    default:
	return AUDIO_FAILURE;
    }
}


static int
audio_hlp_find_cap(ah_ad_cap_comb_t *cap_comb, int prec, int enc)
{
    int i;
    
    for (i = 0; cap_comb[i].ad_prec; i++) {
	if (cap_comb[i].ad_prec == prec && cap_comb[i].ad_enc == enc) 
	    return AUDIO_SUCCESS;
    }
    return AUDIO_FAILURE;
}


static int
audio_hlp_find_chs(uint_t *chs, int channels)
{
    int i;
    
    for (i = 0; chs[i]; i++) {
	if (chs[i] == channels)
	    return AUDIO_SUCCESS;
    }
    return AUDIO_FAILURE;
}


static int
audio_hlp_check_datafmt(ah_ad_ch_cap_t *ch_cap,
			ah_ad_cap_comb_t *cap_comb,
			audio_prinfo_t *fmt,
			device_format_t *device_fmt)
{
    static int fmtconv[] = {
	AUDIO_PRECISION_8, AUDIO_ENCODING_ULAW,
		AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR,
	AUDIO_PRECISION_8, AUDIO_ENCODING_ALAW,
		AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR,
	AUDIO_PRECISION_8, AUDIO_ENCODING_LINEAR,
		AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR,
	0
    };
    int i;
    
    if (audio_hlp_find_cap(cap_comb, fmt->precision, fmt->encoding) 
						== AUDIO_SUCCESS) {
	device_fmt->precision = fmt->precision;
	device_fmt->encoding  = fmt->encoding;
    } else {
	for (i = 0; fmtconv[i]; i += 4) {
	    if (fmtconv[i] == fmt->precision
		&& fmtconv[i+1] == fmt->encoding
		&& audio_hlp_find_cap(cap_comb, fmtconv[i+2], fmtconv[i+3])
						== AUDIO_SUCCESS) {
		device_fmt->precision = fmtconv[i+2];
		device_fmt->encoding  = fmtconv[i+3];
		break;
	    }
	}
	if (!fmtconv[i])
	    return AUDIO_FAILURE;
    }

    if (audio_hlp_find_chs(ch_cap->ad_chs, fmt->channels) == AUDIO_SUCCESS) {
	device_fmt->channels  = fmt->channels;
    } else if (fmt->channels == AUDIO_CHANNELS_MONO 
	       && audio_hlp_find_chs(ch_cap->ad_chs, AUDIO_CHANNELS_STEREO)
						== AUDIO_SUCCESS) {
	device_fmt->channels = AUDIO_CHANNELS_STEREO;
    } else {
	return AUDIO_FAILURE;
    }

    return AUDIO_SUCCESS;
}


static void
audio_hlp_configure_playback(audiohdl_t	      handle,
			     audio_info_t    *newai,
			     device_format_t *newpf,
			     int	      vol_changes,
			     int	      fmt_changes)
{
    audio_prinfo_t *pinfo = &newai->play;
    device_format_t srcfmt;

    if (fmt_changes) {
	AD_SET_FORMAT(handle, AUDIO_NO_CHANNEL, AUDIO_PLAY,
		      newpf->sample_rate, newpf->channels,
		      newpf->precision, newpf->encoding);
	srcfmt.sample_rate = pinfo->sample_rate;
	srcfmt.channels    = pinfo->channels;
	srcfmt.precision   = pinfo->precision;
	srcfmt.encoding    = pinfo->encoding;
	convert_init(&handle->play_conv_ctx, newpf, &srcfmt);
    }

    if (vol_changes & 1<<AD_SET_GAIN)
	AD_SET_CONFIG(handle, AUDIO_NO_CHANNEL,
		      AD_SET_GAIN, AUDIO_PLAY, pinfo->gain, 0);

    if (vol_changes & 1<<AD_SET_GAIN_BAL)
	AD_SET_CONFIG(handle, AUDIO_NO_CHANNEL,
		      AD_SET_GAIN_BAL, AUDIO_PLAY, pinfo->balance, 0);

    if (vol_changes & 1<<AD_SET_PORT)
	AD_SET_CONFIG(handle, AUDIO_NO_CHANNEL,
		      AD_SET_PORT, AUDIO_PLAY, pinfo->port, 0);

    if (vol_changes & 1<<AD_OUTPUT_MUTE)
	AD_SET_CONFIG(handle, AUDIO_NO_CHANNEL,
		      AD_OUTPUT_MUTE, AUDIO_PLAY, newai->output_muted, 0);
}


static void
audio_hlp_configure_record(audiohdl_t	    handle,
			   audio_info_t    *newai,
			   device_format_t *newrf,
			   int		    vol_changes,
			   int		    fmt_changes)
{
    audio_prinfo_t *rinfo = &newai->record;
    device_format_t dstfmt;

    if (fmt_changes) {
	AD_SET_FORMAT(handle, AUDIO_NO_CHANNEL, AUDIO_RECORD,
		      newrf->sample_rate, newrf->channels,
		      newrf->precision, newrf->encoding);
	dstfmt.sample_rate = rinfo->sample_rate;
	dstfmt.channels    = rinfo->channels;
	dstfmt.precision   = rinfo->precision;
	dstfmt.encoding    = rinfo->encoding;
	convert_init(&handle->record_conv_ctx, &dstfmt, newrf);
    }

    if (vol_changes & 1<<AD_SET_GAIN)
	AD_SET_CONFIG(handle, AUDIO_NO_CHANNEL,
		      AD_SET_GAIN, AUDIO_RECORD, rinfo->gain, 0);

    if (vol_changes & 1<<AD_SET_GAIN_BAL)
	AD_SET_CONFIG(handle, AUDIO_NO_CHANNEL,
		      AD_SET_GAIN_BAL, AUDIO_RECORD, rinfo->balance, 0);

    if (vol_changes & 1<<AD_SET_PORT)
	AD_SET_CONFIG(handle, AUDIO_NO_CHANNEL,
		      AD_SET_PORT, AUDIO_RECORD, rinfo->port, 0);

    if (vol_changes & 1<<AD_SET_MONITOR_GAIN)
	AD_SET_CONFIG(handle, AUDIO_NO_CHANNEL,
		      AD_SET_MONITOR_GAIN, AUDIO_RECORD, newai->monitor_gain, 0);
}


int
audio_hlp_audio_setdefaults(audiohdl_t	    handle)
{
    if (audio_hlp_check_sr(&handle->ah_ad_info->ad_play.ad_srs,
			   handle->audio_info.play.sample_rate,
			   &handle->device_play_format) != AUDIO_SUCCESS) {
	return AUDIO_FAILURE;
    }

    if (audio_hlp_check_sr(&handle->ah_ad_info->ad_record.ad_srs,
			   handle->audio_info.record.sample_rate,
			   &handle->device_record_format) != AUDIO_SUCCESS) {
	return AUDIO_FAILURE;
    }

    if (audio_hlp_check_datafmt(&handle->ah_ad_info->ad_play,
				handle->ah_ad_info->ad_play_comb,
				&handle->audio_info.play,
				&handle->device_play_format) != AUDIO_SUCCESS) {
	return AUDIO_FAILURE;
    }

    if (audio_hlp_check_datafmt(&handle->ah_ad_info->ad_record,
				handle->ah_ad_info->ad_rec_comb,
				&handle->audio_info.record,
				&handle->device_record_format) != AUDIO_SUCCESS) {
	return AUDIO_FAILURE;
    }

    audio_hlp_configure_playback(handle,
				 &handle->audio_info,
				 &handle->device_play_format, 
				 ~0, 1);
    audio_hlp_configure_record(handle,
			       &handle->audio_info,
			       &handle->device_record_format,
			       ~0, 1);

    return AUDIO_SUCCESS;
}


static int
audio_hlp_audio_setinfo(audiohdl_t	    handle,
			audio_info_t	   *audio_info,
			queue_t		   *q)
{
    int vol_changes;
    int pfmt_changes;
    int rfmt_changes;
    int changes = 0;
    int play_pause_change = 0;
    int record_pause_change = 0;
    audio_info_t newai;
    device_format_t newpf;
    device_format_t newrf;
    int audio_flags;
    dev_t dev = audio_hlp_get_qptr_dev(q);

    if (AUDIOHLP_MINOR(dev) >= AUDIOHLP_CLONE_MIN)
	audio_flags = handle->minor_inuse[AUDIOHLP_MINOR(dev) - AUDIOHLP_CLONE_MIN];
    else
	audio_flags = 0;

    /*
     * Collect the user's changes on a temporary variable "newai",
     * while we validate them.  If the changes are all OK,
     * copy "newai" back into the global device state.
     */
    bcopy(&handle->audio_info, &newai, sizeof(audio_info_t));
    bcopy(&handle->device_play_format, &newpf, sizeof(device_format_t));
    bcopy(&handle->device_record_format, &newrf, sizeof(device_format_t));

    /*
     * note: sample data format is only changable on the 
     * non-ctl device
     */
    pfmt_changes = 0;
    if (!AUDIOHLP_ISCTL(dev) && (audio_flags & AUDIO_PLAY)) {

	/* copy configuration parameters for the playback side */

	if (audio_info->play.sample_rate != ~0) {
	    if (audio_hlp_check_sr(&handle->ah_ad_info->ad_play.ad_srs,
				   audio_info->play.sample_rate, &newpf)
						!= AUDIO_SUCCESS) {
		AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: play.sample_rate");
		return AUDIO_FAILURE;
	    }
	    pfmt_changes |= newai.play.sample_rate != audio_info->play.sample_rate;
	    newai.play.sample_rate = audio_info->play.sample_rate;
	}

	if (audio_info->play.channels != ~0) {
	    if (audio_info->play.channels != AUDIO_CHANNELS_MONO &&
		audio_info->play.channels != AUDIO_CHANNELS_STEREO) {
		AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: play.channels");
		return AUDIO_FAILURE;
	    }
	    pfmt_changes |= newai.play.channels != audio_info->play.channels;
	    newai.play.channels = audio_info->play.channels;
	}

	if (audio_info->play.encoding != ~0) {
	    switch (audio_info->play.encoding) {
	    case AUDIO_ENCODING_ULAW:
	    case AUDIO_ENCODING_ALAW:
	    case AUDIO_ENCODING_LINEAR:
	    case AUDIO_ENCODING_LINEAR8:
		pfmt_changes |= newai.play.encoding != audio_info->play.encoding; 
		newai.play.encoding = audio_info->play.encoding;
		break;
	    default:
		AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: play.encoding");
		return AUDIO_FAILURE;
	    }
	}
	
	if (audio_info->play.precision != ~0) {
	    if (audio_info->play.precision == AUDIO_PRECISION_8 ||
		( audio_info->play.precision == AUDIO_PRECISION_16 &&
		  newai.play.encoding  == AUDIO_ENCODING_LINEAR) ) {
		
		pfmt_changes |= newai.play.precision != audio_info->play.precision;
		newai.play.precision = audio_info->play.precision;
	    } else {
		AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: play.precision");
		return AUDIO_FAILURE;
	    }
	}

	if (audio_hlp_check_datafmt(&handle->ah_ad_info->ad_play,
				    handle->ah_ad_info->ad_play_comb,
				    &newai.play, &newpf) != AUDIO_SUCCESS) {
	    AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: play.{encoding,precision}");
	    return AUDIO_FAILURE;
	}

	if (audio_info->play.pause != (uchar_t)~0 &&
	    !newai.play.pause != !audio_info->play.pause) {

	    changes |= 1;
	    newai.play.pause = !!audio_info->play.pause;
	    play_pause_change = 1;
	}
    }
    changes |= pfmt_changes;

    rfmt_changes = 0;
    if (!AUDIOHLP_ISCTL(dev) && (audio_flags & AUDIO_RECORD)) {

	/* copy configuation parameters for the recording side */

	if (audio_info->record.sample_rate != ~0) {
	    if (audio_hlp_check_sr(&handle->ah_ad_info->ad_record.ad_srs,
				   audio_info->record.sample_rate, &newrf)
						!= AUDIO_SUCCESS) {
		AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: record.sample_rate");
		return AUDIO_FAILURE;
	    }
	    rfmt_changes |= newai.record.sample_rate != audio_info->record.sample_rate;
	    newai.record.sample_rate = audio_info->record.sample_rate;
	}


	if (audio_info->record.channels != ~0) {
	    if (audio_info->record.channels != AUDIO_CHANNELS_MONO &&
		audio_info->record.channels != AUDIO_CHANNELS_STEREO) {
		AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: record.channels");
		return AUDIO_FAILURE;
	    }
	    rfmt_changes |= newai.record.channels != audio_info->record.channels;
	    newai.record.channels = audio_info->record.channels;
	}

	if (audio_info->record.encoding != ~0) {
	    switch (audio_info->record.encoding) {
	    case AUDIO_ENCODING_ULAW:
	    case AUDIO_ENCODING_ALAW:
	    case AUDIO_ENCODING_LINEAR:
	    case AUDIO_ENCODING_LINEAR8:
		rfmt_changes |= newai.record.encoding != audio_info->record.encoding;
		newai.record.encoding = audio_info->record.encoding;
		break;
	    default:
		AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: record.encoding");
		return AUDIO_FAILURE;
	    }
	}

	if (audio_info->record.precision != ~0) {
	    if (audio_info->record.precision == AUDIO_PRECISION_8 ||
		( audio_info->record.precision == AUDIO_PRECISION_16 &&
		  newai.record.encoding == AUDIO_ENCODING_LINEAR) ) {

		rfmt_changes |= newai.record.precision != audio_info->record.precision;
		newai.record.precision = audio_info->record.precision;
	    } else {
		AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: record.precision");
		return AUDIO_FAILURE;
	    }
	}

	if (audio_hlp_check_datafmt(&handle->ah_ad_info->ad_record,
				    handle->ah_ad_info->ad_rec_comb,
				    &newai.record, &newrf) != AUDIO_SUCCESS) {
	    AUD_DPRINTF_L3(AUDM_ALL, "AUDIO_SETINFO failed: record.{encoding,precision}");
	    return AUDIO_FAILURE;
	}

	if (audio_info->record.pause != (uchar_t)~0 &&
	    !newai.record.pause != !audio_info->record.pause) {
	    changes |= 1;
	    newai.record.pause = !!audio_info->record.pause;
	    record_pause_change = 1;
	}
    }
    changes |= rfmt_changes;

    /* playback volume control */
    vol_changes = 0;
    if (audio_info->play.gain != ~0) {
	if (audio_info->play.gain != newai.play.gain)
	    vol_changes |= 1 << AD_SET_GAIN;
	if (audio_info->play.gain < AUDIO_MIN_GAIN)
	    newai.play.gain = AUDIO_MIN_GAIN;
	else if (audio_info->play.gain > AUDIO_MAX_GAIN)
	    newai.play.gain = AUDIO_MAX_GAIN;
	else
	    newai.play.gain = audio_info->play.gain;
    }
    if (audio_info->play.balance != (uchar_t)~0) {
	if (audio_info->play.balance != newai.play.balance)
	    vol_changes |= 1 << AD_SET_GAIN_BAL;
	/*
	if (audio_info->play.balance < AUDIO_LEFT_BALANCE)
	    newai.play.balance = AUDIO_LEFT_BALANCE;
	*/
	if (audio_info->play.balance > AUDIO_RIGHT_BALANCE)
	    newai.play.balance = AUDIO_RIGHT_BALANCE;
	else
	    newai.play.balance = audio_info->play.balance;
    }
    if (audio_info->play.port != ~0) {
	if (audio_info->play.port != newai.play.port)
	    vol_changes |= 1 << AD_SET_PORT;
	newai.play.port = audio_info->play.port;
    }
    if (audio_info->output_muted != (uchar_t)~0) {
	if (audio_info->output_muted != newai.output_muted) 
	    vol_changes |= 1 << AD_OUTPUT_MUTE;
	newai.output_muted = audio_info->output_muted;
    }

    changes |= vol_changes;

    /*
     * If there were changes to the playback volume parameters above,
     * configure the audio codec hardware accordingly.
     */
    audio_hlp_configure_playback(handle, &newai, &newpf, 
				 vol_changes, pfmt_changes);

    /* record volume control */
    vol_changes = 0;
    if (audio_info->record.gain != ~0) {
	if (audio_info->record.gain != newai.record.gain)
	    vol_changes |= 1 << AD_SET_GAIN;
	if (audio_info->record.gain < AUDIO_MIN_GAIN)
	    newai.record.gain = AUDIO_MIN_GAIN;
	else if (audio_info->record.gain > AUDIO_MAX_GAIN)
	    newai.record.gain = AUDIO_MAX_GAIN;
	else
	    newai.record.gain = audio_info->record.gain;
    }
    if (audio_info->record.balance != (uchar_t)~0) {
	if (audio_info->record.balance != newai.record.balance)
	    vol_changes |= 1 << AD_SET_GAIN_BAL;
	/*
	if (audio_info->record.balance < AUDIO_LEFT_BALANCE)
	    newai.record.balance = AUDIO_LEFT_BALANCE;
	*/
	if (audio_info->record.balance > AUDIO_RIGHT_BALANCE)
	    newai.record.balance = AUDIO_RIGHT_BALANCE;
	else
	    newai.record.balance = audio_info->record.balance;
    }
    if (audio_info->record.port != ~0) {
	if (audio_info->record.port != newai.record.port)
	    vol_changes |= 1 << AD_SET_PORT;
	newai.record.port = audio_info->record.port;
    }
    if (audio_info->monitor_gain != ~0) {
	if (audio_info->monitor_gain != newai.monitor_gain) 
	    vol_changes |= 1 << AD_SET_MONITOR_GAIN;
	if (audio_info->monitor_gain < AUDIO_MIN_GAIN)
	    newai.monitor_gain = AUDIO_MIN_GAIN;
	else if (audio_info->monitor_gain > AUDIO_MAX_GAIN)
	    newai.monitor_gain = AUDIO_MAX_GAIN;
	else
	    newai.monitor_gain = audio_info->monitor_gain;
    }

    changes |= vol_changes;

    /*
     * If there were changes to the recording volume parameters above,
     * configure the hardware accordingly.
     */
    audio_hlp_configure_record(handle, &newai, &newrf,
			       vol_changes, rfmt_changes);

    if (audio_info->play.waiting != (uchar_t)~0) {
	changes |= !newai.play.waiting != !audio_info->play.waiting;
	newai.play.waiting = audio_info->play.waiting;
    }
    if (audio_info->record.waiting != (uchar_t)~0) {
	changes |= !newai.record.waiting != !audio_info->record.waiting;
	newai.record.waiting = audio_info->record.waiting;
    }

    /*
     * save the new audio data encoding and volume parameters 
     * into our state structure
     */
    bcopy(&newai, &handle->audio_info, sizeof(audio_info_t));
    bcopy(&newpf, &handle->device_play_format, sizeof(device_format_t));
    bcopy(&newrf, &handle->device_record_format, sizeof(device_format_t));

    /*
     * special handling for {play,record}.{samples,eof,error}:
     * if these fields are set, update our state structure with
     * the data from the AUDIO_SETINFO ioctl;  modify these fields
     * in the user supplied structure to reflect the state of the
     * device at the time the AUDIO_SETINFO ioctl was issued. 
     */
    if (audio_info->play.samples != ~0) {
	handle->audio_info.play.samples = audio_info->play.samples;
    }
    if (audio_info->play.eof != ~0) {
	handle->audio_info.play.eof = audio_info->play.eof;
    }
    if (audio_info->play.error != (uchar_t)~0) {
	handle->audio_info.play.error = audio_info->play.error;
    }

    if (audio_info->record.samples != ~0) {
	handle->audio_info.record.samples = audio_info->record.samples;
    }
    if (audio_info->record.eof != ~0) {
	handle->audio_info.record.eof = audio_info->record.eof;
    }
    if (audio_info->record.error != (uchar_t)~0) {
	handle->audio_info.record.error = audio_info->record.error;
    }

    bcopy(&newai, audio_info, sizeof(audio_info_t));

    if (play_pause_change) {
	if (newai.play.pause) 
	    AD_PAUSE_PLAY(handle, AUDIO_NO_CHANNEL);
	else
	    AD_START_PLAY(handle, AUDIO_NO_CHANNEL);
    }
    if (record_pause_change) {
	if (newai.record.pause) {
	    AD_STOP_RECORD(handle, AUDIO_NO_CHANNEL);
	} else {
	    AD_START_RECORD(handle, AUDIO_NO_CHANNEL);
	}
    }

    /* send audio status changes notifications */
    if (changes) {
	audio_hlp_status_changed(handle);
    }

    return AUDIO_SUCCESS;
}


/*
 * STREAMS helper to copyin ioctl request structure.
 *
 * We construct a M_COPYIN request (copyreq(9S)) and send it upstream
 * to the streams head.  The streams head performs the requested
 * copyin operation and sends the data downstream with a M_IOCDATA
 * message.
 */
static int
audio_hlp_ioctl_copyin(mblk_t  *mp,
		       caddr_t  arg,
		       size_t   datalen,
		       int     *error)
{
    /* struct iocblk *iocp = (struct iocblk *)mp->b_rptr; */
    struct copyreq *copyreq =  (struct copyreq *)mp->b_rptr;

    if (mp->b_cont) {
	freemsg(mp->b_cont);
	mp->b_cont = NULL;
    }
    mp->b_datap->db_type = M_COPYIN;
    mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
/*
    copyreq->cq_cmd     = iocp->ioc_cmd;
    copyreq->cq_cr      = iocp->ioc_cr;
    copyreq->cq_id      = iocp->ioc_id;
 */
    copyreq->cq_flag    = 0;
    copyreq->cq_addr    = arg;
    copyreq->cq_size    = datalen;
    copyreq->cq_private = (void*)arg;
	
    return AD_WIOCDATA;
}


/*
 * STREAMS helper to copyout ioctl result structure
 *
 * We construct a M_COPYOUT request (copyreq(9S)) plus a data block
 * containing the copyout data and send it upstream to the streams
 * head.  The streams head performs the requested copyout operation
 * and sends back a M_IOCDATA message with the error codes for the
 * copyout operation.
 */
static int
audio_hlp_ioctl_copyout(mblk_t  *mp,
			caddr_t  arg,
			void    *data,
			size_t   datalen,
			int	*error)
{
    /* struct iocblk *iocp = (struct iocblk *)mp->b_rptr; */
    struct copyreq *copyreq =  (struct copyreq *)mp->b_rptr;
    mblk_t *mp2;

    if (mp->b_cont) {
	freemsg(mp->b_cont);
	mp->b_cont = NULL;
    }

    mp2 = allocb(datalen, BPRI_HI);
    if (mp2 == NULL) {
	cmn_err(CE_WARN, "cannot copyout ioctl data, out of streams blocks");
	*error = EAGAIN;
	return AD_NACK;
    }

    if (bpsize(mp2) < datalen) {
	/* oops message too small? */
	cmn_err(CE_WARN, "cannot copyout ioctl data, msg too small");
	freemsg(mp2);
	*error = ENOMEM;
	return AD_NACK;
    }
    bcopy(data, mp2->b_wptr, datalen);
    mp2->b_wptr += datalen;
	
    mp->b_cont = mp2;
    mp->b_datap->db_type = M_COPYOUT;
    mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
/*
    copyreq->cq_cmd     = iocp->ioc_cmd;
    copyreq->cq_cr      = iocp->ioc_cr;
    copyreq->cq_id      = iocp->ioc_id;
*/
    copyreq->cq_flag    = 0;
    copyreq->cq_addr    = arg;
    copyreq->cq_size    = datalen;
    copyreq->cq_private = (void*)arg;

    return AD_WIOCDATA;
}


int
audio_hlp_ioctl(audiohdl_t handle, queue_t *q, mblk_t *mp, int *error)
{
    struct iocblk *iocp = (struct iocblk *)mp->b_rptr;
    mblk_t *mp2;
    caddr_t arg = 0;
    int audio_flags;
    dev_t dev;

    AUD_DPRINTF_L4(AUDM_IOCTL, "audio_hlp_ioctl %x\n", iocp->ioc_cmd);

    if (iocp->ioc_count == TRANSPARENT && (mp2 = mp->b_cont) != NULL)
	/* get the "arg" value from the streams msg */
	bcopy(mp2->b_rptr, &arg, sizeof(arg));
    
    switch (iocp->ioc_cmd) {
    case AUDIO_DRAIN:
	if (AUDIOHLP_ISCTL(audio_hlp_get_qptr_dev(q)))
	    return AD_NACK;

	dev = audio_hlp_get_qptr_dev(q);
	if (AUDIOHLP_MINOR(dev) >= AUDIOHLP_CLONE_MIN)
	    audio_flags = handle->minor_inuse[AUDIOHLP_MINOR(dev) - AUDIOHLP_CLONE_MIN];
	else
	    audio_flags = 0;

	if (audio_flags & AUDIO_PLAY)
	    audio_hlp_audio_drain(q);

	return AD_ACK;

    case AUDIO_SETINFO:
	return audio_hlp_ioctl_copyin(mp, arg, sizeof(audio_info_t), error);

    case AUDIO_GETINFO:
	{
	    audio_info_t tmp_ai;
	    mutex_enter(&handle->lock);
	    bcopy(&handle->audio_info, &tmp_ai, sizeof(audio_info_t));
	    if (handle->ah_ad_info->ad_entry->ad_get_delay != NULL) {
		int delay = AD_GETDELAY(handle, AUDIO_NO_CHANNEL);
		/*cmn_err(CE_NOTE, "^%d - %d\n", tmp_ai.play.samples, delay);*/
		if (handle->device_play_format.sample_rate
		    		!= handle->audio_info.play.sample_rate
		    && handle->device_play_format.sample_rate > 0) {
		    delay = delay
			* handle->audio_info.play.sample_rate
			/ handle->device_play_format.sample_rate;
		}
		tmp_ai.play.samples -= delay;
	    }
	    mutex_exit(&handle->lock);
	    return audio_hlp_ioctl_copyout(mp, arg, &tmp_ai, sizeof(audio_info_t), error);
	}

    case AUDIO_GETDEV:
	return audio_hlp_ioctl_copyout(mp, arg, handle->ah_ad_info->ad_dev_info, sizeof(audio_device_t), error);

#if MIXER
    case AUDIO_MIXERCTL_GET_MODE:
	if (AUDIOHLP_ISCTL(audio_hlp_get_qptr_dev(q))) {
	    /* we only support mixer compat mode for now */
	    int mixer_mode = AM_COMPAT_MODE;
	    return audio_hlp_ioctl_copyout(mp, arg, &mixer_mode, sizeof(mixer_mode), error);
	} else {
	    /* AUDIO_MIXERCTL_GET_MODE is only allowed on the audioctl device */
	    return AD_NACK;
	}

    case AUDIO_MIXER_GET_SAMPLE_RATES:
	return audio_hlp_ioctl_copyin(mp, arg, AUDIO_MIXER_SAMP_RATES_STRUCT_SIZE(0), error);

    case AUDIO_GET_NUM_CHS:
	if (AUDIOHLP_ISCTL(audio_hlp_get_qptr_dev(q))) {
	    /* we only support two channels for now (play&record) */
	    int num_chs = 2;
	    return audio_hlp_ioctl_copyout(mp, arg, &num_chs, sizeof(num_chs), error);
	} else {
	    /* AUDIO_GET_NUM_CHS is only allowed on the audioctl device */
	    return AD_NACK;
	}

#endif

    default:
	return AD_IOCTL(handle, 0, q, mp, error);
    }
}


int
audio_hlp_iocdata(audiohdl_t handle, queue_t *q, mblk_t *mp, int *error)
{
    struct copyresp *copyresp = (struct copyresp *)mp->b_rptr;
    int rval;
    mblk_t *data_mp;
    caddr_t arg = (caddr_t) copyresp->cp_private;

    AUD_DPRINTF_L4(AUDM_IOCTL, "audio_hlp_iocdata %x\n", copyresp->cp_cmd);

    if (mp->b_cont == NULL) return AD_NACK;
    data_mp = mp->b_cont;

    switch (copyresp->cp_cmd) {

    case AUDIO_SETINFO:
	if (bpsize(data_mp) < sizeof(audio_info_t)) {
	    cmn_err(CE_WARN, "cannot copyin ioctl data, msg too small");
	    *error = ENOMEM;
	    return AD_ACK;
	}
	
	if (audio_hlp_audio_setinfo(handle,
				    (audio_info_t*)data_mp->b_rptr,
				    q) != AUDIO_SUCCESS) {
	    *error = EINVAL;
	    return AD_ACK;
	} else {
	    /* 
	     * audio_hlp_ioctl_copyout does a freemsg(mp->b_cont),  but we
	     * use this message's data to pass back data for the M_COPYOUT op.
	     *
	     * So, we do our own freemsg() for the mp->b_cont msg, after
	     * the audio_hlp_ioctl_copyout call when it's safe to free 
	     * the data.
	     */
	    mp->b_cont = NULL;
	    rval = audio_hlp_ioctl_copyout(mp, arg,
					   (audio_info_t*)data_mp->b_rptr,
					   sizeof(audio_info_t),
					   error);
	    freemsg(data_mp);
	    return rval;
	}

#if MIXER
    case AUDIO_MIXER_GET_SAMPLE_RATES:
	{
	    am_sample_rates_t *sample_rates;
	    int saved_num_samp_rates;

	    /* COPYIN was OK, ioctl arg data is in data_mp */
	    if (bpsize(data_mp) < AUDIO_MIXER_SAMP_RATES_STRUCT_SIZE(0)) {
		cmn_err(CE_WARN, "cannot copyin ioctl data, msg too small");
		*error = ENOMEM;
		return AD_ACK;
	    }

	    sample_rates = (am_sample_rates_t *)data_mp->b_rptr;
	    saved_num_samp_rates = sample_rates->num_samp_rates;
;
	    if (AUDIOHLP_ISCTL(audio_hlp_get_qptr_dev(q))
		|| (   sample_rates->type != AUDIO_PLAY
		    && sample_rates->type != AUDIO_RECORD)) {
		return AD_NACK;
	    } else {
		sample_rates->flags = MIXER_SR_LIMITS;
		sample_rates->num_samp_rates = 2;
		if (saved_num_samp_rates > 0)
		    sample_rates->samp_rates[0] = 4000;
		if (saved_num_samp_rates > 1)
		    sample_rates->samp_rates[1] = 48000;

		return audio_hlp_ioctl_copyout(mp, arg, sample_rates,
					       AUDIO_MIXER_SAMP_RATES_STRUCT_SIZE(saved_num_samp_rates), error);
	    }
	}
#endif

    default:
	return AD_IOCDATA(handle, 0, q, mp, error);
    }
}

